;---------------------------------------------------------------------
; Command-Line Interface                    Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

.386                                                                    ; Force 32-bit code
.model flat, stdcall                                                    ; Memory model & calling convention
option casemap :none                                                    ; Case sensitive

include windows.inc                                                     ; Windows typedefs
include kernel32.inc                                                    ; GetCommandLine, ExitProcess
include user32.inc                                                      ; wsprintf
include masm32.inc                                                      ; MemCopy, StdOut
include macros.asm                                                      ; fn

includelib kernel32.lib
includelib user32.lib
includelib masm32.lib

include ..\lib\macros.asm
include ..\VDI\VDI.asm
include ..\JV1\JV1.asm
include ..\JV3\JV3.asm
include ..\DMK\DMK.asm
include ..\OSI\OSI.asm
include ..\TRSDOS\TRSDOS.asm
include ..\DBLDOS\DBLDOS.asm
include ..\DOSPLUS\DOSPLUS.asm
include ..\MICRODOS\MICRODOS.asm
include ..\MULTIDOS\MULTIDOS.asm
include ..\NEWDOS80\NEWDOS80.asm
include ..\RAPIDOS\RAPIDOS.asm
include ..\TRSDOS3\TRSDOS3.asm
include ..\TRSDOS1\TRSDOS1.asm
include ..\TRSDOS1M\TRSDOS1M.asm

;---------------------------------------------------------------------
; Function Prototypes
;---------------------------------------------------------------------

LoadVDI         proto   :DWORD, :HANDLE, :DWORD
LoadOSI         proto   :DWORD, :DWORD, :DWORD
UnloadVDI       proto   :DWORD
UnloadOSI       proto   :DWORD
SetTrsFileDate  proto   :DWORD, :DWORD
SetWinFileDate  proto   :HANDLE, :DWORD
SetTrsFileAttr  proto   :DWORD, :DWORD
SetWinFileAttr  proto   :HANDLE, :DWORD
DumpSector      proto   :DWORD, :DWORD
CompFspec       proto   :DWORD, :DWORD, :DWORD
TrsCat          proto   :DWORD, :DWORD, :DWORD
Trs2Win         proto   :DWORD, :DWORD
Win2Trs         proto   :DWORD, :DWORD
ParseCmdLine    proto
GetFnPtr        proto   :DWORD, :DWORD, :BYTE
MakeArgList     proto   :DWORD, :DWORD, :DWORD
AddArg          proto   :DWORD, :DWORD, :DWORD
ToLower         proto   :DWORD
PrintError      proto   :DWORD

.data

;---------------------------------------------------------------------
; Global Variables
;---------------------------------------------------------------------

;MAX_ARG         equ         16

argv            DWORD       16 dup (?)                                  ; Pointers to argument pointer array
argc            BYTE        0                                           ; Command line argument counter

g_hFile         HANDLE      ?                                           ; Global handle to image file

g_pVDI          DWORD       ?                                           ; Global pointer to VDI object
g_pOSI          DWORD       ?                                           ; Global pointer to OSI object

g_Mask          BYTE        "*.*", 0                                    ; Default search mask

;---------------------------------------------------------------------
; Interface Configuration
;---------------------------------------------------------------------

CONFIG  struct
    File        DWORD       3   dup (NULL)                              ; Pointers to exe file, dsk file and optional file mask
    Command     DWORD       NULL                                        ; Requested operation function pointer
    VDI         DWORD       NULL                                        ; Virtual Disk Interface VTBL
    OSI         DWORD       NULL                                        ; Operating System Interface VTBL
    System      BYTE        0                                           ; Include System files flag
    Invisible   BYTE        0                                           ; Include Invisible files flag
    Attrib      BYTE        0                                           ; Preserve file attributes
    Yes         BYTE        0                                           ; Do no ask for confirmations
CONFIG  ends

g_Config        CONFIG      <>                                          ; User-defined settings

;---------------------------------------------------------------------
; Parameters Array
;---------------------------------------------------------------------

ARG struct
    ArgName     DWORD       ?                                           ; Argument name
    ArgCall     DWORD       ?                                           ; Pointer to handler function
    ArgPass     DWORD       ?                                           ; Value to pass to handler funtion
ARG ends

Arg_List        BYTE        "l", 0                                      ; List files
Arg_Read        BYTE        "r", 0                                      ; Read files
Arg_Write       BYTE        "w", 0                                      ; Write files
Arg_Delete      BYTE        "k", 0                                      ; Delete files
Arg_Dump        BYTE        "d", 0                                      ; Dump sectors
Arg_Sys         BYTE        "s", 0                                      ; Include system files
Arg_Inv         BYTE        "i", 0                                      ; Include invisible files
Arg_Attr        BYTE        "a", 0                                      ; Preserve file attributes
Arg_Yes         BYTE        "y", 0                                      ; Do not ask for confirmations
Arg_JV1         BYTE        "jv1", 0                                    ; JV1
Arg_JV3         BYTE        "jv3", 0                                    ; JV3
Arg_DMK         BYTE        "dmk", 0                                    ; DMK
Arg_CPM         BYTE        "cp", 0                                     ; CP/M
Arg_DD          BYTE        "dd", 0                                     ; DBLDOS
Arg_DP          BYTE        "dp", 0                                     ; DOSPLUS
Arg_MC          BYTE        "mc", 0                                     ; MICRODOS
Arg_MD          BYTE        "md", 0                                     ; MULTIDOS
Arg_ND          BYTE        "nd", 0                                     ; NEWDOS/80
Arg_OS          BYTE        "os", 0                                     ; OS-80
Arg_RD          BYTE        "rd", 0                                     ; RAPIDOS
Arg_T4          BYTE        "t4", 0                                     ; TRSDOS Model 4
Arg_T3          BYTE        "t3", 0                                     ; TRSDOS Model III
Arg_T1          BYTE        "t1", 0                                     ; TRSDOS Model I, Single Density
Arg_TM          BYTE        "tm", 0                                     ; TRSDOS Model I, Mixed Density

Arg_Array       ARG         <offset Arg_List,   offset SetCmd,  offset List>
                ARG         <offset Arg_Read,   offset SetCmd,  offset Read>
                ARG         <offset Arg_Write,  offset SetCmd,  offset Write>
                ARG         <offset Arg_Delete, offset SetCmd,  offset Delete>
                ARG         <offset Arg_Dump,   offset SetCmd,  offset Dump>
                ARG         <offset Arg_Sys,    offset SetFlag, offset g_Config.System>
                ARG         <offset Arg_Inv,    offset SetFlag, offset g_Config.Invisible>
                ARG         <offset Arg_Attr,   offset SetFlag, offset g_Config.Attrib>
                ARG         <offset Arg_Yes,    offset SetFlag, offset g_Config.Yes>
                ARG         <offset Arg_JV1,    offset SetFF,   offset JV1>
                ARG         <offset Arg_JV3,    offset SetFF,   offset JV3>
                ARG         <offset Arg_DMK,    offset SetFF,   offset DMK>
                ARG         <offset Arg_CPM,    offset SetOS,   offset NULL>
                ARG         <offset Arg_DD,     offset SetOS,   offset DBLDOS>
                ARG         <offset Arg_DP,     offset SetOS,   offset DOSPLUS>
                ARG         <offset Arg_MC,     offset SetOS,   offset MICRODOS>
                ARG         <offset Arg_MD,     offset SetOS,   offset MULTIDOS>
                ARG         <offset Arg_ND,     offset SetOS,   offset NEWDOS80>
                ARG         <offset Arg_OS,     offset SetOS,   offset NULL>
                ARG         <offset Arg_RD,     offset SetOS,   offset RAPIDOS>
                ARG         <offset Arg_T4,     offset SetOS,   offset TRSDOS>
                ARG         <offset Arg_T3,     offset SetOS,   offset TRSDOS3>
                ARG         <offset Arg_T1,     offset SetOS,   offset TRSDOS1>
                ARG         <offset Arg_TM,     offset SetOS,   offset TRSDOS1M>

Arg_Count       equ         24

;---------------------------------------------------------------------
; Messages
;---------------------------------------------------------------------

Msg_Rights      BYTE        "TRS-80 Virtual Floppy Disk Manager v0.8", 0Dh, 0Ah
                BYTE        "Copyright (c) 2009-2010 Miguel Dutra", 0Dh, 0Ah
                BYTE        "Icon created by Marco Martin (www.notmart.org)", 0Dh, 0Ah
                BYTE        "Get the latest version at www.mdutra.com/vdisk", 0Dh, 0Ah, 0

Msg_Usage       BYTE        0Dh, 0Ah
                BYTE        "Usage: vdisk [switches] <diskfile> [filespec]", 0Dh, 0Ah, 0Dh, 0Ah
                BYTE        "Commands:", 0Dh, 0Ah
                BYTE        09h, "-l", 09h, 09h, "List disk directory", 0Dh, 0Ah
                BYTE        09h, "-r", 09h, 09h, "Read files from the disk", 0Dh, 0Ah
                BYTE        09h, "-w", 09h, 09h, "Write files to the disk", 0Dh, 0Ah
                BYTE        09h, "-k", 09h, 09h, "Delete (kill) files", 0Dh, 0Ah
                BYTE        09h, "-d", 09h, 09h, "Dump disk sectors", 0Dh, 0Ah
                BYTE        0Dh, 0Ah
                BYTE        "Options:", 0Dh, 0Ah
                BYTE        09h, "-s", 09h, 09h, "Include System files", 0Dh, 0Ah
                BYTE        09h, "-i", 09h, 09h, "Include Invisible files", 0Dh, 0Ah
                BYTE        09h, "-a", 09h, 09h, "Preserve file attributes", 0Dh, 0Ah
                BYTE        09h, "-y", 09h, 09h, "Do not ask for confirmations", 0Dh, 0Ah
                BYTE        0Dh, 0Ah
                BYTE        "File Formats:", 0Dh, 0Ah
                BYTE        09h, "-jv1", 09h, 09h, "Force Jeff Vavasour's JV1 format", 0Dh, 0Ah
                BYTE        09h, "-jv3", 09h, 09h, "Force Jeff Vavasour's JV3 format", 0Dh, 0Ah
                BYTE        09h, "-dmk", 09h, 09h, "Force David Keil's DMK format", 0Dh, 0Ah
                BYTE        0Dh, 0Ah
                BYTE        "Operating Systems:", 0Dh, 0Ah
                BYTE        09h, "-cp", 09h, 09h, "Force a CP/M compatible interface *", 0Dh, 0Ah
                BYTE        09h, "-dd", 09h, 09h, "Force a DBLDOS compatible interface", 0Dh, 0Ah
                BYTE        09h, "-dp", 09h, 09h, "Force a DOSPLUS compatible interface", 0Dh, 0Ah
                BYTE        09h, "-md", 09h, 09h, "Force a MULTIDOS compatible interface", 0Dh, 0Ah
                BYTE        09h, "-mo", 09h, 09h, "Force a MICRODOS compatible interface **", 0Dh, 0Ah
                BYTE        09h, "-nd", 09h, 09h, "Force a NEWDOS/80 compatible interface", 0Dh, 0Ah
                BYTE        09h, "-os", 09h, 09h, "Force a OS-80 compatible interface *", 0Dh, 0Ah
                BYTE        09h, "-rd", 09h, 09h, "Force a RAPIDOS compatible interface", 0Dh, 0Ah
                BYTE        09h, "-t4", 09h, 09h, "Force a TRSDOS compatible interface (Model 4)", 0Dh, 0Ah
                BYTE        09h, "-t3", 09h, 09h, "Force a TRSDOS compatible interface (Model III)", 0Dh, 0Ah
                BYTE        09h, "-t1", 09h, 09h, "Force a TRSDOS compatible interface (Model I, SD)", 0Dh, 0Ah
                BYTE        09h, "-tm", 09h, 09h, "Force a TRSDOS compatible interface (Model I, MD)", 0Dh, 0Ah
                BYTE        0Dh, 0Ah
                BYTE        "*  Not available yet.", 0Dh, 0Ah
                BYTE        "** Experimental, read-only.", 0Dh, 0Ah
                BYTE        0

;---------------------------------------------------------------------
; Entry-Point
;---------------------------------------------------------------------

.code

EntryPoint:

    ; --- Parse command-line arguments ---

            fn      GetCommandLine                                      ; Get pointer to the command line

            fn      ToLower, eax                                        ; Convert command line to lower case

            fn      MakeArgList, offset argv, offset argc, eax          ; Build arguments list

            fn      ParseCmdLine                                        ; Call the handler routine of each identified parameter

            or      eax, eax                                            ; Exit if incorrect parameters were indicated
            jnz     Exit

    ; --- Print credits --- 

            fn      StdErr, offset Msg_Rights                           ; Print copyright information

    ; --- Open image file --- 

            cmp     g_Config.File[1 * sizeof DWORD], NULL               ; Check whether a image file name has been passed
            jne     @F

            fn      StdErr, offset Msg_Usage                            ; Print usage instructions
            jmp     Exit

@@:         fn      CreateFile, g_Config.File[1 * sizeof DWORD], GENERIC_READ+GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL    ; Open file

            cmp     eax, INVALID_HANDLE_VALUE                           ; Check whether an open error occurred
            jne     @F

            fn      GetLastError                                        ; Print error message
            fn      PrintError, eax
            jmp     Exit

@@:         mov     g_hFile, eax                                        ; Save file handle

    ; --- Create global VDI object --- 

            fn      GlobalAlloc, GPTR, 256                              ; Allocate 256 bytes for VDI object

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     CloseFile
                     
@@:         mov     g_pVDI, eax                                         ; Save pointer

    ; --- Create global OSI object --- 

            fn      GlobalAlloc, GPTR, 256                              ; Allocate 256 bytes for OSI object

            or      eax, eax
            jnz     @F
            
            fn      GetLastError
            fn      PrintError, eax
            jmp     DestroyVDI
                     
@@:         mov     g_pOSI, eax                                         ; Save pointer

    ; --- Execute the requested operation --- 

@@:         cmp     g_Config.Command, NULL                              ; Check whether a command has been requested
            jne     @F

            mov     g_Config.Command, List                              ; Assume "-l" as default

@@:         call    g_Config.Command                                    ; Execute the requested command

            fn      PrintError, eax                                     ; Print error message

    ; --- Release allocated resources --- 

DestroyOSI: fn      GlobalFree, g_pOSI                                  ; Destroy global OSI object

DestroyVDI: fn      GlobalFree, g_pVDI                                  ; Destroy global VDI object

CloseFile:  fn      CloseHandle, g_hFile                                ; Close file

Exit:       fn      ExitProcess, 0                                      ; Exit

;---------------------------------------------------------------------
; List                                                 List disk files
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

List        proc

            local   hFPDE:DWORD
            local   FindFlag:BYTE
            local   Filespec[11]:BYTE
            local   FileName[9]:BYTE
            local   FileExt[4]:BYTE
            local   FileSize:DWORD
            local   FileDay:DWORD
            local   FileMonth:DWORD
            local   FileYear:DWORD
            local   FileAttr:DWORD
            local   String[64]:BYTE

    ; --- Load interfaces ---

            fn      LoadVDI, g_pVDI, g_hFile, g_Config.VDI              ; Load Virtual Disk Interface

            or      eax, eax
            jnz     Done

            fn      LoadOSI, g_pOSI, g_pVDI, g_Config.OSI               ; Load Operating System Interface

            or      eax, eax
            jnz     Unload_1

    ; --- Format file mask --- 

            cmp     g_Config.File[2 * sizeof DWORD], NULL               ; Check whether a file mask has been passed
            jne     @F

            mov     g_Config.File[2 * sizeof DWORD], offset g_Mask      ; Point to the default mask

@@:         fn      Win2Trs, g_Config.File[2 * sizeof DWORD], addr Filespec

    ; --- Get next directory entry ---

            mov     FindFlag, OSI_FINDFIRST                             ; Set search flag to "first file"

Dir:        mcall   OSI_VTBL, Find, g_pOSI, FindFlag                    ; Seek the next file

            or      eax, eax                                            ; Check whether there are no more files
            jz      Unload_2

            cmp     eax, -1                                             ; Check whether an error occurred
            jne     @F

            fn      PrintError, ERROR_SEEK                              ; Print error message
            jmp     Error

@@:         mov     hFPDE, eax

    ; --- Retrieve file information ---

            mcall   OSI_VTBL, GetFileName, g_pOSI, hFPDE, addr FileName, sizeof FileName    ; Get file name
            mov     FileName[8], 0

            mcall   OSI_VTBL, GetFileExt, g_pOSI, hFPDE, addr FileExt, sizeof FileExt       ; Get file extension
            mov     FileExt[3], 0

            mcall   OSI_VTBL, GetFileSize, g_pOSI, hFPDE                                    ; Get file size
            mov     FileSize, eax

            mcall   OSI_VTBL, GetFileDate, g_pOSI, hFPDE                                    ; Get file date

            movzx   edx, al                                                                 ;   Split the day
            mov     FileDay, edx

            movzx   edx, ah                                                                 ;   Split the month
            mov     FileMonth, edx

            shr     eax, 16                                                                 ;   Split the year
            mov     FileYear, eax

            mcall   OSI_VTBL, GetFileAttr, g_pOSI, hFPDE                                    ; Get file attributes
            mov     FileAttr, eax

    ; --- Check file attribute filters ---
    
            test    FileAttr, 000000F0h                                 ; Check whether the "System" attribute is set
            jz      @F

            cmp     g_Config.System, 0                                  ; Check whether the "System" allowed is allowed
            je      Next

@@:         test    FileAttr, 00000F00h                                 ; Check whether the "Invisible" attribute is set
            jz      @F

            cmp     g_Config.Invisible, 0                               ; Check whether the "Invisible" attribute is allowed
            je      Next

    ; --- Check file mask ---

@@:         fn      CompFspec, addr FileName, addr Filespec, 8

            or      eax, eax
            jz      Next

            fn      CompFspec, addr FileExt, addr Filespec + 8, 3

            or      eax, eax
            jz      Next

    ; --- Print file information ---

            fn      wsprintf, addr String, "%s%c%s%c%6i%c%04i-%02i-%02i%c%04X%c%c", addr FileName, 09h, addr FileExt, 09h, FileSize, 09h, FileYear, FileMonth, FileDay, 09h, FileAttr, 0Dh, 0Ah

            fn      StdOut, addr String                                 ; Print file data

Next:       mov     FindFlag, OSI_FINDNEXT                              ; Set search flag to "next file"
            jmp     Dir

    ; --- Unload interfaces ---

Error:      mov     eax, ERROR_NO_MORE_FILES                            ; Set error code to "No more files"

Unload_2:   push    eax
            fn      UnloadOSI, g_pOSI                                   ; Unload Operating System Interface
            pop     eax

Unload_1:   push    eax
            fn      UnloadVDI, g_pVDI                                   ; Unload Virtual Disk Interface
            pop     eax

Done:       ret

List        endp

;---------------------------------------------------------------------
; Read                                         Extract files from disk
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

Read        proc

            local   pBuffer:DWORD
            local   hFPDE:DWORD
            local   FindFlag:BYTE
            local   Filespec[11]:BYTE
            local   FileName[9]:BYTE
            local   FileExt[4]:BYTE
            local   FileSize:DWORD
            local   FileDate:DWORD
            local   FileAttr:DWORD
            local   TRSName[12]:BYTE
            local   WinName[13]:BYTE
            local   String[80]:BYTE
            local   hFile:HANDLE
            local   Bytes:DWORD

    ; --- Load interfaces ---

            fn      LoadVDI, g_pVDI, g_hFile, g_Config.VDI              ; Load Virtual Disk Interface

            or      eax, eax
            jnz     Done

            fn      LoadOSI, g_pOSI, g_pVDI, g_Config.OSI               ; Load Operating System Interface

            or      eax, eax
            jnz     Unload_1

    ; --- Allocate memory for file contents ---

            fn      GlobalAlloc, GPTR, 80*30*2*256                      ; 1.2MB should be enough for most TRS-80 files

            or      eax, eax
            jz      Unload_2

            mov     pBuffer, eax                                        ; Save memory pointer

    ; --- Format file mask --- 

            cmp     g_Config.File[2 * sizeof DWORD], NULL               ; Check whether a file mask has been passed
            jne     @F

            mov     g_Config.File[2 * sizeof DWORD], offset g_Mask      ; Point to the default mask

@@:         fn      Win2Trs, g_Config.File[2 * sizeof DWORD], addr Filespec

    ; --- Get next directory entry ---

            mov     FindFlag, OSI_FINDFIRST                             ; Set search flag as "first file"

Dir:        mcall   OSI_VTBL, Find, g_pOSI, FindFlag                    ; Seek the next file

            or      eax, eax                                            ; Check whether there are no more files
            jz      FreeBuffer

            cmp     eax, -1                                             ; Check whether an error occurred
            jne     @F

            fn      PrintError, ERROR_SEEK                              ; Print error message
            jmp     Error

@@:         mov     hFPDE, eax                                          ; Save file handle

    ; --- Retrieve file information ---

            mcall   OSI_VTBL, GetFileName, g_pOSI, hFPDE, addr FileName, sizeof FileName    ; Get file name
            mov     FileName[8], 0

            mcall   OSI_VTBL, GetFileExt, g_pOSI, hFPDE, addr FileExt, sizeof FileExt       ; Get file extension
            mov     FileExt[3], 0

            mcall   OSI_VTBL, GetFileSize, g_pOSI, hFPDE                                    ; Get file size
            mov     FileSize, eax

            or      eax, eax                                                                ; Check whether the file is empty
            jz      Next

            mcall   OSI_VTBL, GetFileDate, g_pOSI, hFPDE                                    ; Get file date
            mov     FileDate, eax

            mcall   OSI_VTBL, GetFileAttr, g_pOSI, hFPDE                                    ; Get file attributes
            mov     FileAttr, eax

    ; --- Check file attribute filters ---
    
            test    FileAttr, 000000F0h                                 ; Check whether the "System" attribute is set
            jz      @F

            cmp     g_Config.System, 0                                  ; Check whether the "System" allowed is allowed
            je      Next

@@:         test    FileAttr, 00000F00h                                 ; Check whether the "Invisible" attribute is set
            jz      @F

            cmp     g_Config.Invisible, 0                               ; Check whether the "Invisible" attribute is allowed
            je      Next

    ; --- Check file mask ---

@@:         fn      CompFspec, addr FileName, addr Filespec, 8

            or      eax, eax
            jz      Next

            fn      CompFspec, addr FileExt, addr Filespec + 8, 3

            or      eax, eax
            jz      Next

    ; --- Convert file name ---

            fn      TrsCat, addr FileName, addr FileExt, addr TRSName   ; Join name and extension (blank-padded)

            fn      Trs2Win, addr TRSName, addr WinName                 ; Convert file name to windows standard

            fn      wsprintf, addr String, "Exporting ""%s""... ", addr WinName

            fn      StdOut, addr String                                 ; Print file name

    ; --- Set file pointer ---

            mcall   OSI_VTBL, Seek, g_pOSI, hFPDE, 0                    ; Set file pointer at beginning of file

            or      eax, eax                                            ; Check for error
            jz      @F

            fn      PrintError, eax                                     ; Print error message
            jmp     Next

    ; --- Read entire file into buffer ---

@@:         mcall   OSI_VTBL, Read, g_pOSI, hFPDE, pBuffer, FileSize    ; Read file

            or      eax, eax                                            ; Check for error
            jz      @F

            fn      PrintError, eax                                     ; Print error message
            jmp     Next

    ; --- Write buffer contents to disk ---

@@:         fn      CreateFile, addr WinName, GENERIC_READ+GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL   ; Create a new windows file

            mov     hFile, eax                                                                              ; Save handle

            fn      WriteFile, hFile, pBuffer, FileSize, addr Bytes, NULL                                   ; Write memory contents to file

            fn      SetWinFileDate, hFile, FileDate                                                         ; Try to set the file date

            cmp     g_Config.Attrib, 0                                                                      ; Check if requested to preserve file attributes
            je      @F

            fn      SetWinFileAttr, addr WinName, FileAttr                                                  ; Try to set the file attributes
            
@@:         fn      CloseHandle, hFile                                                                      ; Close file

    ; --- Move to next file ---

            fnc     wsprintf, addr String, "%i of %i bytes exported.\n", Bytes, FileSize

            fn      StdOut, addr String                                 ; Print written bytes

Next:       mov     FindFlag, OSI_FINDNEXT                              ; Set search flag as "next file"
            jmp     Dir

    ; --- Release allocated resources ---
            
Error:      mov     eax, ERROR_NO_MORE_FILES                            ; Set error code to "No more files"

FreeBuffer: push    eax
            fn      GlobalFree, pBuffer                                 ; Release allocated memory
            pop     eax

Unload_2:   push    eax
            fn      UnloadOSI, g_pOSI                                   ; Unload Operating System Interface
            pop     eax

Unload_1:   push    eax
            fn      UnloadVDI, g_pVDI                                   ; Unload Virtual Disk Interface
            pop     eax

Done:       ret

Read        endp

;---------------------------------------------------------------------
; Write                                            Write files to disk
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

Write       proc

            local   hSource:HANDLE
            local   hTarget:DWORD
            local   hFind:HANDLE
            local   File:WIN32_FIND_DATA
            local   dwWritten:DWORD
            local   dwBytes:DWORD
            local   abBuffer[256]:BYTE
            local   abWinName[13]:BYTE
            local   abTrsName[13]:BYTE
            local   abString[80]:BYTE

    ; --- Check whether a file mask has been passed --- 

            cmp     g_Config.File[2 * sizeof DWORD], NULL
            jne     @F

            mov     eax, ERROR_BAD_ARGUMENTS                            ; One or more arguments are not correct.
            jmp     Done

    ; --- Load interfaces ---

@@:         fn      LoadVDI, g_pVDI, g_hFile, g_Config.VDI              ; Load Virtual Disk Interface

            or      eax, eax
            jnz     Done

            fn      LoadOSI, g_pOSI, g_pVDI, g_Config.OSI               ; Load Operating System Interface

            or      eax, eax
            jnz     Unload_1

    ; --- Find the first file to write ---

            fn      FindFirstFile, g_Config.File[2 * sizeof DWORD], addr File

            cmp     eax, INVALID_HANDLE_VALUE                           ; Check for error
            jne     @F

            fn      GetLastError                                        ; Get error code
            jmp     Unload_2

@@:         mov     hFind, eax                                          ; Save search handle

    ; --- Convert file name ---

Next:       fn      Win2Trs, addr File.cFileName, addr abTrsName

            fn      Trs2Win, addr abTrsName, addr abWinName
            
            fn      wsprintf, addr abString, "Writing ""%s""... ", addr abWinName

            fn      StdOut, addr abString

    ; --- Check source file size ---

            cmp     File.nFileSizeHigh, 0
            je      @F

            fn      PrintError, ERROR_DISK_FULL                         ; There is not enough space on the disk.
            jmp     Error

    ; --- Open source file ---

@@:         fn      CreateFile, addr File.cFileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL

            cmp     eax, INVALID_HANDLE_VALUE                           ; Check for error
            jne     @F

            fn      GetLastError                                        ; Get error code
            jmp     Error
                     
@@:         mov     hSource, eax

    ; --- Create target file ---

            mov     dwWritten, 0                                        ; Reset counter
            
            mcall   OSI_VTBL, Create, g_pOSI, addr abTrsName, File.nFileSizeLow

            or      eax, eax                                            ; Check for error
            jnz     @F

            mcall   OSI_VTBL, GetLastError, g_pOSI                      ; Get error code
            jmp     Error
            
@@:         mov     hTarget, eax                                        ; Save target file handle

            fn      SetTrsFileDate, hTarget, addr File.ftCreationTime   ; Try to set the file date

            cmp     g_Config.Attrib, 0                                  ; Check if requested to preserve file attributes
            je      @F

            fn      SetTrsFileAttr, hTarget, File.dwFileAttributes      ; Try to set the file attributes
            
@@:         mcall   OSI_VTBL, Seek, g_pOSI, hTarget, 0                  ; Set file pointer

            or      eax, eax
            jnz     Error

    ; --- Copy source to target ---

Copy:       fn      ReadFile, hSource, addr abBuffer, sizeof abBuffer, addr dwBytes, NULL   ; Read up to 256 into buffer

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Error
            
@@:         mcall   OSI_VTBL, Write, g_pOSI, hTarget, addr abBuffer, dwBytes                ; Write buffer contents to target

            or      eax, eax
            jnz     Error

            mov     eax, dwBytes                                                            ; Compute total bytes written
            add     dwWritten, eax
            
            mov     eax, dwWritten                                                          ; Compute remaining bytes to write
            cmp     eax, File.nFileSizeLow                                                  ; Continue until bytes written equals file size
            jb      Copy

    ; --- Close file ---

            fn      CloseHandle, hSource                                ; Close source file

            fnc     wsprintf, addr abString, "%i of %i bytes written.\n", dwWritten, File.nFileSizeLow

            fn      StdOut, addr abString

    ; --- Find next file ---

            fn      FindNextFile, hFind, addr File

            or      eax, eax
            jnz     Next
            
            fn      GetLastError

            cmp     eax, ERROR_NO_MORE_FILES                            ; There are no more files.
            jne     Error
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     @F

    ; --- Print final message ---

Error:      push    eax

            fnc     wsprintf, addr abString, "%i of %i bytes written.\n", dwWritten, File.nFileSizeLow

            fn      StdOut, addr abString

            pop     eax
            
@@:         fn      PrintError, eax                                     ; Print error message

    ; --- Release resources ---

Close:      push    eax
            fn      FindClose, hFind                                    ; Close search handle
            pop     eax

Unload_2:   push    eax
            fn      UnloadOSI, g_pOSI                                   ; Unload Operating System Interface
            pop     eax

Unload_1:   push    eax
            fn      UnloadVDI, g_pVDI                                   ; Unload Virtual Disk Interface
            pop     eax

Done:       ret

Write       endp

;---------------------------------------------------------------------
; Delete                                                  Delete files
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

Delete      proc

            local   hFPDE:DWORD
            local   FindFlag:BYTE
            local   Filespec[11]:BYTE
            local   FileName[9]:BYTE
            local   FileExt[4]:BYTE
            local   FileAttr:DWORD
            local   TrsName[12]:BYTE
            local   WinName[13]:BYTE
            local   String[80]:BYTE
            local   hStdIn:HANDLE
            local   Input:INPUT_RECORD
            local   Count:DWORD

    ; --- Load interfaces ---

            fn      LoadVDI, g_pVDI, g_hFile, g_Config.VDI              ; Load Virtual Disk Interface

            or      eax, eax
            jnz     Done

            fn      LoadOSI, g_pOSI, g_pVDI, g_Config.OSI               ; Load Operating System Interface

            or      eax, eax
            jnz     Unload_1

            fn      GetStdHandle, STD_INPUT_HANDLE                      ; Get handle to StdIn
            mov     hStdIn, eax
            
    ; --- Format file mask --- 

            cmp     g_Config.File[2 * sizeof DWORD], NULL               ; Check whether a file mask has been passed
            jne     @F

            mov     g_Config.File[2 * sizeof DWORD], offset g_Mask      ; Point to the default mask

@@:         fn      Win2Trs, g_Config.File[2 * sizeof DWORD], addr Filespec

    ; --- Get next directory entry ---

            mov     FindFlag, OSI_FINDFIRST                             ; Set search flag as "first file"

Dir:        mcall   OSI_VTBL, Find, g_pOSI, FindFlag                    ; Seek the next file

            or      eax, eax                                            ; Check whether there are no more files
            jz      Unload_2

            cmp     eax, -1                                             ; Check whether an error occurred
            jne     @F

            fn      PrintError, ERROR_SEEK                              ; Print error message
            jmp     Error

@@:         mov     hFPDE, eax                                          ; Save file handle

    ; --- Retrieve file information ---

            mcall   OSI_VTBL, GetFileName, g_pOSI, hFPDE, addr FileName, sizeof FileName    ; Get file name
            mov     FileName[8], 0

            mcall   OSI_VTBL, GetFileExt, g_pOSI, hFPDE, addr FileExt, sizeof FileExt       ; Get file extension
            mov     FileExt[3], 0

            mcall   OSI_VTBL, GetFileAttr, g_pOSI, hFPDE                                    ; Get file attributes
            mov     FileAttr, eax

    ; --- Check file attribute filters ---
    
            test    FileAttr, 000000F0h                                 ; Check whether the "System" attribute is set
            jz      @F

            cmp     g_Config.System, 0                                  ; Check whether the "System" allowed is allowed
            je      Next

@@:         test    FileAttr, 00000F00h                                 ; Check whether the "Invisible" attribute is set
            jz      @F

            cmp     g_Config.Invisible, 0                               ; Check whether the "Invisible" attribute is allowed
            je      Next

    ; --- Check file mask ---

@@:         fn      CompFspec, addr FileName, addr Filespec, 8

            or      eax, eax
            jz      Next

            fn      CompFspec, addr FileExt, addr Filespec + 8, 3

            or      eax, eax
            jz      Next

    ; --- Convert file name ---

            fn      TrsCat, addr FileName, addr FileExt, addr TrsName   ; Join name and extension (blank-padded)

            fn      Trs2Win, addr TrsName, addr WinName                 ; Convert file name to windows standard

    ; --- Check if user confirmation is enabled ---

            cmp     g_Config.Yes, 0
            je      @F

            fn      wsprintf, addr String, "Deleting ""%s""... ", addr WinName

            fn      StdOut, addr String                                 ; Print message

            jmp     Del                                                 ; Jump to deletion routine

    ; --- Request user confirmation ---

@@:         fn      wsprintf, addr String, "Delete ""%s""? ", addr WinName

            fn      StdOut, addr String                                 ; Print file name

@@:         fn      ReadConsoleInput, hStdIn, addr Input, 1, addr Count ; Read one event from the console

            cmp     Input.EventType, KEY_EVENT                          ; Ignore all but key events
            jne     @B

            cmp     Input.KeyEvent.bKeyDown, TRUE                       ; Check if key is being pressed or released
            jne     @B
            
            cmp     Input.KeyEvent.wVirtualKeyCode, 'Y'                 ; Check if user pressed 'Y'
            je      Yes

            cmp     Input.KeyEvent.wVirtualKeyCode, 'N'                 ; Check if user pressed 'N'
            je      No

            cmp     Input.KeyEvent.wVirtualKeyCode, 1Bh                 ; Check if user pressed Esc
            je      Break

            cmp     Input.KeyEvent.wVirtualKeyCode, 0Dh                 ; Check if user pressed Enter
            jne     @B

No:         fnc     StdOut, "N\n"                                       ; Print 'N' and advance to next file
            jmp     Next
            
Break:      fnc     StdOut, "\n"                                        ; Exit
            jmp     Unload_2

Yes:        fn      StdOut, "Y "                                        ; Print 'Y'
            
    ; --- Delete file ---

Del:        mcall   OSI_VTBL, Delete, g_pOSI, hFPDE                     ; Delete file

            or      eax, eax                                            ; Check for error
            jz      @F

            fn      PrintError, eax                                     ; Print error message
            jmp     Next

@@:         fnc     StdOut, "Done.\n"                                   ; Print confirmation message

    ; --- Move to next file ---

Next:       mov     FindFlag, OSI_FINDNEXT                              ; Set search flag as "next file"
            jmp     Dir

    ; --- Release interfaces ---
            
Error:      mov     eax, ERROR_NO_MORE_FILES                            ; Set error code to "No more files"

Unload_2:   push    eax
            fn      UnloadOSI, g_pOSI                                   ; Unload Operating System Interface
            pop     eax

Unload_1:   push    eax
            fn      UnloadVDI, g_pVDI                                   ; Unload Virtual Disk Interface
            pop     eax

Done:       ret

Delete      endp

;---------------------------------------------------------------------
; Dump                                                  Dump disk data
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

Dump        proc

            local   Buffer[1024]:BYTE
            local   String[64]:BYTE
            local   LastTrack:BYTE
            local   FirstSector:BYTE
            local   LastSector:BYTE
            local   FirstSide:BYTE
            local   LastSide:BYTE
            local   Track:BYTE
            local   Sector:BYTE
            local   Side:BYTE

    ; --- Load interface ---

            fn      LoadVDI, g_pVDI, g_hFile, g_Config.VDI              ; Load Virtual Disk Interface

            or      eax, eax
            jnz     Done

    ; --- Retrieve the disk geometry ---

            mcall   VDI_VTBL, GetFirstTrack, g_pVDI                     ; Get first track
            mov     Track, al

            mcall   VDI_VTBL, GetLastTrack, g_pVDI                      ; Get last track
            mov     LastTrack, al

            mcall   VDI_VTBL, GetFirstSector, g_pVDI                    ; Get first sector
            mov     FirstSector, al
            mov     Sector, 0                                           ; Always start from 0 (some disks have first track formatted differently)

            mcall   VDI_VTBL, GetLastSector, g_pVDI                     ; Get last sector
            mov     LastSector, al

            mcall   VDI_VTBL, GetFirstSide, g_pVDI                      ; Get first side
            mov     FirstSide, al
            mov     Side, al

            mcall   VDI_VTBL, GetLastSide, g_pVDI                       ; Get last side
            mov     LastSide, al

    ; --- Read and dump the sector ---

@@:         mcall   VDI_VTBL, Read, g_pVDI, addr Buffer, sizeof Buffer, Track, Side, Sector     ; Read sector at first track, first side, sector 0

            or      eax, eax                                            ; On error, skip the sector
            jnz     Next

            push    ecx

            movzx   edx, Track
            movzx   ecx, Side
            movzx   eax, Sector
            
            fn      wsprintf, addr String, "%c%c[%02u:%u:%02u]%c%c", 0Dh, 0Ah, edx, ecx, eax, 0Dh, 0Ah
    
            fn      StdOut, addr String                                 ; Print the sector address (CC:H:SS)

            pop     ecx

            fn      DumpSector, addr Buffer, ecx                        ; Dump sector data

    ; --- Advance to next Sector, Side, Track ---

Next:       inc     Sector                                              ; Increment sector
            mov     al, Sector
            cmp     al, LastSector
            jbe     @B
        
            mov     al, FirstSector                                     ; Reset sector count
            mov     Sector, al

            inc     Side                                                ; Increment side
            mov     al, Side
            cmp     al, LastSide
            jbe     @B

            mov     al, FirstSide                                       ; Reset side count
            mov     Side, al

            inc     Track                                               ; Increment track
            mov     al, Track
            cmp     al, LastTrack
            jbe     @B

            fn      UnloadVDI, g_pVDI                                   ; Unload Virtual Disk Interface

            mov     eax, NO_ERROR

Done:       ret
    
Dump        endp

;---------------------------------------------------------------------
; LoadVDI                                  Load Virtual Disk Interface
;---------------------------------------------------------------------
; pVDI:         Pointer to VDI object
; hFile:        Handle to open diskette file
; Interface:    Selected VDI interface
;---------------------------------------------------------------------
; EAX:          Error code
;---------------------------------------------------------------------

LoadVDI     proc    pVDI:DWORD, hFile:HANDLE, Interface:DWORD

            fnc     StdErr, "Loading Virtual Disk Interface...\n"

    ; --- Load requested interface ---

            mov     eax, Interface

            cmp     eax, NULL
            je      Auto

            mov     edx, pVDI

            mov     (CVDI PTR [edx]).VPTR, eax

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            jmp     Done

    ; --- Try loading the DMK interface ---

Auto:       fnc     StdErr, "Trying DMK...\n"

            mov     edx, pVDI

            mov     (CVDI PTR [edx]).VPTR, offset DMK

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   VDI_VTBL, Unload, pVDI

    ; --- Try the JV3 interface ---

            fnc     StdErr, "Trying JV3...\n"

            mov     edx, pVDI

            mov     (CVDI PTR [edx]).VPTR, offset JV3

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   VDI_VTBL, Unload, pVDI

    ; --- Try the JV1 interface ---

            fnc     StdErr, "Trying JV1...\n"

            mov     edx, pVDI

            mov     (CVDI PTR [edx]).VPTR, offset JV1

            mcall   VDI_VTBL, Load, pVDI, hFile, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   VDI_VTBL, Unload, pVDI

    ; --- Report an error ---

            mov     eax, ERROR_OPEN_FAILED

Done:       ret

LoadVDI     endp

;---------------------------------------------------------------------
; LoadOSI                              Load Operating System Interface
;---------------------------------------------------------------------
; pOSI:         Pointer to OSI object
; pVDI:         Pointer to an initialized VDI object
; Interface:    Selected OSI
;---------------------------------------------------------------------
; EAX:          Error code
;---------------------------------------------------------------------

LoadOSI     proc    pOSI:DWORD, pVDI:DWORD, Interface:DWORD

            fnc     StdErr, "Loading Operating System Interface...\n"

            mov     eax, Interface

    ; --- Load the requested interface ---

            cmp     eax, NULL
            je      Auto

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, eax

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            jmp     Done

    ; --- Try the NEWDOS/80 interface ---

Auto:       fnc     StdErr, "Trying NEWDOS/80...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset NEWDOS80

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the MICRODOS interface ---

            fnc     StdErr, "Trying MICRODOS...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset MICRODOS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the DOSPLUS interface ---

            fnc     StdErr, "Trying DOSPLUS...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset DOSPLUS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the TRSDOS interface ---

            fnc     StdErr, "Trying TRSDOS Model 4...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset TRSDOS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the TRSDOS3 interface ---

            fnc     StdErr, "Trying TRSDOS Model III...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset TRSDOS3

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the TRSDOS1 (SD) interface ---

            fnc     StdErr, "Trying TRSDOS Model I SD...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset TRSDOS1

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the TRSDOS1 (MD) interface ---

            fnc     StdErr, "Trying TRSDOS Model I MD...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset TRSDOS1M

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the RAPIDOS interface ---

            fnc     StdErr, "Trying RAPIDOS...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset RAPIDOS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the MULTIDOS interface ---

            fnc     StdErr, "Trying MULTIDOS...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset MULTIDOS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Try the DBLDOS interface ---

            fnc     StdErr, "Trying DBLDOS...\n"

            mov     edx, pOSI

            mov     (COSI PTR [edx]).VPTR, offset DBLDOS

            mcall   OSI_VTBL, Load, pOSI, pVDI, 0

            fn      PrintError, eax

            or      eax, eax
            jz      Done

            mcall   OSI_VTBL, Unload, pOSI

    ; --- Must report an error ---

            mov     eax, ERROR_NOT_DOS_DISK

Done:       ret

LoadOSI     endp

;---------------------------------------------------------------------
; UnloadVDI                              Unload Virtual Disk Interface
;---------------------------------------------------------------------
; pVDI:     Pointer to VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

UnloadVDI   proc    pVDI:DWORD

            mcall   VDI_VTBL, Unload, pVDI

            fnc     StdErr, "Unloading Virtual Disk Interface...\n"

            ret

UnloadVDI   endp

;---------------------------------------------------------------------
; UnloadOSI                          Unload Operating System Interface
;---------------------------------------------------------------------
; pOSI:     Pointer to OSI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

UnloadOSI   proc    pOSI:DWORD

            mcall   OSI_VTBL, Unload, pOSI

            fnc     StdErr, "Unloading Operating System Interface...\n"

            ret

UnloadOSI   endp

;---------------------------------------------------------------------
; SetTrsFileDate                            Set the date of a TRS file
;---------------------------------------------------------------------
; hFile:    TRS file handle (FPDE)
; pDate:    Pointer to a FILETIME structure
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetTrsFileDate  proc    hFile:DWORD, pDate:DWORD

            local   Time:SYSTEMTIME
            
            fn      FileTimeToSystemTime, pDate, addr Time              ; Convert the date in Day/Month/Year

            or      eax, eax                                            ; Check for error
            jnz     @F
            
            fn      GetLastError                                        ; Get error code
            jmp     Done

@@:         movzx   eax, Time.wYear                                     ; EAX = YYYYMMDD
            shl     eax, 16
            mov     ax, Time.wMonth
            shl     ax, 8
            add     ax, Time.wDay

            mcall   OSI_VTBL, SetFileDate, g_pOSI, hFile, eax           ; Set file date

Done:       ret
            
SetTrsFileDate  endp

;---------------------------------------------------------------------
; SetWinFileDate                        Set the date of a Windows file
;---------------------------------------------------------------------
; hFile:    Windows file handle
; dwDate:   File date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetWinFileDate  proc    hFile:HANDLE, dwDate:DWORD

            local   SystemTime:SYSTEMTIME
            local   FileTime:FILETIME

            mov     eax, dwDate

            movzx   dx, al                                                  ; Fill SystemTime structure
            mov     SystemTime.wDay, dx

            movzx   dx, ah
            mov     SystemTime.wMonth, dx
            
            shr     eax, 16
            mov     SystemTime.wYear, ax
            
            mov     SystemTime.wHour, 0
            mov     SystemTime.wMinute, 0
            mov     SystemTime.wSecond, 0
            mov     SystemTime.wMilliseconds, 0
            
            fn      SystemTimeToFileTime, addr SystemTime, addr FileTime    ; Convert SystemTime in FileTime

            or      eax, eax                                                ; Check for error
            jnz     @F

            fn      GetLastError                                            ; Get error code
            jmp     Done
            
@@:         fn      SetFileTime, hFile, addr FileTime, addr FileTime, addr FileTime

            or      eax, eax                                                ; Check for error
            jnz     @F

            fn      GetLastError                                            ; Get error code
            jmp     Done
            
@@:         mov     eax, NO_ERROR                                       ; The operation completed successfully.
            
Done:       ret

SetWinFileDate  endp

;---------------------------------------------------------------------
; SetTrsFileAttr                      Set the attributes of a TRS file
;---------------------------------------------------------------------
; hFile:    TRS file handle (FPDE)
; dwAttr:   File attributes (Windows standard)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetTrsFileAttr  proc    hFile:DWORD, dwAttr:DWORD

            xor     eax, eax                                            ; Prepare SetFileAttr() parameter
            
            test    dwAttr, FILE_ATTRIBUTE_ARCHIVE                      ; Test the Archive attribute
            jz      @F
            
            or      eax, 0000F000h                                      ; Set the Modified attribute

@@:         test    dwAttr, FILE_ATTRIBUTE_HIDDEN                       ; Test the Hidden attribute
            jz      @F

            or      eax, 00000F00h                                      ; Set the Invisible attribute

@@:         test    dwAttr, FILE_ATTRIBUTE_SYSTEM                       ; Test the System attribute
            jz      @F

            or      eax, 000000F0h                                      ; Set the System attribute

@@:         test    dwAttr, FILE_ATTRIBUTE_READONLY                     ; Test the ReadOnly attribute
            jz      @F

            or      eax, 00000005h                                      ; Set access protection to Read

@@:         mcall   OSI_VTBL, SetFileAttr, g_pOSI, hFile, eax           ; Set file attributes

Done:       ret

SetTrsFileAttr  endp

;---------------------------------------------------------------------
; SetWinFileAttr                  Set the attributes of a Windows file
;---------------------------------------------------------------------
; pFile:    Pointer to Windows filename
; dwAttr:   File attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetWinFileAttr  proc    pFile:DWORD, dwAttr:DWORD

            xor     eax, eax                                            ; Prepare SetFileAttributes() parameter

            test    dwAttr, 0000F000h                                   ; Test the Modified attribute
            jz      @F

            add     eax, FILE_ATTRIBUTE_ARCHIVE                         ; Set the Archive attribute
            
@@:         test    dwAttr, 00000F00h                                   ; Test the Invisible attribute
            jz      @F

            add     eax, FILE_ATTRIBUTE_HIDDEN                          ; Set the Hidden attribute

@@:         test    dwAttr, 000000F0h                                   ; Test the System attribute
            jz      @F

            add     eax, FILE_ATTRIBUTE_SYSTEM                          ; Set the System attribute

@@:         test    dwAttr, 0000000Fh                                   ; Test for any access restrictions
            jz      @F

            add     eax, FILE_ATTRIBUTE_READONLY                        ; Set the ReadOnly attribute

@@:         fn      SetFileAttributes, pFile, eax                       ; Set file attributes

            or      eax, eax                                            ; Check for error
            jnz     @F

            fn      GetLastError                                        ; Get error code
            jmp     Done
            
@@:         mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

SetWinFileAttr  endp

;---------------------------------------------------------------------
; DumpSector                           Dump data in both hex and ascii
;---------------------------------------------------------------------
; pData:    Pointer to the data buffer
; dwLen:    Length of the data buffer
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

DumpSector  proc    pData:DWORD, dwLen:DWORD

            local   Line[65]:BYTE
            
            mov     Line[sizeof Line - 1], 0

            push    ebx
            push    esi
            push    edi
            
            mov     ebx, _T("0123456789ABCDEF")
NewLine:    lea     esi, Line
            lea     edi, Line+48

            mov     ecx, sizeof Line - 1
@@:         mov     byte ptr [esi+ecx-1], ' '
            loop    @B

Get:        mov     edx, pData

            mov     al, [edx]
            shr     al, 4
            xlatb
            mov     [esi], al

            mov     al, [edx]
            and     al, 0Fh
            xlatb
            mov     [esi+1], al

            mov     al, [edx]
            cmp     al, ' '
            jae     @F
            mov     al, '.'
@@:         mov     [edi], al

            add     esi, 3
            inc     edi
            inc     pData
            dec     dwLen

            lea     eax, Line+64
            cmp     eax, edi
            je      @F
            
            cmp     dwLen, 0
            jne     Get

@@:         fn      StdOut, addr Line
            fnc     StdOut, "\n"

            cmp     dwLen, 0
            jne     NewLine
            
            pop     edi
            pop     esi
            pop     ebx

Done:       ret

DumpSector  endp

;---------------------------------------------------------------------
; CompFspec                                          Compare filespecs
;---------------------------------------------------------------------
; pSource:  Pointer to first filespec
; pTarget:  Pointer to second filespec
; Len:      Length
;---------------------------------------------------------------------
; EAX:      FALSE:Don't Match, TRUE:Match
;---------------------------------------------------------------------

CompFspec   proc    pSource:DWORD, pTarget:DWORD, Len:DWORD

            push    esi
            push    edi

            mov     esi, pSource
            mov     edi, pTarget
            mov     ecx, Len

Compare:    mov     ah, [esi]                                           ; Move one char from each fspec to a register
            mov     al, [edi]

            cmp     ah, '?'                                             ; "?" means that any char is to be allowed in this position
            je      Next
    
            cmp     al, '?'
            je      Next

            cmp     ah, '*'                                             ; "*" means that any char is to be allowed from here to the end
            je      Match

            cmp     al, '*'
            je      Match

            or      ah, 'a' - 'A'                                       ; Make both chars the same case
            or      al, 'a' - 'A'

            cmp     ah, al                                              ; Compare them
            jne     Diff

Next:       inc     esi                                                 ; Move to the next char
            inc     edi
            loop    Compare

Match:      mov     eax, TRUE                                           ; Return TRUE
            jmp     Done

Diff:       mov     eax, FALSE                                          ; Return FALSE

Done:       pop     edi
            pop     esi

            ret

CompFspec   endp

;---------------------------------------------------------------------
; TrsCat               Concatenate Name, Extension and a trailing null 
;---------------------------------------------------------------------
; pFileName:    Pointer to 8-byte file name
; pFileExt:     Pointer to 3-byte file extension
; pBuffer:      Pointer to target buffer
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

TrsCat      proc    pFileName:DWORD, pFileExt:DWORD, pBuffer:DWORD

            push    esi
            push    edi

            mov     esi, pFileName                                      ; Copy file name
            mov     edi, pBuffer
            mov     ecx, 8
            cld
            rep     movsb
            
            mov     esi, pFileExt                                       ; Append file extension
            mov     ecx, 3
            rep     movsb

            mov     byte ptr [edi], 0                                   ; Append trailing null

            pop     edi
            pop     esi

            ret

TrsCat      endp

;---------------------------------------------------------------------
; Trs2Win                     Convert TRS filename to Windows filename
;---------------------------------------------------------------------
; pTrsName: Pointer to TRS-formatted file name and extension
; pWinName: Pointer to buffer for the windows-formatted filename
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

Trs2Win     proc    pTrsName:DWORD, pWinName:DWORD

            push    esi
            push    edi

            mov     esi, pTrsName                                       ; Set pointer at the file name
            mov     edi, pWinName
            mov     ecx, 8

CopyName:   cmp     byte ptr [esi], ' '                                 ; Copy the file name
            je      CheckExt
            movsb
            loop    CopyName

CheckExt:   mov     esi, pTrsName                                       ; Set pointer at the file extension
            add     esi, 8
            mov     ecx, 3

            cmp     byte ptr [esi], ' '                                 ; Check whether an extension exists
            je      AddNull

            mov     al, '.'                                             ; Add a dot
            stosb

CopyExt:    movsb                                                       ; Copy the file extension
            cmp     byte ptr [esi], ' '
            je      AddNull
            loop    CopyExt
            
AddNull:    mov     al, 0                                               ; Add a trailing null
            stosb

            pop     edi
            pop     esi

            ret

Trs2Win     endp

;---------------------------------------------------------------------
; Win2Trs                     Convert Windows filename to TRS filename
;---------------------------------------------------------------------
; pWinName: Pointer to Windows-formatted filename
; pTrsName: Pointer to buffer for TRS-formatted filename and extension
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

Win2Trs     proc    pWinName:DWORD, pTrsName:DWORD

            local   abName[MAX_PATH]:BYTE

            push    esi
            push    edi
            
            fn      GetShortPathName, pWinName, addr abName, sizeof abName

            or      eax, eax                                            ; Check for error
            jnz     BkSlash

            mov     esi, pWinName                                       ; Try using the original name instead
            jmp     CopyName
            
BkSlash:    mov     ecx, eax                                            ; Find the last backslash in the string (if any)
            lea     edx, abName
@@:         cmp     byte ptr [edx+ecx-1], '\'
            loopne  @B
            jcxz    @F
            
            add     edx, ecx                                            ; Point to the filename right after the last backslash
            inc     edx
            
@@:         mov     esi, edx                                            ; Copy the file name, up to 8 chars
CopyName:   mov     edi, pTrsName
            mov     ecx, 9
            cld

@@:         lodsb

            or      al, al                                              ; Reached the end of the string?
            jz      Fill_1
            
            cmp     al, '.'                                             ; Reached the beginning of the extension part?
            je      Fill_1
            
            stosb
            loop    @B
            
Fill_1:     mov     al, ' '                                             ; Fill the remaining name space with blanks
            dec     ecx
            rep     stosb

CopyExt:    mov     ecx, 3                                              ; Copy the file extension, up to 3 chars
            
            cmp     byte ptr [esi-1], 0                                 ; Check whether an extension exists at all
            je      Fill_2
            
@@:         lodsb

            or      al, al                                              ; Reached the end of the string?
            jz      Fill_2
            
            stosb
            loop    @B

Fill_2:     mov     al, ' '                                             ; Fill the remaining extension space with blanks
            rep     stosb

Done:       pop     edi
            pop     esi

            ret

Win2Trs     endp

;---------------------------------------------------------------------
; SetCmd                                             Parameter handler
;---------------------------------------------------------------------
; pCmd:     Pointer to the command routine
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetCmd      proc    pCmd:DWORD

            cmp     g_Config.Command, NULL                              ; Check whether this is the first attempt to define a command
            je      @F

            fnc     StdErr, "Attempt to set multiple commands.\n"       ; Print error message

            mov     eax, ERROR_INVALID_PARAMETER
            jmp     Done

@@:         mov     eax, pCmd                                           ; Set pointer to the command routine
            mov     g_Config.Command, eax

            mov     eax, NO_ERROR

Done:       ret

SetCmd      endp

;---------------------------------------------------------------------
; SetFlag                                            Parameter handler
;---------------------------------------------------------------------
; pVar:     Pointer to the flag being set
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetFlag     proc    pVar:DWORD

            mov     eax, pVar                                           ; Get target variable address
            inc     byte ptr [eax]                                      ; Increment target variable
            mov     eax, NO_ERROR
            ret

SetFlag     endp

;---------------------------------------------------------------------
; SetFF                                              Parameter handler
;---------------------------------------------------------------------
; pVTBL:    Pointer to the interface VTBL
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetFF       proc    pVTBL:DWORD

            cmp     g_Config.VDI, NULL                                  ; Check whether this is the first attempt to define a format
            je      @F

            fnc     StdErr, "Attempt to set multiple file formats.\n"   ; Print error message

            mov     eax, ERROR_INVALID_PARAMETER
            jmp     Done

@@:         mov     eax, pVTBL                                          ; Save the requested Virtual Disk Interface VTBL
            mov     g_Config.VDI, eax

            mov     eax, NO_ERROR

Done:       ret

SetFF       endp

;---------------------------------------------------------------------
; SetOS                                              Parameter handler
;---------------------------------------------------------------------
; pVTBL:    Pointer to the interface VTBL
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

SetOS       proc    pVTBL:DWORD

            cmp     g_Config.OSI, NULL                                  ; Check whether this is the first attempt to define an OS
            je      @F

            fnc     StdErr, "Attempt to set multiple operating systems.\n"

            mov     eax, ERROR_INVALID_PARAMETER
            jmp     Done

@@:         mov     eax, pVTBL                                          ; Save the requested Operating System Interface VTBL
            mov     g_Config.OSI, eax

            mov     eax, NO_ERROR

Done:       ret

SetOS       endp

;---------------------------------------------------------------------
; ParseCmdLine           Identify arguments and call related functions
;---------------------------------------------------------------------
; No input
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ParseCmdLine    proc

            local   Count:BYTE

            mov     Count, 0
            
            movzx   ecx, argc                                           ; Get argument count and pointer to pointers list
            mov     edx, offset argv

Scan:       push    ecx
            push    edx

            mov     edx, [edx]
            
            cmp     byte ptr [edx], '-'                                 ; Check whether we're handling a switch
            jne     Fspec

            inc     edx                                                 ; Skip the "-"
            
            fn      GetFnPtr, edx, offset Arg_Array, Arg_Count          ; Get the associated function call address

            or      eax, eax                                            ; Check whether an invalid parameter has been indicated
            jnz     @F
            
            fnc     StdErr, "Unknown switch.\n"                         ; Print error message
            
            mov     eax, ERROR_INVALID_PARAMETER
            jmp     Abort

@@:         push    edx
            call    eax                                                 ; Call the parameter handler

            or      eax, eax                                            ; Test for an error condition
            jnz     Abort
            jmp     Next

Fspec:      cmp     Count, 3
            jb      @F

            fnc     StdErr, "Too many filespecs.\n"                     ; Print error message
            
            mov     eax, ERROR_INVALID_PARAMETER
            jmp     Abort

@@:         mov     eax, sizeof DWORD                                   ; Store edx at g_Config.File[Count]
            mul     Count
            add     eax, offset g_Config.File
            mov     [eax], edx
            
            inc     Count                                               ; Increment Count

Next:       pop     edx
            pop     ecx

            add     edx, sizeof DWORD                                   ; Advance to the next argument
            loop    Scan

            mov     eax, NO_ERROR
            jmp     Done

Abort:      add     esp, sizeof DWORD * 2                               ; Balance the stack

Done:       ret

ParseCmdLine    endp

;---------------------------------------------------------------------
; GetFnPtr     Return the function pointer associated with a parameter
;---------------------------------------------------------------------
; pName:    Name to be searched
; pArray:   Pointer to ARG array to look at
; Count:    Number of items in the array
;---------------------------------------------------------------------
; EAX:      Pointer to function or NULL if not found
; EDX:      Argument value to pass to the function call
;---------------------------------------------------------------------

GetFnPtr    proc    pName:DWORD, pArray:DWORD, Count:BYTE

            local   FnPtr:DWORD
            local   FnArg:DWORD

            mov     FnPtr, NULL
            mov     FnArg, 0

            mov     eax, pName                                          ; Name to search for
            mov     edx, pArray                                         ; Array to look at
            movzx   ecx, Count                                          ; Number of array items

Scan:       push    eax
            push    ebx
            push    ecx
            push    edx

            mov     ebx, (ARG PTR [edx]).ArgName                    

@@:         mov     cl, [eax]                                           ; Compare ParamName with pName
            cmp     cl, [ebx]
            jne     @F

            inc     eax
            inc     ebx
            
            or      cl, cl
            jnz     @B

            mov     eax, (ARG PTR [edx]).ArgCall
            mov     FnPtr, eax                                          ; Save function call address

            mov     eax, (ARG PTR [edx]).ArgPass
            mov     FnArg, eax                                          ; Save function argument value

            mov     dword ptr [esp + sizeof DWORD * 1], 1               ; Exit the loop (mov stack_ECX, 1)

@@:         pop     edx
            pop     ecx
            pop     ebx
            pop     eax

            add     edx, sizeof ARG                                     ; Advance to the next parameter
            loop    Scan

            mov     eax, FnPtr
            mov     edx, FnArg

            ret

GetFnPtr    endp

;---------------------------------------------------------------------
; MakeArgList                   Build a list of command-line arguments
;---------------------------------------------------------------------
; pArgV:    Pointer to the argument pointer array (argv[16]:DWORD)
; pArgC:    Pointer to the argument counter variable (argc:BYTE)
; pCmd:     Pointer to the command-line string
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

MakeArgList proc    pArgV:DWORD, pArgC:DWORD, pCmd:DWORD

            local   IgnoreBlanks:BYTE
            local   QuotesOpen:BYTE

            mov     IgnoreBlanks, 1
            mov     QuotesOpen, 0

            mov     eax, pArgC
            mov     byte ptr [eax], 0
            
            mov     eax, pCmd

Scan:       cmp     byte ptr [eax], 0                                   ; End of line?
            je      Done

            cmp     byte ptr [eax], ' '                                 ; Is blank?
            sete    dl

            cmp     byte ptr [eax], '"'                                 ; Is double quotes?
            sete    dh

            shl     dl, 1                                               ; Assemble a decision bitmap
            add     dl, dh
            shl     dl, 1
            add     dl, IgnoreBlanks
            shl     dl, 1
            add     dl, QuotesOpen

            cmp     dl, 1000y                                           ; End of argument
            je      EndArg

            cmp     dl, 0010y                                           ; Beginning of new argument
            je      NewArg

            cmp     dl, 0110y                                           ; Opening quotes
            je      NewQArg

            cmp     dl, 0111y                                           ; Closing quotes
            je      EndQArg

Next:       inc     eax                                                 ; Advance to the next char
            jmp     Scan

EndArg:     mov     byte ptr [eax], 0                                   ; Add trailing null
            mov     IgnoreBlanks, 1                                     ; Ignore blanks until next argument
            jmp     Next

NewArg:     fn      AddArg, eax, pArgV, pArgC                           ; Add pointer to arguments list
            mov     IgnoreBlanks, 0                                     ; Stop ignoring blanks
            jmp     Next

NewQArg:    inc     eax                                                 ; Skip the quotes mark
            fn      AddArg, eax, pArgV, pArgC                           ; Add pointer to arguments list
            mov     QuotesOpen, 1                                       ; Flag the open quotes
            mov     IgnoreBlanks, 1                                     ; Ignore blanks inside quotes
            jmp     Next

EndQArg:    mov     byte ptr [eax], 0                                   ; Remove the quotes mark
            mov     QuotesOpen, 0                                       ; Reset the open quotes flag
            jmp     Next

Done:       ret

MakeArgList endp

;---------------------------------------------------------------------
; AddArg                 Insert pointer in the argument pointers array
;---------------------------------------------------------------------
; pArg:     Pointer to the argument to add
; pArgV:    Pointer to the argument pointer array (argv[16]:DWORD)
; pArgC:    Pointer to the argument counter variable (argc:BYTE)
;---------------------------------------------------------------------
; EAX:      Pointer to the same argument
;---------------------------------------------------------------------

AddArg      proc    pArg:DWORD, pArgV:DWORD, pArgC:DWORD

            mov     edx, pArgC
            cmp     byte ptr [edx], 16
            jnb     Done

            mov     eax, sizeof DWORD                                   ; Compute the address of the next slot
            mul     byte ptr [edx]
            add     eax, pArgV

            mov     edx, pArg                                           ; Save pointer to the argument
            mov     [eax], edx

            mov     edx, pArgC
            inc     byte ptr [edx]                                      ; Increment the argument counter

Done:       mov     eax, pArg
            ret

AddArg      endp

;---------------------------------------------------------------------
; ToLower                                 Convert string to lower case
;---------------------------------------------------------------------
; pStr:     Pointer to a null-terminated string
;---------------------------------------------------------------------
; EAX:      Pointer to the same string
;---------------------------------------------------------------------

ToLower     proc    pStr:DWORD

            mov     eax, pStr
            
Scan:       mov     dl, [eax]

            cmp     dl, 0
            je      Done

            cmp     dl, 'A'
            jb      Next

            cmp     dl, 'Z'
            ja      Next

            or      byte ptr [eax], 'a' - 'A'

Next:       inc     eax
            jmp     Scan

Done:       mov     eax, pStr

            ret

ToLower     endp

;---------------------------------------------------------------------
; PrintError                     Print formatted system error messages
;---------------------------------------------------------------------
; dwError:  Error code to print
;---------------------------------------------------------------------
; EAX:      Original error code
;---------------------------------------------------------------------

PrintError  proc    dwError:DWORD

            local   pMessage:PTR

            fn      FormatMessage, FORMAT_MESSAGE_FROM_SYSTEM+FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, dwError, 0, addr pMessage, 0, NULL

            or      eax ,eax
            jz      @F

            fn      StdErr, pMessage

@@:         fn      LocalFree, pMessage

            mov     eax, dwError
            
            ret

PrintError  endp

END EntryPoint

