;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

TrsCat          proto   :DWORD, :DWORD, :DWORD
Trs2Win         proto   :DWORD, :DWORD
ParseCmdLine    proto   :DWORD
AddArgPtr       proto   :DWORD
ToUpper         proto   :DWORD
FormatError     proto   :DWORD, :DWORD, :DWORD
ShowError       proto   :DWORD

;---------------------------------------------------------------------
; TrsCat               Concatenate Name, Extension and a trailing null 
;---------------------------------------------------------------------
; pBuffer:      Pointer to target buffer
; pFileName:    Pointer to 8-byte file name
; pFileExt:     Pointer to 3-byte file extension
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

TrsCat      proc    pBuffer:DWORD, pFileName:DWORD, pFileExt:DWORD

            push    esi
            push    edi

            mov     esi, pFileName
            mov     edi, pBuffer
            mov     ecx, 8
            cld
            rep     movsb

            mov     esi, pFileExt
            mov     ecx, 3
            rep     movsb

            mov     byte ptr [edi], 0

            pop     edi
            pop     esi

            ret

TrsCat      endp

;---------------------------------------------------------------------
; Trs2Win                  Convert TRS-80 filename to Windows filename
;---------------------------------------------------------------------
; pTrsName: Pointer to TRS-formatted file name and extension
; pWinName: Pointer to buffer for the windows-formatted filename
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

Trs2Win     proc    pTrsName:DWORD, pWinName:DWORD

            push    esi
            push    edi

            mov     esi, pTrsName                                       ; Set pointer at the file name
            mov     edi, pWinName
            mov     ecx, 8

CopyName:   cmp     byte ptr [esi], ' '                                 ; Copy the file name
            je      CheckExt
            movsb
            loop    CopyName

CheckExt:   mov     esi, pTrsName                                       ; Set pointer at the file extension
            add     esi, 8
            mov     ecx, 3

            cmp     byte ptr [esi], ' '
            je      AddNull

            mov     al, '.'                                             ; Add a dot
            stosb

CopyExt:    movsb                                                       ; Copy the file extension
            cmp     byte ptr [esi], ' '
            je      AddNull
            loop    CopyExt
            
AddNull:    mov     al, 0                                               ; Add a trailing null
            stosb

            pop     edi
            pop     esi

            ret

Trs2Win     endp

;---------------------------------------------------------------------
; ParseCmdLine                    Build a list of individual arguments
;---------------------------------------------------------------------
; pCmdLine: Pointer to command-line string
;---------------------------------------------------------------------
; No return
;---------------------------------------------------------------------

ParseCmdLine        proc    pCmdLine:DWORD 

            local   IgnoreBlanks:BYTE
            local   QuotesOpen:BYTE

            mov     IgnoreBlanks, 1
            mov     QuotesOpen, 0

            mov     argc, 0                                             ; Reset argument count

            mov     eax, pCmdLine

Scan:       cmp     byte ptr [eax], 0                                   ; End of line?
            je      Done

            cmp     byte ptr [eax], ' '                                 ; Is blank?
            sete    dl

            cmp     byte ptr [eax], '"'                                 ; Is double quotes?
            sete    dh

            shl     dl, 1                                               ; Assemble a decision bitmap
            add     dl, dh
            shl     dl, 1
            add     dl, IgnoreBlanks
            shl     dl, 1
            add     dl, QuotesOpen

            cmp     dl, 1000y                                           ; End of argument
            je      EndArg

            cmp     dl, 0010y                                           ; Beginning of new argument
            je      NewArg

            cmp     dl, 0110y                                           ; Opening quotes
            je      NewQArg

            cmp     dl, 0111y                                           ; Closing quotes
            je      EndQArg

Next:       inc     eax                                                 ; Advance to the next char
            jmp     Scan

EndArg:     mov     byte ptr [eax], 0                                   ; Add trailing null
            mov     IgnoreBlanks, 1                                     ; Ignore blanks until next argument
            jmp     Next

NewArg:     fn      AddArgPtr, eax                                  ; Add pointer to arguments list
            mov     IgnoreBlanks, 0                                     ; Stop ignoring blanks
            jmp     Next

NewQArg:    inc     eax                                                 ; Skip the quotes mark
            fn      AddArgPtr, eax                                  ; Add pointer to arguments list
            mov     QuotesOpen, 1                                       ; Flag the open quotes
            mov     IgnoreBlanks, 1                                     ; Ignore blanks inside quotes
            jmp     Next

EndQArg:    mov     byte ptr [eax], 0                                   ; Remove the quotes mark
            mov     QuotesOpen, 0                                       ; Reset the open quotes flag
            jmp     Next

Done:       ret

ParseCmdLine        endp

;---------------------------------------------------------------------
; AddArgPtr                       Add an argument to the pointers list
;---------------------------------------------------------------------
; pArg: Pointer to the argument
;---------------------------------------------------------------------
; EAX:  Pointer to the same argument
;---------------------------------------------------------------------

AddArgPtr   proc    pArg:DWORD

            cmp     argc, 16                                            ; Check whether the pointers list is not full
            jnb     Done

            mov     eax, sizeof DWORD                                   ; Compute the address of the next slot
            mul     argc
            add     eax, offset argv

            mov     edx, pArg                                           ; Save pointer to the argument
            mov     [eax], edx

            inc     argc                                                ; Increment the argument counter

Done:       mov     eax, edx

            ret

AddArgPtr   endp

;---------------------------------------------------------------------
; ToUpper                                 Convert string to upper case
;---------------------------------------------------------------------
; pString:  Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Pointer to the same string
;---------------------------------------------------------------------

ToUpper     proc    pString:DWORD

Scan:       mov     dl, [eax]

            cmp     dl, 0
            je      Done

            cmp     dl, 'a'
            jb      Next

            cmp     dl, 'z'
            ja      Next

            sub     byte ptr [eax], 'a' - 'A'

Next:       inc     eax
            jmp     Scan

Done:       mov     eax, pString

            ret

ToUpper     endp

;---------------------------------------------------------------------
; FormatError                                  Format an error message
;---------------------------------------------------------------------

FormatError proc    Error:DWORD, pBuffer:DWORD, Len:DWORD

            fn      FormatMessage, FORMAT_MESSAGE_FROM_SYSTEM, NULL, Error, 0, pBuffer, Len, NULL

            or      eax, eax
            jnz     Done

            cmp     Len, 43
            jb      Done

            fn      wsprintf, pBuffer, "Error %u (GetLastError=%u)", Error, eax
            
Done:       ret

FormatError endp

;---------------------------------------------------------------------
; ShowError                          Display a formatted error message
;---------------------------------------------------------------------

ShowError   proc    Error:DWORD

            local   Message[1024]:BYTE

            fn      FormatError, Error, addr Message, sizeof Message

            fn      MessageBox, NULL, addr Message, NULL, MB_ICONERROR+MB_OK

            ret

ShowError   endp

