;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

GUI_Main            proto   :HWND, :UINT, :WPARAM, :LPARAM
GUI_Main_Cmd        proto   :HWND, :WPARAM
GUI_Main_Notif      proto   :HWND, :LPARAM
GUI_Main_Init       proto   :HWND
GUI_Main_End        proto   :HWND
GUI_Main_Resize     proto   :HWND, :LPARAM
GUI_Main_HdrClick   proto   :HWND, :DWORD

;---------------------------------------------------------------------
; GUI_Main                                       Main dialog procedure
;---------------------------------------------------------------------

GUI_Main    proc    hWnd:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM

    ; --- Handle the WM_INITDIALOG message ---
    
            cmp     uMsg, WM_INITDIALOG
            jne     @F

            fn      GUI_Main_Init, hWnd
                
            mov     eax, TRUE
            jmp     Done
                
    ; --- Handle the WM_CLOSE message ---
    
@@:         cmp     uMsg, WM_CLOSE
            jne     @F

            fn      GUI_Main_End, hWnd

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the WM_SIZE message ---

@@:         cmp     uMsg, WM_SIZE
            jne     @F

            fn      GUI_Main_Resize, hWnd, lParam

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the WM_COMMAND message ---
    
@@:         cmp     uMsg, WM_COMMAND
            jne     @F

            fn      GUI_Main_Cmd, hWnd, wParam
            jmp     Done

    ; --- Handle the WM_NOTIFY message ---

@@:         cmp     uMsg, WM_NOTIFY
            jne     @F

            fn      GUI_Main_Notif, hWnd, lParam
            jmp     Done

    ; --- Message not processed ---

@@:         mov     eax, FALSE
            
Done:       ret

GUI_Main    endp

;---------------------------------------------------------------------
; GUI_Main_Cmd                            Main dialog command handling
;---------------------------------------------------------------------

GUI_Main_Cmd        proc    hWnd:HWND, wParam:WPARAM

    ; --- Handle the IDM_FILE_OPEN command ---
    
            cmp     wParam, IDM_FILE_OPEN
            jne     @F
                
            fn      GUI_OpenFileFromSysDlg, hWnd

            mov     eax, TRUE
            jmp     Done
                    
    ; --- Handle the IDM_FILE_CLOSE command ---
    
@@:         cmp     wParam, IDM_FILE_CLOSE
            jne     @F
                
            fn      GUI_CloseFile, hWnd
            
            mov     eax, TRUE
            jmp     Done
                    
    ; --- Handle the IDM_FILE_EXPORT command ---
    
@@:         cmp     wParam, IDM_FILE_EXPORT
            jne     @F
                
            fn      GUI_Export, hWnd
            
            mov     eax, TRUE
            jmp     Done
                    
    ; --- Handle the IDM_FILE_EXIT command ---
    
@@:         cmp     wParam, IDM_FILE_EXIT
            jne     @F
            
            fn      GUI_Main_End, hWnd

            mov     eax, TRUE
            jmp     Done
                    
    ; --- Handle the IDM_TOOLS_PROP command ---
    
@@:         cmp     wParam, IDM_TOOLS_PROP
            jne     @F

            fn      GetModuleHandle, NULL

            fn      DialogBoxParam, eax, IDD_PROP, hWnd, offset GUI_Prop, NULL
                
            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_TOOLS_EDITOR command ---
    
@@:         cmp     wParam, IDM_TOOLS_EDITOR
            jne     @F

            fn      GetModuleHandle, NULL

            fn      DialogBoxParam, eax, IDD_VIEW, hWnd, offset GUI_Viewer, NULL
                
            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_OPTIONS_ASSOCIATE command ---
    
@@:         cmp     wParam, IDM_OPTIONS_ASSOCIATE
            jne     @F

            fn      GUI_Menu_Associate, hWnd

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_OPTIONS_AUTODETECT command ---
    
@@:         cmp     wParam, IDM_OPTIONS_AUTODETECT
            jne     @F

            fn      GUI_Menu_AutoDetect, hWnd

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_OPTIONS_HIDESYS command ---
    
@@:         cmp     wParam, IDM_OPTIONS_HIDESYS
            jne     @F

            fn      GUI_Menu_HideSys, hWnd

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_OPTIONS_HIDEINV command ---
    
@@:         cmp     wParam, IDM_OPTIONS_HIDEINV
            jne     @F

            fn      GUI_Menu_HideInv, hWnd

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the IDM_ABOUT command ---
    
@@:         cmp     wParam, IDM_HELP_ABOUT
            jne     @F

            fn      GetModuleHandle, NULL

            fn      DialogBoxParam, eax, IDD_ABOUT, hWnd, offset GUI_About, NULL
                
            mov     eax, TRUE
            jmp     Done

    ; --- Command not processed ---
    
@@:         mov     eax, FALSE

Done:       ret

GUI_Main_Cmd        endp

;---------------------------------------------------------------------
; GUI_Main_Notif                     Main dialog notification handling
;---------------------------------------------------------------------

GUI_Main_Notif      proc    hWnd:HWND, lParam:LPARAM

            mov     eax, lParam                             

        ; --- Handle the LVN_COLUMNCLICK notification ---

            cmp     (NMHDR PTR [eax]).code, LVN_COLUMNCLICK             ; Was a column header clicked?
            jne     @F

            fn      GUI_Main_HdrClick, (NMHDR PTR [eax]).hwndFrom, (NM_LISTVIEW PTR [eax]).iSubItem

            mov     eax, TRUE
            jmp     Done

    ; --- Notification not processed ---
    
@@:         mov     eax, FALSE

Done:       ret

GUI_Main_Notif      endp

;---------------------------------------------------------------------
; GUI_Main_Init                             Main dialog initialization
;---------------------------------------------------------------------

GUI_Main_Init       proc    hWnd:HWND

            local   hIcon:HICON
            local   hList:HWND
            local   hMenu:HMENU
            local   Rect:RECT

    ; --- Load the application icon --- 

            fn      GetModuleHandle, NULL

            fn      LoadIcon, eax, IDI_ICON
            mov     hIcon, eax
            
            fn      SendMessage, hWnd, WM_SETICON, ICON_BIG, hIcon
            fn      SendMessage, hWnd, WM_SETICON, ICON_SMALL, hIcon

    ; --- Initialize the list control --- 

            fn      GetDlgItem, hWnd, IDC_LIST
            mov     hList, eax

            fn      SendMessage, hList, LVM_INSERTCOLUMN, 0, offset GUI_LVC_Name
            fn      SendMessage, hList, LVM_INSERTCOLUMN, 1, offset GUI_LVC_Extension
            fn      SendMessage, hList, LVM_INSERTCOLUMN, 2, offset GUI_LVC_Size
            fn      SendMessage, hList, LVM_INSERTCOLUMN, 3, offset GUI_LVC_Date
            fn      SendMessage, hList, LVM_INSERTCOLUMN, 4, offset GUI_LVC_Attributes

    ; --- Load the program settings ---

            fn      GUI_LoadSettings

    ; --- Set window position and size ---

            fn      GetDesktopWindow

            mov     edx, eax

            fn      GetWindowRect, edx, addr Rect

            mov     eax, Rect.right
            sub     eax, Rect.left
            sub     eax, GUI_Config.WindowWidth
            shr     eax, 1
            mov     Rect.left, eax

            mov     eax, Rect.bottom
            sub     eax, Rect.top
            sub     eax, GUI_Config.WindowHeight
            shr     eax, 1
            mov     Rect.top, eax

            fn      SetWindowPos, hWnd, HWND_TOP, Rect.left, Rect.top, GUI_Config.WindowWidth, GUI_Config.WindowHeight, SWP_SHOWWINDOW

    ; --- Get the file association status ---

            fn      GUI_GetClass

            mov     GUI_Config.Associate, eax

    ; --- Adjust the menu options states ---

            fn      GetMenu, hWnd                                       ; Get a handle to the Menu
            mov     hMenu, eax

            mov     edx, offset GUI_Settings
            mov     ecx, GUI_MMI_Count

@@:         mov     eax, (GUI_SETTING PTR [edx]).pValue                 ; Get the setting's current value

            cmp     dword ptr [eax], 0                                  ; Check whether it is non-zero
            je      Next

            mov     eax, (GUI_SETTING PTR [edx]).pMII                   ; Get the setting's MMI

            mov     (MENUITEMINFO PTR [eax]).fState, MFS_CHECKED        ; Check the menu option

            push    ecx
            push    edx

            fn      SetMenuItemInfo, hMenu, (MENUITEMINFO PTR [eax]).wID, FALSE, eax

            pop     edx
            pop     ecx

Next:       add     edx, sizeof GUI_SETTING
            loop    @B

    ; --- Check whether a file has been passed through the command-line ---

            fn      GUI_OpenFileFromCmdLine, hWnd

Done:       ret

GUI_Main_Init       endp

;---------------------------------------------------------------------
; GUI_Main_End                                Main dialog finalization
;---------------------------------------------------------------------

GUI_Main_End        proc    hWnd:HWND

            local   Rect:RECT

    ; --- Check whether there is a file open ---

            cmp     GUI_hFile, INVALID_HANDLE_VALUE
            je      @F

            fn      GUI_CloseFile, hWnd

    ; --- Save current settings ---

@@:         fn      IsZoomed, hWnd                                      ; Is the window maximized?
            jnz     Maximized

            fn      GetWindowRect, hWnd, addr Rect
            jmp     @F

Maximized:  fn      GetDesktopWindow

            mov     edx, eax

            fn      GetWindowRect, edx, addr Rect

@@:         mov     eax, Rect.right
            sub     eax, Rect.left
            mov     GUI_Config.WindowWidth, eax

            mov     eax, Rect.bottom
            sub     eax, Rect.top
            mov     GUI_Config.WindowHeight, eax

            fn      GUI_SaveSettings

 ; --- Terminate the dialog window ---

            fn      EndDialog, hWnd, 0

Done:       ret

GUI_Main_End        endp

;---------------------------------------------------------------------
; GUI_Main_Resize                       Main dialog resizing procedure
;---------------------------------------------------------------------

GUI_Main_Resize     proc    hWnd:HWND, lParam:LPARAM

            local   hList:HWND

            fn      GetDlgItem, hWnd, IDC_LIST                          ; Get handle to list control
            mov     hList, eax

            mov     edx, lParam                                         ; Get the new window size
            movzx   ecx, dx                                             ; Width
            shr     edx, 16                                             ; Height
            
            fn      MoveWindow, hList, 0, 0, ecx, edx, TRUE             ; Resize the list control
            
            ret

GUI_Main_Resize     endp

;---------------------------------------------------------------------
; GUI_Main_HdrClick                              Header click handling
;---------------------------------------------------------------------

GUI_Main_HdrClick   proc    hWnd:HWND, Column:DWORD

            mov     ecx, Column
            btc     GUI_Main_Sort_Sel, ecx                              ; Remember the selected column
            setc    GUI_Main_Sort_Order                                 ; Invert the sorting order if appropriate
            
            fn      SendMessage, hWnd, LVM_SORTITEMS, Column, offset GUI_ColumnSort
            
            ret

GUI_Main_HdrClick   endp

