;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

GUI_Menu_Associate  proto   :HWND
GUI_Menu_AutoDetect proto   :HWND
GUI_Menu_HideSys    proto   :HWND
GUI_Menu_HideInv    proto   :HWND
GUI_Menu_SetState   proto   :HWND, :DWORD, :DWORD

;---------------------------------------------------------------------
; GUI_Menu_Associate           Handle the File Association menu option
;---------------------------------------------------------------------
; hWnd:     Window handle
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

GUI_Menu_Associate  proc    hWnd:HWND

            cmp     GUI_Config.Associate, 0
            je      Set

            fn      GUI_ResetClass

            or      eax, eax
            jnz     Error

            jmp     @F

Set:        fn      GUI_SetClass

            or      eax, eax
            jnz     Error

            fn      MessageBox, NULL, "DSK/DMK files are now associated with VDISK", "Success", MB_ICONINFORMATION+MB_OK

@@:         xor     GUI_Config.Associate, 1

            fn      GUI_Menu_SetState, hWnd, GUI_MMI_FA, GUI_Config.Associate

            jmp     Done

Error:      fn      ShowError, eax

Done:       ret

GUI_Menu_Associate  endp

;---------------------------------------------------------------------
; GUI_Menu_AutoDetect               Handle the Auto-Detect menu option
;---------------------------------------------------------------------
; hWnd:     Window handle
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

GUI_Menu_AutoDetect proc    hWnd:HWND

            xor     GUI_Config.AutoDetect, 1                            ; Invert the current selection

            fn      GUI_Menu_SetState, hWnd, GUI_MMI_AD, GUI_Config.AutoDetect

            ret

GUI_Menu_AutoDetect endp

;---------------------------------------------------------------------
; GUI_Menu_HideSys            Handle the Hide System Files menu option
;---------------------------------------------------------------------
; hWnd:     Window handle
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

GUI_Menu_HideSys    proc    hWnd:HWND

            xor     GUI_Config.HideSystem, 1                            ; Invert the current selection

            fn      GUI_Menu_SetState, hWnd, GUI_MMI_SS, GUI_Config.HideSystem

            cmp     GUI_hFile, INVALID_HANDLE_VALUE                     ; Check whether a file is open
            je      Done

            fn      GUI_List, hWnd                                      ; Refresh directory listing

Done:       ret

GUI_Menu_HideSys    endp

;---------------------------------------------------------------------
; GUI_Menu_HideInv         Handle the Hide Invisible Files menu option
;---------------------------------------------------------------------
; hWnd:     Window handle
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

GUI_Menu_HideInv    proc    hWnd:HWND

            xor     GUI_Config.HideInvisible, 1                         ; Invert the current selection

            fn      GUI_Menu_SetState, hWnd, GUI_MMI_SI, GUI_Config.HideInvisible

            cmp     GUI_hFile, INVALID_HANDLE_VALUE                     ; Check whether a is file open
            je      Done

            fn      GUI_List, hWnd                                      ; Refresh directory listing

Done:       ret

GUI_Menu_HideInv    endp

;---------------------------------------------------------------------
; GUI_Menu_SetState                              Set menu option state
;---------------------------------------------------------------------
; hWnd:     Window handle
; Item:     Menu item to be inverted (0 - n)
; State:    New menu state (0 or 1)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

GUI_Menu_SetState   proc    hWnd:HWND, Item:DWORD, State:DWORD

            local   hMenu:HMENU

            fn      GetMenu, hWnd                                       ; Get handle to the Menu
            mov     hMenu, eax

            mov     eax, sizeof GUI_SETTING                             ; Get pointer to GUI_Settings[Item]
            mul     Item
            add     eax, offset GUI_Settings

            mov     edx, (GUI_SETTING PTR [eax]).pValue                 ; Get pointer to the setting's current value

            mov     ecx, State                                          ; Set new value
            mov     [edx], ecx

            mov     edx, (GUI_SETTING PTR [eax]).pMII                   ; Get pointer to the setting's Menu Item Info

            cmp     State, 0
            jne     @F

            mov     ecx, MFS_UNCHECKED
            jmp     SetState

@@:         mov     ecx, MFS_CHECKED

SetState:   mov     (MENUITEMINFO PTR [edx]).fState, ecx                ; Set new menu state

            fn      SetMenuItemInfo, hMenu, (MENUITEMINFO PTR [edx]).wID, FALSE, edx

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     eax, NO_ERROR

Done:       ret

GUI_Menu_SetState   endp

