;---------------------------------------------------------------------
; Graphical User Interface             Copyright (C) 2009 Miguel Dutra
;---------------------------------------------------------------------

GUI_Prop            proto   :HWND, :UINT, :WPARAM, :LPARAM
GUI_Prop_Init       proto   :HWND

;---------------------------------------------------------------------
; GUI_Prop                                      Disk properties dialog
;---------------------------------------------------------------------

GUI_Prop    proc    hWnd:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM

    ; --- Handle the WM_INITDIALOG message ---
    
            cmp     uMsg, WM_INITDIALOG
            jne     @F

            fn      GUI_Prop_Init, hWnd

            mov     eax, TRUE
            jmp     Done
                
    ; --- Handle the WM_CLOSE message ---
    
@@:         cmp     uMsg, WM_CLOSE
            jne     @F

            fn      EndDialog, hWnd, 0

            mov     eax, TRUE
            jmp     Done

    ; --- Handle the WM_COMMAND message ---

@@:         cmp     uMsg, WM_COMMAND
            jne     @F

            cmp     wParam, IDC_PROP_CLOSE
            jne     @F

            fn      EndDialog, hWnd, 0

            mov     eax, TRUE
            jmp     Done

    ; --- Message not processed ---

@@:         mov     eax, FALSE
            
Done:       ret

GUI_Prop    endp

;---------------------------------------------------------------------
; GUI_Prop_Init                  Disk properties dialog initialization
;---------------------------------------------------------------------

GUI_Prop_Init       proc    hWnd:HWND

            local   hControl:HWND
            local   String[32]:BYTE

            cmp     GUI_hFile, INVALID_HANDLE_VALUE
            je      Done

    ; --- Init the File Format edit control --- 

            fn      GetDlgItem, hWnd, IDC_PROP_FF
            mov     hControl, eax

            mov     ecx, GUI_SEL_VDI_Count
            mov     edx, offset GUI_SEL_VDI_Array
           
FindVDI:    mov     eax, (GUI_SEL_VDI PTR [edx]).pVTBL

            cmp     eax, GUI_VDI.VPTR
            jne     NextVDI

            fn      SendMessage, hControl, WM_SETTEXT, 0, (GUI_SEL_VDI PTR [edx]).pName
            jmp     @F

NextVDI:    add     edx, sizeof GUI_SEL_VDI
            loop    FindVDI

    ; --- Init the Operating System edit control --- 

@@:         fn      GetDlgItem, hWnd, IDC_PROP_OS
            mov     hControl, eax

            mov     ecx, GUI_SEL_OSI_Count
            mov     edx, offset GUI_SEL_OSI_Array
           
FindOSI:    mov     eax, (GUI_SEL_OSI PTR [edx]).pVTBL

            cmp     eax, GUI_OSI.VPTR
            jne     NextOSI

            fn      SendMessage, hControl, WM_SETTEXT, 0, (GUI_SEL_OSI PTR [edx]).pName
            jmp     @F

NextOSI:    add     edx, sizeof GUI_SEL_OSI
            loop    FindOSI

    ; --- Init the Tracks edit control --- 

@@:         mcall   VDI_VTBL, GetTracks, addr GUI_VDI

            fn      wsprintf, addr String, "%3u", eax

            fn      GetDlgItem, hWnd, IDC_PROP_TRACKS

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the Sides edit control --- 

            mcall   VDI_VTBL, GetSides, addr GUI_VDI

            fn      wsprintf, addr String, "%3u", eax

            fn      GetDlgItem, hWnd, IDC_PROP_SIDES

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the Sides edit control --- 

            mcall   VDI_VTBL, GetSectorsPerTrack, addr GUI_VDI

            fn      wsprintf, addr String, "%3u", eax

            fn      GetDlgItem, hWnd, IDC_PROP_SECTORS

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the First Sector edit control --- 

            mcall   VDI_VTBL, GetFirstSector, addr GUI_VDI
            push    eax

            mcall   VDI_VTBL, GetFirstSide, addr GUI_VDI
            push    eax

            mcall   VDI_VTBL, GetFirstTrack, addr GUI_VDI
            push    eax

            fn      wsprintf, addr String, "%02u:%u:%02u"

            add     esp, 3*4

            fn      GetDlgItem, hWnd, IDC_PROP_1STSECTOR

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the Disk Name edit control --- 

            fn      memfill, addr String, sizeof String, 0

            mcall   OSI_VTBL, GetDiskName, addr GUI_OSI, addr String, sizeof String

            fn      GetDlgItem, hWnd, IDC_PROP_DISKNAME

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the Disk Date edit control --- 

            fn      memfill, addr String, sizeof String, 0

            mcall   OSI_VTBL, GetDiskDate, addr GUI_OSI, addr String, sizeof String

            fn      GetDlgItem, hWnd, IDC_PROP_DISKDATE

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

    ; --- Init the DOS Version edit control --- 

            mcall   OSI_VTBL, GetDosVersion, addr GUI_OSI

            mov     edx, eax
            and     edx, 0Fh
            push    edx

            shr     eax, 4
            push    eax

            fn      wsprintf, addr String, "%u.%u"

            add     esp, 2*4

            fn      GetDlgItem, hWnd, IDC_PROP_DOSVER

            mov     edx, eax

            fn      SendMessage, edx, WM_SETTEXT, 0, addr String

Done:       ret

GUI_Prop_Init       endp

