;---------------------------------------------------------------------
; MICRODOS Operating System Interface       Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

.data?

;---------------------------------------------------------------------
; MICRODOS File Structure
;---------------------------------------------------------------------

MO_File struct
    FileName            BYTE        8   dup (?)
    Extension           BYTE        3   dup (?)
    FileSize            DWORD       ?
    FileAttr            DWORD       ?
    FilePointer         DWORD       ?
MO_File ends

;---------------------------------------------------------------------
; MICRODOS Class
;---------------------------------------------------------------------

CMICRODOS   struct
    VPTR                DWORD       ?                                   ; Pointer to VTBL
    pVDI                DWORD       ?                                   ; Pointer to Virtual Disk Interface
    FirstTrack          BYTE        ?                                   ; Number of the first track on disk (0 or 1)
    FirstSide           BYTE        ?                                   ; Number of the first disk side (0 or 1)
    FirstSector         BYTE        ?                                   ; Number of the first sector on track (0 or 1)
    Sides               BYTE        ?                                   ; Number of disk sides (1 or 2)
    SectorsPerTrack     BYTE        ?                                   ; Number of sectors per track
    Sectors             WORD        ?                                   ; Total number of disk sectors
    FileIndex           BYTE        ?                                   ; Current file index
    dwLastError         DWORD       ?                                   ; Code of last error occurred in an internal function
    Files               MO_File     2 dup (<>)                          ; System and data files
CMICRODOS   ends

.data

;---------------------------------------------------------------------
; MICRODOS Virtual Table
;---------------------------------------------------------------------

MICRODOS_VTBL   struct
    Load                DWORD       MO_Load                             ; OSI-derived functions
    Unload              DWORD       MO_Unload
    Find                DWORD       MO_Find
    Open                DWORD       MO_Open
    Create              DWORD       MO_Create
    Seek                DWORD       MO_Seek
    Read                DWORD       MO_Read
    Write               DWORD       MO_Write
    Rename              DWORD       MO_Rename
    Delete              DWORD       MO_Delete
    GetLastError        DWORD       MO_GetLastError
    GetDosVersion       DWORD       MO_GetDosVersion
    SetDosVersion       DWORD       MO_SetDosVersion
    GetDiskName         DWORD       MO_GetDiskName
    SetDiskName         DWORD       MO_SetDiskName
    GetDiskDate         DWORD       MO_GetDiskDate
    SetDiskDate         DWORD       MO_SetDiskDate
    GetFilePos          DWORD       MO_GetFilePos
    GetFileName         DWORD       MO_GetFileName
    GetFileExt          DWORD       MO_GetFileExt
    GetFileSize         DWORD       MO_GetFileSize
    GetFileDate         DWORD       MO_GetFileDate
    SetFileDate         DWORD       MO_SetFileDate
    GetFileAttr         DWORD       MO_GetFileAttr
    SetFileAttr         DWORD       MO_SetFileAttr
    InitFiles           DWORD       MO_InitFiles                        ; MICRODOS-specific functions
    ReadSysFile         DWORD       MO_ReadSysFile
    ReadDatFile         DWORD       MO_ReadDatFile
    Sector2CHS          DWORD       MO_Sector2CHS
MICRODOS_VTBL   ends

MICRODOS                MICRODOS_VTBL   <>

;---------------------------------------------------------------------
; MICRODOS Boot Signature
;---------------------------------------------------------------------

MO_Signature            BYTE        "MICRODOS"

;---------------------------------------------------------------------
; MICRODOS File Templates
;---------------------------------------------------------------------

MO_SysFile              MO_File     <"MICRODOS", "SYS", 20*256, 0010h>
MO_DatFile              MO_File     <"MICRODOS", "TXT",      0, 0000h>

.code

;---------------------------------------------------------------------
; MO_Load                                         Initialize interface
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Load     proc    pMO:DWORD, pVDI:DWORD

            local   Boot[256]:BYTE
            local   Tracks:BYTE
            local   Sides:BYTE
            local   SectorsPerTrack:BYTE
            local   FirstTrack:BYTE
            local   FirstSide:BYTE
            local   FirstSector:BYTE
            local   Sectors:WORD

    ; --- Get disk information ---

            mcall   VDI_VTBL, GetTracks, pVDI
            mov     Tracks, al

            mcall   VDI_VTBL, GetSides, pVDI
            mov     Sides, al

            mcall   VDI_VTBL, GetSPT, pVDI
            mov     SectorsPerTrack, al

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     FirstTrack, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     FirstSide, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     FirstSector, al

    ; --- Load the boot sector ---

            mcall   VDI_VTBL, Read, pVDI, addr Boot, sizeof Boot, FirstTrack, FirstSide, FirstSector
            
            or      eax, eax
            jnz     Done

    ; --- Check whether it's a MICRODOS disk ---

            push    esi
            push    edi

            mov     esi, offset MO_Signature
            lea     edi, Boot
            add     edi, 241
            mov     ecx, sizeof MO_Signature
            cld
            repe    cmpsb

            pop     edi
            pop     esi

            jne     IntError

    ; --- Compute the total number of disk sectors ---

            mov     al, Tracks                                          ; Sectors = Tracks * SectorsPerTrack * Sides
            mul     Sides
            mul     SectorsPerTrack
            mov     Sectors, ax

    ; --- Update member class variables ---
    
            mov     edx, pMO
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (CMICRODOS PTR [edx]).pVDI, eax

            mov     al, FirstTrack                                      ; Update number of the first track on disk
            mov     (CMICRODOS PTR [edx]).FirstTrack, al

            mov     al, FirstSide                                       ; Update number of the first disk side
            mov     (CMICRODOS PTR [edx]).FirstSide, al

            mov     al, FirstSector                                     ; Update number of the first track on disk
            mov     (CMICRODOS PTR [edx]).FirstSector, al

            mov     al, Sides                                           ; Update number of disk sides
            mov     (CMICRODOS PTR [edx]).Sides, al

            mov     al, SectorsPerTrack                                 ; Update number of sectors per track
            mov     (CMICRODOS PTR [edx]).SectorsPerTrack, al

            mov     ax, Sectors                                         ; Update the total number of disk sectors
            mov     (CMICRODOS PTR [edx]).Sectors, ax

            mov     (CMICRODOS PTR [edx]).FileIndex, 0                  ; Reset the current file index

    ; --- Initialize the file structures ---
    
            mcall   MICRODOS_VTBL, InitFiles, pMO
            jmp     Done

IntError:   mov     eax, ERROR_NOT_DOS_DISK

Done:       ret

MO_Load     endp

;---------------------------------------------------------------------
; MO_Unload                                          Release interface
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Unload   proc    pMO:DWORD

            mov     eax, NO_ERROR
            ret

MO_Unload   endp

;---------------------------------------------------------------------
; MO_Find                                Search files in the directory
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; Flag:     OSI_FINDFIRST or OSI_FINDNEXT
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

MO_Find     proc    pMO:DWORD, Flag:BYTE

            mov     edx, pMO

            inc     (CMICRODOS PTR [edx]).FileIndex                     ; Increment the file index

            cmp     Flag, OSI_FINDFIRST                                 ; Requested to look for first file?
            jne     @F

            mov     (CMICRODOS PTR [edx]).FileIndex, 0                  ; Reset the file index back to zero

@@:         cmp     (CMICRODOS PTR [edx]).FileIndex, 2                  ; Check the array boundary
            jb      @F

            mov     eax, pMO
            mov     (CMICRODOS PTR [eax]).dwLastError, ERROR_NO_MORE_FILES  ; There are no more files.
            mov     eax, NULL
            jmp     Done

@@:         mov     eax, sizeof MO_File                                 ; Compute pointer to Files[FileIndex]
            mul     (CMICRODOS PTR [edx]).FileIndex
            lea     edx, (CMICRODOS PTR [edx]).Files
            add     eax, edx

Done:       ret

MO_Find     endp

;---------------------------------------------------------------------
; MO_Open                                   Return handle by file name
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pName:    Pointer to a 11-byte blank-padded string with the filename
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

MO_Open     proc    pMO:DWORD, pFile:DWORD

            mov     eax, pMO
            lea     eax, (CMICRODOS PTR [eax]).Files                    ; Make EAX point to Files[0]
            mov     ecx, 2

Scan:       push    ecx
            push    esi
            push    edi

            lea     esi, (MO_File PTR [eax]).FileName                   ; Compare file name and extension
            mov     edi, pFile
            mov     ecx, 8+3
            cld
            repe    cmpsb

            pop     edi
            pop     esi
            pop     ecx

            je      Done                                                ; If they match, end scan

            add     eax, sizeof MO_File                                 ; Otherwise advance to next Files[] and loop
            loop    Scan

            mov     eax, pMO
            mov     (CMICRODOS PTR [eax]).dwLastError, ERROR_FILE_NOT_FOUND ; The system cannot find the file specified.
            
            mov     eax, NULL                                           ; Return null

Done:       ret
    
MO_Open     endp

;---------------------------------------------------------------------
; MO_Create                                          Create a new file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pName:    Pointer to a 11-byte blank-padded string with the filename
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

MO_Create   proc    pMO:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_Create   endp

;---------------------------------------------------------------------
; MO_Seek                                            Move file pointer
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwPos:    New pointer position
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Seek     proc    pMO:DWORD, hFile:DWORD, Pos:DWORD

            mov     edx, hFile

            mov     eax, (MO_File PTR [edx]).FileSize                   ; Compare the file size with the requested position
            cmp     eax, Pos
            ja      @F

            mov     eax, ERROR_HANDLE_EOF
            jmp     Done

@@:         mov     eax, Pos                                            ; Update the file pointer
            mov     (MO_File PTR [edx]).FilePointer, eax

            mov     eax, NO_ERROR

Done:       ret
            
MO_Seek     endp

;---------------------------------------------------------------------
; MO_Read                                          Read data from file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; pBuffer:  Pointer to the receiving buffer
; Bytes:    Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Read     proc    pMO:DWORD, hFile:DWORD, pBuffer:DWORD, Bytes:DWORD

            lea     eax, (CMICRODOS PTR [edx]).Files                    ; Determine whether we're reading from file[0] or others
            cmp     eax, hFile
            jne     @F

            mcall   MICRODOS_VTBL, ReadSysFile, pMO, hFile, pBuffer, Bytes
            jmp     Done

@@:         mcall   MICRODOS_VTBL, ReadDatFile, pMO, hFile, pBuffer, Bytes

Done:       ret

MO_Read     endp

;---------------------------------------------------------------------
; MO_Write                                          Write data to file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; pBuffer:  Pointer to the data buffer
; Bytes:    Number of bytes to write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Write    proc    pMO:DWORD, hFile:DWORD, pBuffer:DWORD, Bytes:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_Write    endp

;---------------------------------------------------------------------
; MO_Rename                                                Rename file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pOldName: Current file name
; pNewName: Desired file name 
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Rename   proc    pMO:DWORD, pOldName:DWORD, pNewName:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_Rename   endp

;---------------------------------------------------------------------
; MO_Delete                                                Delete file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pFile:    Name of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_Delete   proc    pMO:DWORD, pFile:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_Delete   endp

;---------------------------------------------------------------------
; MO_GetLastError                  Return the last internal error code
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
;---------------------------------------------------------------------
; EAX:      Last error code
;---------------------------------------------------------------------

MO_GetLastError     proc    pMO:DWORD

            mov     edx, pMO
            mov     eax, (CMICRODOS PTR [edx]).dwLastError
            ret

MO_GetLastError     endp

;---------------------------------------------------------------------
; MO_GetDosVersion                              Return the DOS version
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

MO_GetDosVersion    proc    pMO:DWORD

            mov     eax, 01h
            ret
    
MO_GetDosVersion    endp

;---------------------------------------------------------------------
; MO_SetDosVersion                               Set a new DOS version
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; bVersion: New DOS version to set
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

MO_SetDosVersion    proc    pMO:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_SetDosVersion    endp

;---------------------------------------------------------------------
; MO_GetDiskName                                  Return the disk name
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

MO_GetDiskName      proc    pMO:DWORD, pStr:DWORD, Len:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_GetDiskName      endp

;---------------------------------------------------------------------
; MO_SetDiskName                                     Set the disk name
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pName:    Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_SetDiskName      proc    pMO:DWORD, pName:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_SetDiskName      endp

;---------------------------------------------------------------------
; MO_GetDiskDate                                  Return the disk date
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_GetDiskDate      proc    pMO:DWORD, pStr:DWORD, Len:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_GetDiskDate      endp

;---------------------------------------------------------------------
; MO_SetDiskDate                                     Set the disk date
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pDate:    Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_SetDiskDate      proc    pMO:DWORD, pDate:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_SetDiskDate      endp

;---------------------------------------------------------------------
; MO_GetFilePos                            Return current file pointer
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
;---------------------------------------------------------------------
; EAX:      Current pointer position
;---------------------------------------------------------------------

MO_GetFilePos   proc    pMO:DWORD, hFile:DWORD

            mov     eax, hFile
            mov     eax, (MO_File PTR [eax]).FilePointer
            ret

MO_GetFilePos   endp

;---------------------------------------------------------------------
; MO_GetFileName                                  Return the file name
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_GetFileName  proc    pMO:DWORD, hFile:DWORD, pStr:DWORD, Len:DWORD

            cmp     Len, sizeof MO_File.FileName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, hFile
            lea     edx, (MO_File PTR [edx]).FileName

            fn      MemCopy, edx, pStr, sizeof MO_File.FileName

            mov     eax, NO_ERROR

Done:       ret
    
MO_GetFileName  endp

;---------------------------------------------------------------------
; MO_GetFileExt                              Return the file extension
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; pStr:     Pointer to the receiving string buffer
; Len:      Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_GetFileExt   proc    pMO:DWORD, hFile:DWORD, pStr:DWORD, Len:DWORD

            cmp     Len, sizeof MO_File.Extension
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, hFile
            lea     edx, (MO_File PTR [edx]).Extension

            fn      MemCopy, edx, pStr, sizeof MO_File.Extension

            mov     eax, NO_ERROR

Done:       ret
    
MO_GetFileExt   endp

;---------------------------------------------------------------------
; MO_GetFileSize                                  Return the file size
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

MO_GetFileSize  proc    pMO:DWORD, hFile:DWORD

            mov     eax, hFile
            mov     eax, (MO_File PTR [eax]).FileSize
            ret

MO_GetFileSize  endp

;---------------------------------------------------------------------
; MO_GetFileDate                                  Return the file date
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

MO_GetFileDate  proc    pMO:DWORD, hFile:DWORD

            mov     eax, 00000000h
            ret

MO_GetFileDate  endp

;---------------------------------------------------------------------
; MO_SetFileDate                                  Change the file date
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; Date:     New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_SetFileDate  proc    pMO:DWORD, hFile:DWORD, Date:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_SetFileDate  endp

;---------------------------------------------------------------------
; MO_GetFileAttributes                      Return the file attributes
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

MO_GetFileAttr  proc    pMO:DWORD, hFile:DWORD

            mov     eax, hFile
            mov     eax, (MO_File PTR [eax]).FileAttr
            ret

MO_GetFileAttr  endp

;---------------------------------------------------------------------
; MO_SetFileAttributes                      Change the file attributes
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; Attr:     New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_SetFileAttr  proc    pMO:DWORD, hFile:DWORD, Attr:DWORD

            mov     eax, ERROR_NOT_SUPPORTED
            ret
    
MO_SetFileAttr  endp

;---------------------------------------------------------------------
; MO_InitFiles             Initialize the member class file structures
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_InitFiles    proc    pMO:DWORD

            local   pVDI:DWORD
            local   Buffer[256]:BYTE
            local   FileSize:DWORD

            mov     edx, pMO

            mov     eax, (CMICRODOS PTR [edx]).pVDI
            mov     pVDI, eax

    ; --- Copy the System File template ---

            lea     edx, (CMICRODOS PTR [edx]).Files[0 * sizeof MO_File]

            fn      MemCopy, offset MO_SysFile, edx, sizeof MO_File

    ; --- Copy the Data File template ---

            mov     edx, pMO
            lea     edx, (CMICRODOS PTR [edx]).Files[1 * sizeof MO_File]

            fn      MemCopy, offset MO_DatFile, edx, sizeof MO_File

    ; --- Compute the data file size ---

            mov     edx, pMO
            movzx   ecx, (CMICRODOS PTR [edx]).Sectors                  ; Get the total number of disk sectors minus the number of system sectors
            sub     ecx, 20
            mov     edx, 20                                             ; Skip the system sectors in this counting

            mov     FileSize, 0                                         ; Reset the file size accumulator

@@:         push    ecx
            push    edx

            mcall   MICRODOS_VTBL, Sector2CHS, pMO, dx

            mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, cl, dh, dl

            pop     edx
            pop     ecx

            or      eax, eax
            jnz     Done

            movzx   eax, Buffer[0]                                      ; The first byte of each sector contains the string length
            add     FileSize, eax

            inc     edx                                                 ; Advance to the next sector and loop
            loop    @B

    ; --- Update the data file size ---

            mov     edx, pMO

            mov     eax, FileSize
            mov     (CMICRODOS PTR [edx]).Files[1 * sizeof MO_File].FileSize, eax

            mov     eax, NO_ERROR

Done:       ret
    
MO_InitFiles    endp

;---------------------------------------------------------------------
; MO_ReadSysFile                            Read data from system file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; pBuffer:  Pointer to the receiving buffer
; Bytes:    Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_ReadSysFile  proc    pMO:DWORD, hFile:DWORD, pBuffer:DWORD, Bytes:DWORD

            local   pVDI:DWORD
            local   Buffer[256]:BYTE
            local   Pointer:DWORD
            local   Sector:WORD

            mov     edx, pMO

            mov     eax, (CMICRODOS PTR [edx]).pVDI
            mov     pVDI, eax

    ; --- Compute the starting sector ---

            mov     eax, (MO_File PTR [eax]).FilePointer

            mov     Pointer, eax

            shr     eax, 8

            mov     Sector, ax

    ; --- Read sector ---

Read:       mcall   MICRODOS_VTBL, Sector2CHS, pMO, Sector

            mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, cl, dh, dl

            or      eax, eax
            jnz     Done

    ; --- Copy data from internal buffer ---

            mov     ecx, Pointer
            and     ecx, 000000FFh

            lea     edx, Buffer
            add     edx, ecx

            neg     ecx
            add     ecx, sizeof Buffer

            cmp     ecx, Bytes
            jbe     @F

            mov     ecx, Bytes

@@:         push    ecx

            fn      MemCopy, edx, pBuffer, ecx

            pop     ecx

    ; --- Advance to the next block and loop until all bytes are read ---

            inc     Sector
            add     pBuffer, ecx
            sub     Bytes, ecx
            jnz     Read

            mov     eax, NO_ERROR

Done:       ret
        
MO_ReadSysFile  endp

;---------------------------------------------------------------------
; MO_ReadDatFile                              Read data from data file
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; hFile:    Pointer to a MO_File structure
; pBuffer:  Pointer to the receiving buffer
; Bytes:    Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

MO_ReadDatFile  proc    pMO:DWORD, hFile:DWORD, pBuffer:DWORD, Bytes:DWORD

            local   pVDI:DWORD
            local   Buffer[256]:BYTE
            local   Pointer:DWORD
            local   Count:DWORD
            local   Sector:WORD

            mov     edx, pMO

            mov     eax, (CMICRODOS PTR [edx]).pVDI
            mov     pVDI, eax

            mov     eax, hFile

            mov     eax, (MO_File PTR [eax]).FilePointer
            mov     Pointer, eax

            mov     Count, 0
            mov     Sector, 20

    ; --- Read sector ---

Read:       mcall   MICRODOS_VTBL, Sector2CHS, pMO, Sector

            mcall   VDI_VTBL, Read, pVDI, addr Buffer, sizeof Buffer, cl, dh, dl

            or      eax, eax
            jnz     Done

    ; --- Check the stored string length ---

            movzx   eax, Buffer[0]
            add     Count, eax

            mov     eax, Pointer
            cmp     eax, Count
            jb      Copy

            inc     Sector
            jmp     Read

    ; --- Copy data from internal buffer ---

Copy:       movzx   ecx, Buffer[0]
            sub     ecx, Count
            add     ecx, Pointer

            lea     edx, Buffer
            add     edx, ecx
            inc     edx

            movzx   eax, Buffer[0]
            neg     ecx
            add     ecx, eax

            cmp     ecx, Bytes
            jbe     @F

            mov     ecx, Bytes

@@:         push    ecx

            fn      MemCopy, edx, pBuffer, ecx

            pop     ecx

    ; --- Advance to the next block and loop until all bytes are read ---

            mov     eax, Count
            mov     Pointer, eax

            inc     Sector
            add     pBuffer, ecx
            sub     Bytes, ecx
            jnz     Read

            mov     eax, NO_ERROR

Done:       ret
        
MO_ReadDatFile  endp

;---------------------------------------------------------------------
; MO_Sector2CHS                       Convert relative sector to C/H/S
;---------------------------------------------------------------------
; pMO:      Pointer to the calling MICRODOS object
; Sector:   Relative sector number
;---------------------------------------------------------------------
; CL:       Cylinder
; DH:       Head
; DL:       Sector
;---------------------------------------------------------------------

MO_Sector2CHS   proc    pMO:DWORD, Sector:WORD

            mov     edx, pMO

            mov     al, (CMICRODOS PTR [edx]).SectorsPerTrack           ; SectorsPerTrack * Sides
            mul     (CMICRODOS PTR [edx]).Sides

            mov     cx, ax
            mov     ax, Sector                                          ; Cylinder (AL) = (RelSector / (SectorsPerTrack * Sides)) + FirstTrack
            div     cl                                                  ; AH = Remainder
            add     al, (CMICRODOS PTR [edx]).FirstTrack

            mov     cl, al                                              ; CL = Cylinder
            
            movzx   ax, ah                                              ; Head (AL) = (Remainder / SectorsPerTrack) + FirstSide
            div     (CMICRODOS PTR [edx]).SectorsPerTrack               ; Sector (AH) = Remainder + FirstSector
            add     al, (CMICRODOS PTR [edx]).FirstSide
            add     ah, (CMICRODOS PTR [edx]).FirstSector

            mov     dh, al                                              ; DH = Head
            mov     dl, ah                                              ; DL = Sector

Done:       ret

MO_Sector2CHS   endp

