;---------------------------------------------------------------------
; NEWDOS80/80 Operating System Interface      Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; NEWDOS80 Types of Interface (TI)
;---------------------------------------------------------------------

ND_PDRIVE_TI_A          equ         0000000100000000y                   ; Standard disk interface
ND_PDRIVE_TI_B          equ         0000001000000000y                   ; Omnikron mapper type interface
ND_PDRIVE_TI_C          equ         0000010000000000y                   ; Percom doubler type interface
ND_PDRIVE_TI_D          equ         0000100000000000y                   ; Apparat disk controller type interface
ND_PDRIVE_TI_E          equ         0001000000000000y                   ; LNW type interface
ND_PDRIVE_TI_H          equ         1000000000000000y                   ; Head settle delay required
ND_PDRIVE_TI_I          equ         0010000000000000y                   ; Lowest numbered sector on each track is sector 1
ND_PDRIVE_TI_J          equ         0000000000000001y                   ; Track numbers start from 1
ND_PDRIVE_TI_K          equ         0000000000000100y                   ; Track 0 is in opposite density to that of the other tracks
ND_PDRIVE_TI_L          equ         0000000000001000y                   ; Two step pulses between tracks required
ND_PDRIVE_TI_M          equ         0000000000010000y                   ; Standard NEWDOS80/80 Model III diskette

;---------------------------------------------------------------------
; NEWDOS80 Types of Drives (TD)
;---------------------------------------------------------------------

ND_PDRIVE_TD_5SDSS      equ         0                                   ; "A": 5 inch, single density, single sided drive
ND_PDRIVE_TD_8SDSS      equ         1                                   ; "B": 8 inch, single density, single sided drive
ND_PDRIVE_TD_5SDDS      equ         2                                   ; "C": 5 inch, single density, double sided drive
ND_PDRIVE_TD_8SDDS      equ         3                                   ; "D": 8 inch, single density, double sided drive
ND_PDRIVE_TD_5DDSS      equ         4                                   ; "E": 5 inch, double density, single sided drive
ND_PDRIVE_TD_8DDSS      equ         5                                   ; "F": 8 inch, double density, single sided drive
ND_PDRIVE_TD_5DDDS      equ         6                                   ; "G": 5 inch, double density, double sided drive
ND_PDRIVE_TD_8DDDS      equ         7                                   ; "H": 8 inch, double density, double sided drive

;---------------------------------------------------------------------
; NEWDOS80 PDRIVE flags
;---------------------------------------------------------------------

ND_PDRIVE_Flags_TSR     equ         00000011y                           ; Track Stepping Rate
ND_PDRIVE_Flags_TI      equ         00011100y                           ; Type of Interface
ND_PDRIVE_Flags_M       equ         00100000y                           ; Standard NEWDOS80/80 Model III diskette
ND_PDRIVE_Flags_K       equ         01000000y                           ; Track 0 is in opposite density to that of the other tracks
ND_PDRIVE_Flags_8I      equ         10000000y                           ; 8-inch drives only

ND_PDRIVE_Flags_H       equ         00000001y                           ; Head settle delay required
ND_PDRIVE_Flags_DS      equ         00000010y                           ; Double sided
ND_PDRIVE_Flags_I       equ         00001000y                           ; Lowest numbered sector on each track is sector 1
ND_PDRIVE_Flags_L       equ         00100000y                           ; Two step pulses between tracks required
ND_PDRIVE_Flags_J       equ         01000000y                           ; Track numbers start from 1
ND_PDRIVE_Flags_DD      equ         10000000y                           ; Double density

;---------------------------------------------------------------------
; NEWDOS80 File Primary Directory Entry (FPDE) flags
;---------------------------------------------------------------------

ND_FPDE_Extended        equ         0000000010000000y                   ; 0 indicates FPDE, vice FXDE
ND_FPDE_System          equ         0000000001000000y                   ; 1 if a system file
ND_FPDE_Active          equ         0000000000010000y                   ; 1 indicates FDE allocated to a file
ND_FPDE_Invisible       equ         0000000000001000y                   ; 1 if the file has the invisible attribute
ND_FPDE_Access          equ         0000000000000111y                   ; Access level code
ND_FPDE_Allocation      equ         1000000000000000y                   ; 0 file may allocate more space as needed
ND_FPDE_Release         equ         0100000000000000y                   ; 0 the DOS may deallocate any excess granules above the EOF
ND_FPDE_Modified        equ         0010000000000000y                   ; 1 at least one sector of the file has been written to

;---------------------------------------------------------------------
; NEWDOS80 directory operations
;---------------------------------------------------------------------

ND_DIR_READ             equ         0                                   ; Read directory from disk
ND_DIR_WRITE            equ         1                                   ; Write directory to disk

.data?

;---------------------------------------------------------------------
; NEWDOS80 PDRIVE table
;---------------------------------------------------------------------

ND_PDRIVE   struct
                        BYTE        ?                                   ; Disk Directory Starting Lump
    bLumps              BYTE        ?                                   ; Lumps count
    bFlags_1            BYTE        ?                                   ; Flags
    bTC_                BYTE        ?                                   ; Track Count
    bSPT                BYTE        ?                                   ; Sectors Per Track
    bGPL                BYTE        ?                                   ; Granules Per Lump
                        BYTE        ?                                   ; ?
    bFlags_2            BYTE        ?                                   ; Flags
    bDDSL               BYTE        ?                                   ; Disk Directory Starting Lump
    bDDGA               BYTE        ?                                   ; Disk Directory Granule Allocation
    bSPG                BYTE        ?                                   ; Sectors Per Granule
                        BYTE        ?                                   ; ?
    bTSR                BYTE        ?                                   ; Track Stepping Rate
    wTI                 WORD        ?                                   ; Type of Interface
    bTD                 BYTE        ?                                   ; Type of Drive
ND_PDRIVE   ends

;---------------------------------------------------------------------
; NEWDOS80 Granule Allocation Table (GAT)
;---------------------------------------------------------------------

ND_GAT  struct
    abAllocTable        BYTE        192 dup (?)                         ; Granule free/allocated table
                        BYTE        14  dup (?)                         ; Undefined
    wPassword           WORD        ?                                   ; Disk encoded password
    abDiskName          BYTE        8   dup (?)                         ; Disk name
    abDiskDate          BYTE        8   dup (?)                         ; Disk date
    abCommand           BYTE        32  dup (?)                         ; AUTO command or 0Dh in the first byte if no command at all
ND_GAT  ends

;---------------------------------------------------------------------
; NEWDOS80 File Extent Element
;---------------------------------------------------------------------

ND_Extent   struct
    bLump               BYTE        ?                                   ; 255: End of Extents, 254: Next byte is DEC, 0-253: Lump to area start
    bGranules           BYTE        ?                                   ; Bits 7-5: Granules to the start, Bits 4-0: Contiguous granules less one
ND_Extent   ends

;---------------------------------------------------------------------
; NEWDOS80 File Primary Directory Entry (FPDE)
;---------------------------------------------------------------------

ND_FPDE struct
    wFlags              WORD        ?                                   ; Entry flags and file attributes
                        BYTE        ?                                   ; Undefined 
    bEOF                BYTE        ?                                   ; The lower order byte of the file's EOF
    bLRL                BYTE        ?                                   ; Logical Record Lenght (0 = 256) in bytes
    abFileName          BYTE        8   dup (?)                         ; File name, padded on right with blanks
    abExtension         BYTE        3   dup (?)                         ; File extension, padded on right with blanks
    wUpdatePassword     WORD        ?                                   ; The encode of the update password
    wAccessPassword     WORD        ?                                   ; The encode of the access password
    wNext               WORD        ?                                   ; The middle and high order bytes of the EOF
    Extent              ND_Extent   4   dup (<>)                        ; Four extent elements
    FXDE                ND_Extent   <>                                  ; Link to File Extended Directory Entry
ND_FPDE ends

;---------------------------------------------------------------------
; NEWDOS80 File Extended Directory Entry (FXDE)
;---------------------------------------------------------------------

ND_FXDE struct
    bFlags              BYTE        ?                                   ; Bits 7 and 4 are both 1 indicating an active FXDE
    bDEC                BYTE        ?                                   ; The DEC for previous FXDE or FPDE of this file
                        BYTE        3   dup (?)                         ; Unused
    abFileName          BYTE        8   dup (?)                         ; File name, padded on right with blanks
    abExtension         BYTE        3   dup (?)                         ; File extension, padded on right with blanks
                        BYTE        6   dup (?)                         ; Unused
    Extent              ND_Extent   4   dup (<>)                        ; Four extent elements
    FXDE                ND_Extent   <>                                  ; File Extended Directory Entry
ND_FXDE ends

;---------------------------------------------------------------------
; NEWDOS80 Class
;---------------------------------------------------------------------

CNEWDOS80   struct
    VPTR                DWORD       ?                                   ; Pointer to VTBL
    pVDI                DWORD       ?                                   ; Pointer to Virtual Disk Interface
    pDir                DWORD       ?                                   ; Pointer to Directory
    PDRIVE              ND_PDRIVE   <>                                  ; PDRIVE Table
    wDirSector          WORD        ?                                   ; Relative sector of the disk directory
    bDirSectors         BYTE        ?                                   ; Number of sectors in the disk directory
    bFirstTrack         BYTE        ?                                   ; Number of the first disk track
    bFirstSide          BYTE        ?                                   ; Number of the first disk side
    bSides              BYTE        ?                                   ; Number of disk sides
    dwSeekPointer       DWORD       ?                                   ; Current file pointer set by Seek()
    wSeekSector         WORD        ?                                   ; Current relative sector computed by Seek()
    wLastScan           WORD        ?                                   ; Last Directory Entry Code in a HIT scan
    dwLastError         DWORD       ?                                   ; Code of last error occurred in an internal function
CNEWDOS80   ends

.data

;---------------------------------------------------------------------
; NEWDOS80 Virtual Table
;---------------------------------------------------------------------

NEWDOS80_VTBL   struct                                                  ; Declare the structure and initializers
    Load                DWORD       ND_Load                             ; OSI-derived functions
    Unload              DWORD       ND_Unload
    Find                DWORD       ND_Find
    Open                DWORD       ND_Open
    Create              DWORD       ND_Create
    Seek                DWORD       ND_Seek
    Read                DWORD       ND_Read
    Write               DWORD       ND_Write
    Rename              DWORD       ND_Rename
    Delete              DWORD       ND_Delete
    GetLastError        DWORD       ND_GetLastError
    GetDosVersion       DWORD       ND_GetDosVersion
    SetDosVersion       DWORD       ND_SetDosVersion
    GetDiskName         DWORD       ND_GetDiskName
    SetDiskName         DWORD       ND_SetDiskName
    GetDiskDate         DWORD       ND_GetDiskDate
    SetDiskDate         DWORD       ND_SetDiskDate
    GetFilePos          DWORD       ND_GetFilePos
    GetFileName         DWORD       ND_GetFileName
    GetFileExt          DWORD       ND_GetFileExt
    GetFileSize         DWORD       ND_GetFileSize
    GetFileDate         DWORD       ND_GetFileDate
    SetFileDate         DWORD       ND_SetFileDate
    GetFileAttr         DWORD       ND_GetFileAttr
    SetFileAttr         DWORD       ND_SetFileAttr
    DirRW               DWORD       ND_DirRW                            ; NEWDOS80-specific functions
    GetFDE              DWORD       ND_GetFDE
    GetExtPtr           DWORD       ND_GetExtPtr
    AllocGAT            DWORD       ND_AllocGAT
    FreeGAT             DWORD       ND_FreeGAT
    ScanHIT             DWORD       ND_ScanHIT
    SetHIT              DWORD       ND_SetHIT
    DEC2FDE             DWORD       ND_DEC2FDE
    FDE2DEC             DWORD       ND_FDE2DEC
    Hash                DWORD       ND_Hash
    Sector2CHS          DWORD       ND_Sector2CHS
NEWDOS80_VTBL   ends

NEWDOS80                NEWDOS80_VTBL <>                                ; Define a variable of that type

.code

;---------------------------------------------------------------------
; ND_Load                                         Initialize interface
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Load     proc    pND:DWORD, pVDI:DWORD

            local   abPDRIVE[256]:BYTE
            local   wDirSector:WORD
            local   bDirSectors:BYTE
            local   pDir:DWORD
            local   bTracks:BYTE
            local   bRetry:BYTE

    ; --- Find out and update the first track and side ---

            mov     edx, pND

            mcall   VDI_VTBL, GetFirstTrack, pVDI                       ; Get first track
            mov     (CNEWDOS80 PTR [edx]).bFirstTrack, al

            mcall   VDI_VTBL, GetFirstSide, pVDI                        ; Get first side
            mov     (CNEWDOS80 PTR [edx]).bFirstSide, al

            mcall   VDI_VTBL, GetTracks, pVDI                           ; Get number of tracks
            mov     bTracks, al
            
            mov     bRetry, 0                                           ; Reset the retry counter
            
    ; --- Load the PDRIVE table from the 3rd disk sector ---

Read:       mov     edx, pND

            mcall   VDI_VTBL, Read, pVDI, addr abPDRIVE, sizeof abPDRIVE, (CNEWDOS80 PTR [edx]).bFirstTrack, (CNEWDOS80 PTR [edx]).bFirstSide, 2
            
            or      eax, eax
            jnz     Done

    ; --- Check whether the PDRIVE table is valid ---

            lea     edx, abPDRIVE

            cmp     (ND_PDRIVE PTR [edx]).bTD, 7                        ; Type of Drive must be between 0 (A) and 7 (H)
            ja      Retry

            cmp     (ND_PDRIVE PTR [edx]).bGPL, 2                       ; Granules Per Lump must be between 2 and 8
            jb      Retry
 
            cmp     (ND_PDRIVE PTR [edx]).bGPL, 8
            ja      Retry

            mov     al, (ND_PDRIVE PTR [edx]).bTC_                      ; Get Track Count

            test    (ND_PDRIVE PTR [edx]).wTI, ND_PDRIVE_TI_K           ; Check if track 0 is being ignored
            jz      @F
            
            inc     al                                                  ; Adjust track count for comparison
            
@@:         cmp     al, 35                                              ; Track Count must be at least 35
            jae     @F

    ; --- Try looking for the PDRIVE table in Track #1 in mixed density disks ---

Retry:      cmp     bRetry, 0                                           ; Check if haven't done tried this before
            jne     Error

            inc     bRetry                                              ; Register this second attempt
            
            mcall   VDI_VTBL, GetDensity, pVDI                          ; Get disk density
            
            cmp     al, 2                                               ; Check whether it is a mixed density disk
            jne     Error
            
            mov     edx, pND
            inc     (CNEWDOS80 PTR [edx]).bFirstTrack                   ; Try reading the PDRIVE table from track 1
            jmp     Read

    ; --- Compute granule units ---

@@:         cmp     (ND_PDRIVE PTR [edx]).bSPG, 0                       ; Check if Sectors per Granule is defined
            jne     @F

            mov     al, (ND_PDRIVE PTR [edx]).bSPT                      ; SPG = (SPT * Tracks) / (Lumps / GPL)
            mul     (ND_PDRIVE PTR [edx]).bTC_
            div     (ND_PDRIVE PTR [edx]).bLumps
            cbw
            div     (ND_PDRIVE PTR [edx]).bGPL
            mov     (ND_PDRIVE PTR [edx]).bSPG, al
            jmp     Copy
            
@@:         movzx   ax, (ND_PDRIVE PTR [edx]).bDDGA                     ; Convert DDSA into DDGA
            div     (ND_PDRIVE PTR [edx]).bSPG
            mov     (ND_PDRIVE PTR [edx]).bDDGA, al

            or      ah, ah                                              ; There must be no remainder after last division
            jnz     Error

    ; --- Copy the first PDRIVE record ---

Copy:       mov     edx, pND

            fn      MemCopy, addr abPDRIVE, addr (CNEWDOS80 PTR [edx]).PDRIVE, sizeof ND_PDRIVE

    ; --- Calculate the directory sectors ---

            lea     edx, abPDRIVE

            mov     al, (ND_PDRIVE PTR [edx]).bSPG                      ; Sector = Directory Starting Lump * (Granules Per Lump * Sectors Per Granule)
            mul     (ND_PDRIVE PTR [edx]).bGPL
            mul     (ND_PDRIVE PTR [edx]).bDDSL

            mov     wDirSector, ax

            movzx   eax, (ND_PDRIVE PTR [edx]).bDDGA                    ; Sectors = Directory Granules * Sectors Per Granule * 256 bytes per sector
            mul     (ND_PDRIVE PTR [edx]).bSPG
            
            mov     bDirSectors, al
            
    ; --- Allocate memory for the directory ---

            shl     eax, 8

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pDir, eax                                           ; Save pointer

    ; --- Update member class variables ---
    
            mov     edx, pND
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (CNEWDOS80 PTR [edx]).pVDI, eax

            mov     eax, pDir                                           ; Update pointer to directory buffer
            mov     (CNEWDOS80 PTR [edx]).pDir, eax

            mov     ax, wDirSector                                      ; Update directory's relative sector number
            mov     (CNEWDOS80 PTR [edx]).wDirSector, ax

            mov     al, bDirSectors                                     ; Update number of directory sectors
            mov     (CNEWDOS80 PTR [edx]).bDirSectors, al

    ; --- Load directory into memory ---

            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_READ
            
            or      eax, eax
            jnz     Done

    ; --- Compute and validate the disk's granule units ---

            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done

Error:      mov     eax, ERROR_NOT_DOS_DISK   

Done:       ret

ND_Load     endp

;---------------------------------------------------------------------
; ND_Unload                                          Release interface
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Unload   proc    pND:DWORD

            mov     eax, pND

            fn      GlobalFree, (CNEWDOS80 PTR [eax]).pDir

            or      eax, eax
            jz      Done

            fn      GetLastError

Done:       ret

ND_Unload   endp

;---------------------------------------------------------------------
; ND_Find                                Search files in the directory
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; bFlag:    OSI_FINDFIRST or OSI_FINDNEXT
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

ND_Find     proc    pND:DWORD, bFlag:BYTE


            local   bMode:BYTE

            mov     bMode, 0
            
            cmp     bFlag, OSI_FINDFIRST
            je      Find

            mov     bMode, 1

Find:       mcall   NEWDOS80_VTBL, ScanHIT, pND, bMode, 0                   ; Scan the first/next non-empty slot

            cmp     eax, -1
            jne     @F
            
            mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_NO_MORE_FILES  ; There are no more files.
            mov     eax, NULL
            jmp     Done

@@:         mcall   NEWDOS80_VTBL, DEC2FDE, pND, al                         ; Get pointer to directory entry based on DEC

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_INVALID_ACCESS ; The access code is invalid.
            mov     eax, NULL
            jmp     Done

@@:         test    (ND_FPDE PTR [eax]).wFlags, ND_FPDE_Extended            ; Check whether the entry is primary or extended
            jnz     Find

            test    (ND_FPDE PTR [eax]).wFlags, ND_FPDE_Active              ; Check if the FPDE is active
            jz      Find

Done:       ret
    
ND_Find     endp

;---------------------------------------------------------------------
; ND_Open                                        Find an existing file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pName:    Pointer to a 11-byte blank-padded string with the filename
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

ND_Open     proc    pND:DWORD, pName:DWORD

            local   bMode:BYTE
            local   bHash:BYTE

            mov     bMode, 0
            
            mcall   NEWDOS80_VTBL, Hash, pND, pName                         ; Hash the file name
            mov     bHash, al

Find:       mcall   NEWDOS80_VTBL, ScanHIT, pND, bMode, bHash               ; Scan the first/next non-empty slot

            cmp     eax, -1
            jne     @F
            
            mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_FILE_NOT_FOUND ; The system cannot find the file specified.
            mov     eax, NULL
            jmp     Done

@@:         mov     bMode, 1                                                ; Set to "Find Next" on next scan
    
            mcall   NEWDOS80_VTBL, DEC2FDE, pND, al                         ; Convert DEC into pointer to FPDE

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_INVALID_ACCESS ; The access code is invalid.
            mov     eax, NULL
            jmp     Done

@@:         test    (ND_FPDE PTR [eax]).wFlags, ND_FPDE_Extended            ; Check whether the entry is primary or extended
            jnz     Find

            test    (ND_FPDE PTR [eax]).wFlags, ND_FPDE_Active              ; Check if the FPDE is active
            jz      Find

            push    esi
            push    edi
            
            lea     esi, (ND_FPDE PTR [eax]).abFileName
            mov     edi, pName
            mov     ecx, sizeof ND_FPDE.abFileName + sizeof ND_FPDE.abExtension
            cld
            repe    cmpsb                                                   ; Compare the two file names
            
            pop     edi
            pop     esi
            jne     Find

Done:       ret
    
ND_Open     endp

;---------------------------------------------------------------------
; ND_Create                                          Create a new file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pName:    Pointer to a 11-byte blank-padded string with the filename
; dwSize:   File size
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

ND_Create   proc    pND:DWORD, pName:DWORD, dwSize:DWORD

            local   pFPDE:DWORD
            local   pFDE:DWORD
            local   pLastFDE:DWORD
            local   bGranules:BYTE
            local   bExtCount:BYTE
            local   wExtent:WORD
            local   bAllocated:BYTE

            mov     bExtCount, 0

    ; --- Find an available directory entry ---

            mcall   NEWDOS80_VTBL, GetFDE, pND

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
            jmp     Error

@@:         mov     pFPDE, eax
            
    ; --- Fill out the entry ---

            mov     (ND_FPDE PTR [eax]).wFlags, ND_FPDE_Active              ; Mark entry as active

            mov     (ND_FPDE PTR [eax]).wUpdatePassword, 4296h              ; Update password is blanks
            mov     (ND_FPDE PTR [eax]).wAccessPassword, 4296h              ; Access password is blanks

            mov     edx, dwSize                                             ; Set End Of File (EOF)
            mov     (ND_FPDE PTR [eax]).bEOF, dl
            
            shr     edx, 8                                                  ; Set Ending Record Number (ERN)
            mov     (ND_FPDE PTR [eax]).wNext, dx

            cmp     (ND_FPDE PTR [eax]).bEOF, 0                             ; If EOF is not zero increment ERN
            je      @F
            
            inc     (ND_FPDE PTR [eax]).wNext
            
@@:         fn      MemCopy, pName, addr (ND_FPDE PTR [eax]).abFileName, sizeof ND_FPDE.abFileName + sizeof ND_FPDE.abExtension

    ; --- Update the Hash Index Table ---
    
            mcall   NEWDOS80_VTBL, Hash, pND, pName                         ; Hash the file name

            mcall   NEWDOS80_VTBL, SetHIT, pND, pFPDE, al                   ; Update HIT slot with the new hash code
            
    ; --- Compute the required number of granules ---
    
            mov     edx, pND
            lea     edx, (CNEWDOS80 PTR [edx]).PDRIVE
            
            mov     eax, pFPDE
            mov     pFDE, eax
            
            mov     ax, (ND_FPDE PTR [eax]).wNext                           ; Granules = Sectors / SectorsPerGranule
            div     (ND_PDRIVE PTR [edx]).bSPG
            or      ah, ah
            jz      @F
            inc     al
@@:         mov     bGranules, al

    ; --- Create an extent with as much granules as possible ---
    
NextExt:    mcall   NEWDOS80_VTBL, AllocGAT, pND, bGranules

            or      eax, eax                                                ; Check against null pointer
            jz      Done

            mov     wExtent, ax
            mov     bAllocated, cl

    ; --- Save the new extent in the extents table ---

            inc     bExtCount

            mcall   NEWDOS80_VTBL, GetExtPtr, pND, pFDE, bExtCount, 1       ; Get pointer to next extent

            or      eax, eax                                                ; Check if null
            jz      AllocFDE

            cmp     eax, -1                                                 ; Check for error
            jne     @F

            mov     eax, ERROR_INVALID_TABLE                                ; Invalid or unknown table specified.
            jmp     Error

@@:         push    wExtent                                                 ; Save extent
            pop     word ptr [eax]

    ; --- Check if all required granules have been allocated ---
    
            mov     al, bAllocated
            sub     bGranules, al
            jnz     NextExt

    ; --- Save Directory ---
    
            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE

            or      eax, eax
            jnz     Error

            mov     eax, pFPDE                                              ; Return pointer to FPDE
            jmp     Done

    ; --- Obtain a new directory entry ---

AllocFDE:   mov     eax, pFDE                                               ; Save current pointer
            mov     pLastFDE, eax

            mcall   NEWDOS80_VTBL, GetFDE, pND                              ; Get a new pointer

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
            jmp     Error

@@:         mov     pFDE, eax                                               ; Save new pointer

    ; --- Set a forward link to the next entry ---

            mcall   NEWDOS80_VTBL, FDE2DEC, pND, pFDE                       ; Get its corresponding DEC

            mov     edx, pLastFDE
            
            mov     (ND_FPDE PTR [edx]).FXDE.bLump, 0FEh                    ; 0xFE indicates a link exists
            mov     (ND_FPDE PTR [edx]).FXDE.bGranules, al                  ; This value is the DEC of the next entry
            
    ; --- Set a backward link to the previous entry ---

            mcall   NEWDOS80_VTBL, FDE2DEC, pND, pLastFDE                   ; Get DEC of previous entry

            mov     edx, pFDE

            mov     (ND_FXDE PTR [edx]).bFlags, ND_FPDE_Active+ND_FPDE_Extended ; Mark the new entry as active and extended
            mov     (ND_FXDE PTR [edx]).bDEC, al                                ; In a FXDE this byte contains the backward link

    ; --- Copy the filename and set the HIT slot ---
    
            fn      MemCopy, pName, addr (ND_FXDE PTR [edx]).abFileName, sizeof ND_FXDE.abFileName + sizeof ND_FXDE.abExtension

            mcall   NEWDOS80_VTBL, Hash, pND, pName                         ; Hash the file name

            mcall   NEWDOS80_VTBL, SetHIT, pND, pFDE, al                    ; Update HIT slot with the hash code
            
            jmp     NextExt

Error:      push    eax
            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_READ                  ; Restore previous directory state
            pop     eax

            mov     edx, pND
            mov     (CNEWDOS80 PTR [edx]).dwLastError, eax                  ; Set last error
            
            mov     eax, NULL                                               ; Return null pointer

Done:       ret

ND_Create   endp

;---------------------------------------------------------------------
; ND_Seek               Convert a file position into a relative sector
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Handle to File Primary Directory Entry
; dwPos:    New pointer position
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Seek     proc    pND:DWORD, pFPDE:DWORD, dwPos:DWORD

            local   bSPG:BYTE
            local   bExtent:BYTE
            local   wSector:WORD

    ; --- Initialize variables ---

            mov     eax, pND
            lea     eax, (CNEWDOS80 PTR [eax]).PDRIVE
            mov     al, (ND_PDRIVE PTR [eax]).bSPG                      ; Copy SectorsPerGranule to a local var
            mov     bSPG, al

            mov     eax, dwPos                                          ; Sector = Pos / 256 bytes per sector
            shr     eax, 8
            mov     wSector, ax

    ; --- Find in which extent the requested sector is ---

            mov     bExtent, 1
            
Scan:       mcall   NEWDOS80_VTBL, GetExtPtr, pND, pFPDE, bExtent, 0    ; Get pointer to extent

            or      eax, eax                                            ; Check if EOF
            jnz     @F
            
            mov     eax, ERROR_HANDLE_EOF                               ; Reached the end of the file.
            mov     wSector, -1
            jmp     Update

@@:         cmp     eax, -1                                             ; Check for error
            jne     @F

            mov     eax, ERROR_FILE_CORRUPT                             ; The file or directory is corrupted and unreadable.
            mov     wSector, -1
            jmp     Update

@@:         mov     ax, word ptr [eax]                                  ; Get the extent
   
            and     ah, 00011111y                                       ; Compute the number of sectors contained in this extent
            inc     ah
            shr     ax, 8
            mul     bSPG

            cmp     ax, wSector                                         ; Check if the requested sector is in the current extent
            ja      Found

            sub     wSector, ax                                         ; Subtract from sectors required in the next extent(s) and repeat
            inc     bExtent
            jmp     Scan

    ; --- Convert the extent information in a relative sector ---

Found:      mov     ax, word ptr [edx]                                  ; AL = Lump

            and     ah, 11100000y                                       ; AH = Initial granule from the start of the lump
            shr     ah, 5
            movzx   cx, ah

            mov     edx, pND
            lea     edx, (CNEWDOS80 PTR [edx]).PDRIVE

            mul     (ND_PDRIVE PTR [edx]).bGPL                          ; Granule = Lump * GPL + initial granule
            add     ax, cx

            movzx   cx, (ND_PDRIVE PTR [edx]).bSPG                      ; Sector = Granule * SPG
            mul     cx

            add     wSector, ax                                         ; Add to the remaining sector offset

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

    ; --- Update member variables ---

Update:     mov     edx, pND

            push    dwPos
            pop     (CNEWDOS80 PTR [edx]).dwSeekPointer                 ; Update current file pointer

            push    wSector
            pop     (CNEWDOS80 PTR [edx]).wSeekSector                   ; Update current relative sector

Done:       ret
            
ND_Seek     endp

;---------------------------------------------------------------------
; ND_Read                                          Read data from file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to the receiving buffer
; dwBytes:  Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Read     proc    pND:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            local   pVDI:DWORD
            local   abBuffer[256]:BYTE
            local   dwSeekPointer:DWORD
            local   dwFileSize:DWORD

    ; --- Initialize local variables ---
    
            mov     edx, pND

            mov     eax, (CNEWDOS80 PTR [edx]).pVDI                     ; Copy pVDI to local var
            mov     pVDI, eax

            mov     eax, (CNEWDOS80 PTR [edx]).dwSeekPointer            ; Copy SeekPointer to local var
            mov     dwSeekPointer, eax

            mcall   NEWDOS80_VTBL, GetFileSize, pND, pFPDE              ; Get the file size
            mov     dwFileSize, eax

    ; --- Check if last seek was successful ---

Next:       mov     edx, pND

            cmp     (CNEWDOS80 PTR [edx]).wSeekSector, -1
            jne     @F

            mov     eax, ERROR_INVALID_ADDRESS                          ; Attempt to access invalid address.
            jmp     Done

    ; --- Read the sector pointed by SeekSector ---

@@:         mcall   NEWDOS80_VTBL, Sector2CHS, pND, (CNEWDOS80 PTR [edx]).wSeekSector   ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Read, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl    ; Read sector

            or      eax, eax
            jnz     Done

    ; --- Determine the starting transfer point ---

            mov     eax, dwSeekPointer                                  ; Compute SeekPointer % 256
            and     eax, 0FFh                                           

            lea     edx, abBuffer                                       ; Advance over internal buffer pointer 
            add     edx, eax

    ; --- Determine the ending transfer point ---

            mov     ecx, dwFileSize                                     ; Compute how far SeekPointer is from the end of file
            sub     ecx, dwSeekPointer
            ja      @F
            
            mov     eax, ERROR_READ_FAULT                               ; The system cannot read from the specified device.
            jmp     Done

@@:         cmp     ecx, 256                                            ; Check if SeekPointer is within the last sector
            jae     @F
            
            mov     ecx, dwFileSize                                     ; Set the ending point to FileSize % 256
            and     ecx, 0FFh
            jmp     J1

@@:         mov     ecx, 256                                            ; Set the ending point to the end of sector

    ; --- Determine the transfer length ---

J1:         sub     ecx, eax

            cmp     ecx, dwBytes                                        ; If it is more than the remaining requested bytes to read
            jbe     @F

            mov     ecx, dwBytes                                        ; Reduce it to the number of remaining requested bytes to read

    ; --- Transfer data from the internal buffer to the user buffer ---

@@:         push    ecx
            fn      MemCopy, edx, pBuffer, ecx
            pop     ecx

    ; --- Update pointers and counters ---

            sub     dwBytes, ecx                                        ; Compute remaining bytes to read
            add     pBuffer, ecx                                        ; Advance buffer pointer
            add     dwSeekPointer, ecx                                  ; Advance file pointer

            mcall   NEWDOS80_VTBL, Seek, pND, pFPDE, dwSeekPointer      ; Move the file pointer to the most current position

            cmp     dwBytes, 0                                          ; Exit when read it all
            jnz     Next
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

ND_Read     endp

;---------------------------------------------------------------------
; ND_Write                                          Write data to file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to the data buffer
; dwBytes:  Number of bytes to write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Write    proc    pND:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            local   pVDI:DWORD
            local   abBuffer[256]:BYTE
            local   dwSeekPointer:DWORD
            local   dwFileSize:DWORD
            local   dwPtr:DWORD
            local   dwLen:DWORD

    ; --- Initialize local variables ---
    
            mov     edx, pND

            mov     eax, (CNEWDOS80 PTR [edx]).pVDI                     ; Copy pVDI to local var
            mov     pVDI, eax

            mov     eax, (CNEWDOS80 PTR [edx]).dwSeekPointer            ; Copy SeekPointer to local var
            mov     dwSeekPointer, eax

            mcall   NEWDOS80_VTBL, GetFileSize, pND, pFPDE              ; Get the file size
            mov     dwFileSize, eax

    ; --- Check if last seek was successful ---

Next:       mov     edx, pND

            cmp     (CNEWDOS80 PTR [edx]).wSeekSector, -1
            jne     @F

            mov     eax, ERROR_INVALID_ADDRESS                          ; Attempt to access invalid address.
            jmp     Done

    ; --- Determine the starting transfer point ---

@@:         mov     eax, dwSeekPointer                                  ; Compute SeekPointer % 256
            and     eax, 0FFh                                           

            lea     edx, abBuffer                                       ; Advance over internal buffer pointer 
            add     edx, eax
            
            mov     dwPtr, edx                                          ; Save starting point

    ; --- Determine the ending transfer point ---

            mov     ecx, dwFileSize                                     ; Compute how far SeekPointer is from the end of file
            sub     ecx, dwSeekPointer
            ja      @F
            
            mov     eax, ERROR_WRITE_FAULT                              ; The system cannot write to the specified device.
            jmp     Done

@@:         cmp     ecx, 256                                            ; Check if SeekPointer is within the last sector
            jae     @F
            
            mov     ecx, dwFileSize                                     ; Set the ending point to FileSize % 256
            and     ecx, 0FFh
            jmp     J1

@@:         mov     ecx, 256                                            ; Set the ending point to the end of sector

    ; --- Determine the transfer length ---

J1:         sub     ecx, eax

            cmp     ecx, dwBytes                                        ; If it is more than the remaining requested bytes to read
            jbe     @F

            mov     ecx, dwBytes                                        ; Reduce it to the number of remaining requested bytes to read

@@:         mov     dwLen, ecx                                          ; Save transfer length
            
    ; --- If needed, read the sector before writing to it ---

            cmp     ecx, 256                                                            ; Check if transfer length is of one entire 256-byte sector
            je      @F

            mov     edx, pND

            mcall   NEWDOS80_VTBL, Sector2CHS, pND, (CNEWDOS80 PTR [edx]).wSeekSector   ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Read, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl    ; Read sector

            or      eax, eax
            jnz     Done

    ; --- Transfer data from the user buffer to the internal buffer ---

@@:         fn      MemCopy, pBuffer, dwPtr, dwLen

    ; --- Write sector ---

            mov     edx, pND

            mcall   NEWDOS80_VTBL, Sector2CHS, pND, (CNEWDOS80 PTR [edx]).wSeekSector   ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Write, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl   ; Write sector

            or      eax, eax
            jnz     Done

    ; --- Update pointers and counters ---

            mov     ecx, dwLen
            
            sub     dwBytes, ecx                                        ; Compute remaining bytes to write
            add     pBuffer, ecx                                        ; Advance buffer pointer
            add     dwSeekPointer, ecx                                  ; Advance file pointer

            mcall   NEWDOS80_VTBL, Seek, pND, pFPDE, dwSeekPointer      ; Move the file pointer to the most current position

            cmp     dwBytes, 0                                          ; Exit when done
            jnz     Next

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

ND_Write    endp

;---------------------------------------------------------------------
; ND_Rename                                                Rename file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the FPDE of the file to rename
; pName:    Pointer to a 11-byte blank-padded string with the new name
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Rename   proc    pND:DWORD, pFPDE:DWORD, pName:DWORD

    ; --- Replace the old name with the new one ---
   
            mov     eax, pFPDE
            
            fn      MemCopy, pName, addr (ND_FPDE PTR [eax]).abFileName, sizeof ND_FPDE.abFileName + sizeof ND_FPDE.abExtension

    ; --- Update the Directory Entry Code ---
        
            mcall   NEWDOS80_VTBL, Hash, pND, pName                         ; Hash the file name

            mcall   NEWDOS80_VTBL, SetHIT, pND, pFPDE, al                   ; Set HIT slot with the new hash code

    ; --- Save Directory ---
    
            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE

            ret
    
ND_Rename   endp

;---------------------------------------------------------------------
; ND_Delete                                                Delete file
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the FPDE of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_Delete   proc    pND:DWORD, pFPDE:DWORD

            local   bExtent:BYTE
            
    ; --- Inactivate directory entry ---
    
            mov     edx, pFPDE

            and     byte ptr (ND_FPDE PTR [edx]), not ND_FPDE_Active        ; Reset 4th bit of FPDE Attributes[0]

    ; --- Release corresponding HIT slot ---

            mcall   NEWDOS80_VTBL, SetHIT, pND, pFPDE, 0                    ; Replace current hash code by zero

    ; --- Loop through all extents releasing allocated granules ---

            mov     bExtent, 1

Del:        mcall   NEWDOS80_VTBL, GetExtPtr, pND, pFPDE, bExtent, 0        ; Get pointer to extent

            or      eax, eax                                                ; Check if EOF
            jz      Commit

            cmp     eax, -1                                                 ; Check for error
            jne     @F

            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_READ                  ; Restore previous directory state

            mov     eax, ERROR_FILE_CORRUPT                                 ; The file or directory is corrupted and unreadable.
            jmp     Done

@@:         mcall   NEWDOS80_VTBL, FreeGAT, pND, word ptr [eax]             ; Release granules allocated in this extent

            or      eax, eax                                                ; Check if any errors occurred
            jnz     Commit

            inc     bExtent                                                 ; Advance to next extent and repeat
            jmp     Del

    ; --- Save Directory ---
    
Commit:     mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE                 ; Save directory

Done:       ret

ND_Delete   endp

;---------------------------------------------------------------------
; ND_GetLastError                  Return the last internal error code
;---------------------------------------------------------------------
; pND:      Pointer to the calling OSI object
;---------------------------------------------------------------------
; EAX:      Last error code
;---------------------------------------------------------------------

ND_GetLastError proc    pND:DWORD

            mov     eax, pND
            mov     eax, (CNEWDOS80 PTR [eax]).dwLastError
            ret

ND_GetLastError endp

;---------------------------------------------------------------------
; ND_GetDosVersion                              Return the DOS version
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

ND_GetDosVersion    proc    pND:DWORD

            mov     eax, 0
            ret
    
ND_GetDosVersion    endp

;---------------------------------------------------------------------
; ND_SetDosVersion                               Set a new DOS version
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; bVersion: New DOS version to set
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

ND_SetDosVersion    proc    pOSI:DWORD, bVersion:BYTE

            mov     eax, ERROR_NOT_SUPPORTED                            ; The request is not supported.
            ret
            
ND_SetDosVersion    endp

;---------------------------------------------------------------------
; ND_GetDiskName                                  Return the disk name
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

ND_GetDiskName  proc    pND:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof ND_GAT.abDiskName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pND
            mov     edx, (CNEWDOS80 PTR [edx]).pDir
            lea     edx, (ND_GAT PTR [edx]).abDiskName

            fn      MemCopy, edx, pStr, sizeof ND_GAT.abDiskName

            mov     eax, NO_ERROR

Done:       ret
    
ND_GetDiskName  endp

;---------------------------------------------------------------------
; ND_SetDiskName                                     Set the disk name
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pName:    Pointer to null-terminated string
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_SetDiskName  proc    pND:DWORD, pName:DWORD

            mov     eax, pND
            mov     eax, (CNEWDOS80 PTR [eax]).pDir
            lea     eax, (ND_GAT PTR [eax]).abDiskName

            fn      MemCopy, pName, eax, sizeof ND_GAT.abDiskName

            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE             ; Commit changes in the Directory structure

            ret

ND_SetDiskName  endp

;---------------------------------------------------------------------
; ND_GetDiskDate                                  Return the disk date
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_GetDiskDate  proc    pND:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof ND_GAT.abDiskDate
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pND
            mov     edx, (CNEWDOS80 PTR [edx]).pDir
            lea     edx, (ND_GAT PTR [edx]).abDiskDate

            fn      MemCopy, edx, pStr, sizeof ND_GAT.abDiskDate

            mov     eax, NO_ERROR

Done:       ret
    
ND_GetDiskDate  endp

;---------------------------------------------------------------------
; ND_SetDiskDate                                     Set the disk date
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pDate:    Pointer to a 8-byte blank-padded string with the new date
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_SetDiskDate  proc    pND:DWORD, pDate:DWORD

            mov     eax, pND
            mov     eax, (CNEWDOS80 PTR [eax]).pDir
            lea     eax, (ND_GAT PTR [eax]).abDiskDate

            fn      MemCopy, pDate, eax, sizeof ND_GAT.abDiskDate

            mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE             ; Commit changes in the Directory structure

            ret
    
ND_SetDiskDate  endp

;---------------------------------------------------------------------
; ND_GetFilePos                            Return current file pointer
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Handle to File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Current pointer position or -1 for an invalid position
;---------------------------------------------------------------------

ND_GetFilePos   proc    pND:DWORD, pFPDE:DWORD

            mov     eax, pND
            mov     eax, (CNEWDOS80 PTR [eax]).dwSeekPointer
            ret
    
ND_GetFilePos   endp

;---------------------------------------------------------------------
; ND_GetFileName                                  Return the file name
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_GetFileName  proc    pND:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof ND_FPDE.abFileName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pFPDE
            lea     edx, (ND_FPDE PTR [edx]).abFileName

            fn      MemCopy, edx, pStr, sizeof ND_FPDE.abFileName

            mov     eax, NO_ERROR

Done:       ret
    
ND_GetFileName  endp

;---------------------------------------------------------------------
; ND_GetFileExt                              Return the file extension
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_GetFileExt   proc    pND:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof ND_FPDE.abExtension
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER
            jmp     Done

@@:         mov     edx, pFPDE
            lea     edx, (ND_FPDE PTR [edx]).abExtension

            fn      MemCopy, edx, pStr, sizeof ND_FPDE.abExtension

            mov     eax, NO_ERROR

Done:       ret
    
ND_GetFileExt   endp

;---------------------------------------------------------------------
; ND_GetFileSize                                  Return the file size
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

ND_GetFileSize  proc    pND:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (ND_FPDE PTR [edx]).wNext                    ; Number of file sectors
            shl     eax, 8

            movzx   ecx, (ND_FPDE PTR [edx]).bEOF                     ; Number of bytes within the last sector
            neg     cl
            sub     eax, ecx

            ret
            
ND_GetFileSize  endp

;---------------------------------------------------------------------
; ND_GetFileDate                                  Return the file date
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

ND_GetFileDate  proc    pND:DWORD, pFPDE:DWORD

            mov     eax, 0                                              ; NEWDOS80 doesn't support file dates
            ret

ND_GetFileDate  endp

;---------------------------------------------------------------------
; ND_SetFileDate                                  Change the file date
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwDate:   New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_SetFileDate  proc    pND:DWORD, pFPDE:DWORD, dwDate:DWORD

            mov     eax, ERROR_NOT_SUPPORTED                            ; The request is not supported.
            ret
    
ND_SetFileDate  endp

;---------------------------------------------------------------------
; ND_GetFileAttr                            Return the file attributes
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

ND_GetFileAttr  proc    pND:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (ND_FPDE PTR [edx]).wFlags                     ; Get the file access level
            and     ax, ND_FPDE_Access

            test    (ND_FPDE PTR [edx]).wFlags, ND_FPDE_System          ; Test the System attribute
            jz      @F

            bts     eax, 4
            
@@:         test    (ND_FPDE PTR [edx]).wFlags, ND_FPDE_Invisible       ; Test the Invisible attribute
            jz      @F

            bts     eax, 8
            
@@:         test    (ND_FPDE PTR [edx]).wFlags, ND_FPDE_Modified        ; Test the Modified attribute
            jz      Done

            bts     eax, 12
            
Done:       ret
    
ND_GetFileAttr  endp

;---------------------------------------------------------------------
; ND_SetFileAttr                            Change the file attributes
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwAttr:   New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_SetFileAttr  proc    pND:DWORD, pFPDE:DWORD, dwAttr:DWORD

            mov     edx, pFPDE
            mov     eax, dwAttr

            and     (ND_FPDE PTR [edx]).wFlags, not (ND_FPDE_Modified+ND_FPDE_Access+ND_FPDE_System+ND_FPDE_Invisible)

            and     ax, ND_FPDE_Access                                  ; Set the new access level
            or      (ND_FPDE PTR [edx]).wFlags, ax

            test    dwAttr, 000000F0h                                   ; Check whether to set the System attribute
            jz      @F
            
            or      (ND_FPDE PTR [edx]).wFlags, ND_FPDE_System

@@:         test    dwAttr, 00000F00h                                   ; Check whether to set the Invisible attribute
            jz      @F
            
            or      (ND_FPDE PTR [edx]).wFlags, ND_FPDE_Invisible

@@:         test    dwAttr, 0000F000h                                   ; Check whether to set the Modified attribute
            jz      @F
            
            or      (ND_FPDE PTR [edx]).wFlags, ND_FPDE_Modified

@@:         mcall   NEWDOS80_VTBL, DirRW, pND, ND_DIR_WRITE             ; Commit changes in the Directory structure
    
Done:       ret
    
ND_SetFileAttr  endp

;---------------------------------------------------------------------
; ND_DirRW                          Read or Write the entire directory
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; bMode:    0:Read, 1:Write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------      

ND_DirRW    proc    pND:DWORD, bMode:BYTE

            local   pVDI:DWORD
            local   pBuffer:DWORD
            local   wDirSector:WORD
            local   bDirSectors:BYTE

            mov     edx, pND

            mov     eax, (CNEWDOS80 PTR [edx]).pVDI
            mov     pVDI, eax
            
            mov     eax, (CNEWDOS80 PTR [edx]).pDir
            mov     pBuffer, eax

            mov     ax, (CNEWDOS80 PTR [edx]).wDirSector
            mov     wDirSector, ax

            mov     al, (CNEWDOS80 PTR [edx]).bDirSectors
            mov     bDirSectors, al

Next:       mcall   NEWDOS80_VTBL, Sector2CHS, pND, wDirSector          ; Convert relative setor to C/H/S

            cmp     bMode, ND_DIR_READ
            jne     DirWrite
            
DirRead:    mcall   VDI_VTBL, Read, pVDI, pBuffer, 256, cl, dh, dl      ; Read sector
            jmp     @F

DirWrite:   mcall   VDI_VTBL, Write, pVDI, pBuffer, 256, cl, dh, dl     ; Write sector

@@:         or      eax, eax
            jnz     Done

            add     pBuffer, 256                                        ; Advance pointer to the next memory block

            inc     wDirSector                                          ; Advance to next sector
            dec     bDirSectors                                         ; Decrease number of sectors to read
            jnz     Next                                                ; Continue while not zero
                        
            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            
Done:       ret

ND_DirRW    endp

;---------------------------------------------------------------------
; ND_GetFDE      Return a pointer to an available file directory entry
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

ND_GetFDE   proc    pND:DWORD

            mcall   NEWDOS80_VTBL, ScanHIT, pND, 2, 0                   ; Scan first empty slot

            cmp     eax, -1
            jne     @F

            mov     eax, NULL
            jmp     Done

@@:         mcall   NEWDOS80_VTBL, DEC2FDE, pND, al                     ; Convert DEC into pointer to FPDE

            or      eax, eax                                            ; Check against null pointer
            jz      Done

            push    eax
            mov     ecx, sizeof ND_FPDE
@@:         mov     byte ptr [eax], 0                                   ; Fill entry with zeroes
            inc     eax
            loop    @B
            pop     eax

            push    eax
            lea     eax, (ND_FPDE PTR [eax]).Extent
            mov     ecx, 4+1
@@:         not     word ptr [eax]                                      ; Set extents as "not used" (0xFFFF)
            add     eax, sizeof ND_Extent
            loop    @B
            pop     eax

Done:       ret
            
ND_GetFDE   endp

;---------------------------------------------------------------------
; ND_GetExtPtr                             Return pointer to an extent
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to FPDE
; bExtent:  Required extent (starts at 1)
; bIEOF:    Ignore EOF (0:No, 1:Yes)
;---------------------------------------------------------------------
; EAX:      Pointer to extent or NULL if not found or -1 on error
;---------------------------------------------------------------------

ND_GetExtPtr    proc    pND:DWORD, pFPDE:DWORD, bExtent:BYTE, bIEOF:BYTE

            mov     eax, pFPDE

Set:        lea     edx, (ND_FPDE PTR [eax]).Extent                     ; Point to the extents table

            mov     ecx, 5                                              ; Scan up to 5 extents

Get:        mov     ax, word ptr [edx]                                  ; Get one extent

            cmp     bIEOF, 0
            jne     @F
            
            cmp     al, 255                                             ; 255 means no more extents for this file
            je      EOF

@@:         cmp     al, 254                                             ; 254 means other byte is a DEC linking to a FXDE
            je      Link

            dec     bExtent                                             ; Check whether the required extent has been achieved
            jnz     Next

            mov     eax, edx                                            ; Return pointer in EAX
            jmp     Done
            
Next:       add     edx, sizeof ND_Extent                               ; Advance to the next extent
            loop    Get

EOF:        mov     eax, NULL                                           ; Return null if not found
            jmp     Done

Link:       mcall   NEWDOS80_VTBL, DEC2FDE, pND, ah                     ; Get pointer to the FXDE

            or      eax, eax                                            ; Check against null pointer
            jnz     Set

            mov     eax, -1

Done:       ret

ND_GetExtPtr    endp

;---------------------------------------------------------------------
; ND_AllocGAT                             Allocate contiguous granules
;---------------------------------------------------------------------
; pND:          Pointer to the calling NEWDOS80 object
; bGranules:    Contiguous granules to allocate (up to 32)
;---------------------------------------------------------------------
; EAX:          New extent or NULL on error
; ECX:          Number of contiguous granules allocated
;---------------------------------------------------------------------

ND_AllocGAT proc    pND:DWORD, bGranules:BYTE

            local   bLumps:BYTE
            local   bGPL:BYTE
            local   bInitLump:BYTE
            local   bInitGran:BYTE
            local   bAllocGrans:BYTE
            
    ; --- Validate parameters ---

            cmp     bGranules, 0                                        ; Required number of granules cannot be zero
            ja      Init

            mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_INVALID_PARAMETER  ; The parameter is incorrect.
            mov     eax, NULL
            jmp     Done
            
    ; --- Initialize variables ---

Init:       mov     edx, pND
            lea     edx, (CNEWDOS80 PTR [edx]).PDRIVE

            mov     al, (ND_PDRIVE PTR [edx]).bLumps                    ; Copy Lumps Count to local var
            mov     bLumps, al
                        
            mov     al, (ND_PDRIVE PTR [edx]).bGPL                      ; Copy Granules Per Lump to local var
            mov     bGPL, al

            mov     bAllocGrans, 0                                      ; Clear allocated granules counter

            mov     edx, pND
            mov     edx, (CNEWDOS80 PTR [edx]).pDir                     ; EDX = Pointer to GAT
            xor     eax, eax                                            ; AL = Current lump
            xor     ecx, ecx                                            ; CL = Current granule

    ; --- Look for the first empty slot in GAT ---

Search:     cmp     bAllocGrans, 0                                      ; Find the first empty slot then continue up to the first non-empty slot
            jnz     Allocate
            
            mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Next

            cmp     al, bLumps                                          ; Check if reached a valid lump
            jnb     Error

            mov     bInitLump, al                                       ; Save initial lump
            mov     bInitGran, cl                                       ; Save initial granule

    ; --- Allocate as much contiguous granules as required ---

Allocate:   mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Finish

            or      [eax+edx], ch                                       ; Set granule as used

            inc     bAllocGrans                                         ; Increment count of granules allocated
            
            cmp     bAllocGrans, 32                                     ; Count of granules must fit in 5 bits
            je      Finish

            dec     bGranules                                           ; Decrement count of required granules
            jz      Finish                                              ; Exit when done

    ; --- Advance to next granule/lump as needed ---
    
Next:       inc     cl                                                  ; Advance to next granule
            cmp     cl, bGPL                                            ; Repeat for up to Granules Per Lump
            jb      Search

            mov     cl, 0                                               ; Reset current granule counter            

            inc     eax                                                 ; Advance to next cylinder

            cmp     al, bLumps                                          ; Check if still a valid lump
            jnb     Error

            cmp     eax, sizeof ND_GAT.abAllocTable                     ; Repeat for all allocation table bytes
            jb      Search

Error:      mov     eax, pND
            mov     (CNEWDOS80 PTR [eax]).dwLastError, ERROR_DISK_FULL  ; There is not enough space on the disk.
            mov     eax, NULL
            jmp     Done

    ; --- Set return registers ---

Finish:     movzx   eax, bInitLump                                      ; AL = Initial Lump

            mov     ah, bAllocGrans                                     ; AH (5 LSB) = Contiguous Granules - 1
            dec     ah

            shl     bInitGran, 5                                        ; AH (3 MSB) = Initial Granule
            or      ah, bInitGran

            movzx   ecx, bAllocGrans                                    ; ECX = Contiguous granules allocated
            
Done:       ret

ND_AllocGAT endp

;---------------------------------------------------------------------
; ND_FreeGAT                            Deallocate contiguous granules
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; wExtent:  Extent to deallocate
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

ND_FreeGAT  proc    pND:DWORD, wExtent:WORD

            local   bGPL:BYTE
            local   bLump:BYTE
            local   bGranules:BYTE
            local   bGranule:BYTE

    ; --- Initialize local variables ---
                
            mov     eax, pND
            lea     eax, (CNEWDOS80 PTR [eax]).PDRIVE
            mov     al, (ND_PDRIVE PTR [eax]).bGPL                      ; Copy Granules Per Lump to local var
            mov     bGPL, al

            mov     ax, wExtent

            mov     bLump, al                                           ; Get the initial cylinder
                
            mov     bGranules, ah                                       ; Get the number of granules in the extent
            and     bGranules, 00011111y
            inc     bGranules

            mov     bGranule, ah                                        ; Get the initial granule
            shr     bGranule, 5

    ; --- Deallocate all granules in the extent ---

NextLump:   mov     edx, pND                                            ; Point to GAT[Lump]
            mov     edx, (CNEWDOS80 PTR [edx]).pDir
            movzx   eax, bLump
            add     edx, eax

ResetGran:  mov     cl, bGranule                                        ; Reset bGranule at GAT[Lump]
            mov     al, 1
            shl     al, cl
            not     al
            and     [edx], al

            dec     bGranules                                           ; Check whether all have been done
            jnz     @F
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done
            
@@:         inc     bGranule                                            ; Advance to the next granule

            mov     al, bGranule                                        ; Check the granules per lump limit
            cmp     al, bGPL
            jb      ResetGran

            mov     bGranule, 0                                         ; Advance to the next cylinder
            inc     bLump
            
            cmp     bLump, sizeof ND_GAT.abAllocTable                   ; Check the lump against the allocation table limit
            jb      NextLump

            mov     eax, ERROR_FLOPPY_WRONG_CYLINDER                    ; Mismatch between the floppy disk sector ID field and the floppy disk controller track address.

Done:       ret

ND_FreeGAT  endp

;---------------------------------------------------------------------
; ND_ScanHIT                                 Scan the Hash Index Table
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; bMode:    0: Return first non-empty (optionally matching) slot
;           1: Return next non-empty (optionally matching) slot
;           2: Return first empty slot
; bHash:    Optional hash code for matching
;---------------------------------------------------------------------
; EAX:      DEC or -1 if not found
;---------------------------------------------------------------------

ND_ScanHIT  proc    pND:DWORD, bMode:BYTE, bHash:BYTE

            local   bCols:BYTE
            local   bRows:BYTE
            local   bCol:BYTE
            local   bRow:BYTE

            mov     edx, pND

            mov     al, (CNEWDOS80 PTR [edx]).bDirSectors               ; Compute max HIT cols
            sub     al, 2
            mov     bCols, al

            mov     bRows, 256 / sizeof ND_FPDE                         ; Compute max HIT rows

            mov     ax, (CNEWDOS80 PTR [edx]).wLastScan                 ; Recover last scan position
            
            mov     bCol, ah
            mov     bRow, al

            mov     edx, (CNEWDOS80 PTR [edx]).pDir                     ; Point to the HIT
            add     edx, 256

            cmp     bMode, 1                                            ; Jump if "find next" requested
            je      Next

            mov     bCol, 0                                             ; Otherwise restart from 0,0
            mov     bRow, 0

Scan:       movzx   eax, bRow                                           ; Compute current position (Row x sizeof FPDE + Col)
            shl     al, 5
            add     al, bCol
            
            cmp     eax, 31                                             ; The 32nd byte of HIT contains the count of extra FDE sectors
            je      Next

            cmp     byte ptr [edx+eax], 0                               ; Test the slot
            setne   cl

            cmp     bMode, 2                                            ; Invert result if requested mode is "2"
            jne     @F

            xor     cl, 1

@@:         or      cl, cl                                              ; Jump if search result has been achieved
            jnz     Found

Next:       inc     bCol                                                ; Advance to the next column

            mov     al, bCol                                            ; Check the end of columns
            cmp     al, bCols
            jb      Scan

            mov     bCol, 0

            inc     bRow                                                ; Advance to the next row

            mov     al, bRow                                            ; Check the end of rows
            cmp     al, bRows
            jb      Scan

            mov     eax, -1                                             ; Not found: return -1
            jmp     Done

Found:      cmp     bHash, 0                                            ; Check whether a hash code has been indicated
            je      @F

            mov     cl, bHash                                           ; Compare with the value in the slot
            cmp     byte ptr [edx+eax], cl
            jne     Next

@@:         mov     ch, bCol                                            ; Save current position for next scan
            mov     cl, bRow
            
            mov     edx, pND
            mov     (CNEWDOS80 PTR [edx]).wLastScan, cx
            
Done:       ret

ND_ScanHIT  endp

;---------------------------------------------------------------------
; ND_SetHIT                          Set a HIT slot to a new hash code
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the FPDE whose HIT slot is to be set
; bHash:    New hash code (0 indicates the slot is free)
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

ND_SetHIT   proc    pND:DWORD, pFPDE:DWORD, bHash:BYTE

            mcall   NEWDOS80_VTBL, FDE2DEC, pND, pFPDE                  ; Compute DEC from pointer to FPDE

            mov     edx, pND                                            ; Compute pointer to the corresponding HIT slot
            mov     edx, (CNEWDOS80 PTR [edx]).pDir
            add     edx, 256*1
            add     edx, eax                                            ; EAX = DEC

            mov     al, bHash                                           ; Set the new hash code
            mov     [edx], al

            ret

ND_SetHIT   endp

;---------------------------------------------------------------------
; ND_DEC2FDE                    Return pointer to File Directory Entry
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; bDEC:     Directory Entry Code (DEC)
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

ND_DEC2FDE  proc    pND:DWORD, bDEC:BYTE

            movzx   eax, bDEC
            
            mov     ah, al
            and     ah, 00011111y                                       ; Strip column bits to get the sector offset
            and     al, 11100000y                                       ; Strip row bits to get the FDE offset (already multiplied by 32 bytes)

            mov     edx, pND

            cmp     ah, (CNEWDOS80 PTR [edx]).bDirSectors               ; Check whether this is will be a valid pointer
            jb      @F

            mov     eax, NULL                                           ; Return NULL if not
            jmp     Done

@@:         add     eax, (CNEWDOS80 PTR [edx]).pDir                     ; Make a pointer to the directory
            add     eax, 256*2                                          ; Skip GAT and HIT sectors

Done:       ret

ND_DEC2FDE  endp

;---------------------------------------------------------------------
; ND_FDE2DEC                          Convert pointer to FPDE into DEC
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Directory Entry Code (index to HIT slot containg the hash)
;---------------------------------------------------------------------

ND_FDE2DEC  proc    pND:DWORD, pFPDE:DWORD

            mov     edx, pND
            mov     eax, pFPDE
            sub     eax, (CNEWDOS80 PTR [edx]).pDir                     ; Subtract pointer to FPDE from pointer to directory buffer
            sub     eax, 256*2                                          ; Subtract GAT and HIT sector sizes
            shr     eax, 5                                              ; Divide by the size of each directory entry (32 bytes)
            ror     al, 3                                               ; Reorganize bits: 11111000 -> 00011111
            ret

ND_FDE2DEC  endp

;---------------------------------------------------------------------
; ND_Hash                                             Hash a file name
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; pFile:    Pointer to a blank-padded file name and extension
;---------------------------------------------------------------------
; AL:       Computed hash code
;---------------------------------------------------------------------

ND_Hash     proc    pND:DWORD, pFile:DWORD

            mov     edx, pFile
            mov     ecx, 11
            xor     eax, eax
            
@@:         xor     al, byte ptr [edx]
            inc     edx
            rol     al, 1
            loop    @B
            
            or      al, al
            jnz     Done
            
            inc     al

Done:       ret

ND_Hash     endp

;---------------------------------------------------------------------
; ND_Sector2CHS                       Convert relative sector to C/H/S
;---------------------------------------------------------------------
; pND:      Pointer to the calling NEWDOS80 object
; wSector:  Relative sector number
;---------------------------------------------------------------------
; CL:       Cylinder
; DH:       Head
; DL:       Sector
;---------------------------------------------------------------------

ND_Sector2CHS   proc    pND:DWORD, wSector:WORD

            mov     edx, pND
            lea     edx, (CNEWDOS80 PTR [edx]).PDRIVE
            
            mov     ch, (ND_PDRIVE PTR [edx]).bSPT                      ; Get Sectors Per Cylinder
            
            test    (ND_PDRIVE PTR [edx]).bTD, 00000010b                ; Test if disk is double sided (only TD=2/3/6/7 have bit 1 set)
            jz      @F
            
            shr     ch, 1                                               ; Convert Sectors Per Cylinder into Sectors Per Track

@@:         mov     ax, wSector                                         ; Cylinder = (Relative Sector / Sectors Per Cylinder) + FirstTrack
            div     (ND_PDRIVE PTR [edx]).bSPT
            mov     cl, al                                              ; CL = Cylinder

            movzx   ax, ah                                              ; Head = (Remainder / Sectors Per Track) + FirstSide
            div     ch                                                  ; Sector = Last remainder

            mov     dh, al                                              ; DH = Head
            mov     dl, ah                                              ; DL = Sector

            mov     eax, pND            
            add     cl, (CNEWDOS80 PTR [eax]).bFirstTrack               ; Adjust track
            add     dh, (CNEWDOS80 PTR [eax]).bFirstSide                ; Adjust side

            ret

ND_Sector2CHS   endp

