;---------------------------------------------------------------------
; Operating System Interface                Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; OSI_FindFiles global definitions
;---------------------------------------------------------------------

OSI_FINDFIRST           equ     1
OSI_FINDNEXT            equ     0

.data?

;---------------------------------------------------------------------
; OSI Class
;---------------------------------------------------------------------

COSI    struct
    VPTR                DWORD   ?                                       ; Pointer to the object's VTBL
COSI    ends

.data

;---------------------------------------------------------------------
; OSI Virtual Table
;---------------------------------------------------------------------

OSI_VTBL    struct                                                      ; Declare the structure and initializers
    Load                DWORD   OSI_Load
    Unload              DWORD   OSI_Unload
    Find                DWORD   OSI_Find
    Open                DWORD   OSI_Open
    Create              DWORD   OSI_Create
    Seek                DWORD   OSI_Seek
    Read                DWORD   OSI_Read
    Write               DWORD   OSI_Write
    Rename              DWORD   OSI_Rename
    Delete              DWORD   OSI_Delete
    GetLastError        DWORD   OSI_GetLastError
    GetDosVersion       DWORD   OSI_GetDosVersion
    SetDosVersion       DWORD   OSI_SetDosVersion
    GetDiskName         DWORD   OSI_GetDiskName
    SetDiskName         DWORD   OSI_SetDiskName
    GetDiskDate         DWORD   OSI_GetDiskDate
    SetDiskDate         DWORD   OSI_SetDiskDate
    GetFilePos          DWORD   OSI_GetFilePos
    GetFileName         DWORD   OSI_GetFileName
    GetFileExt          DWORD   OSI_GetFileExt
    GetFileSize         DWORD   OSI_GetFileSize
    GetFileDate         DWORD   OSI_GetFileDate
    SetFileDate         DWORD   OSI_SetFileDate
    GetFileAttr         DWORD   OSI_GetFileAttr
    SetFileAttr         DWORD   OSI_SetFileAttr
OSI_VTBL    ends

OSI                     OSI_VTBL    <>                                  ; Define a variable of that type

.code

;---------------------------------------------------------------------
; OSI_Load                                        Initialize interface
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Load    proc    pOSI:DWORD, pVDI:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Load    endp

;---------------------------------------------------------------------
; OSI_Unload                                         Release interface
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Unload  proc    pOSI:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Unload  endp

;---------------------------------------------------------------------
; OSI_Find                               Search files in the directory
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; bFlag:    OSI_FINDFIRST or OSI_FINDNEXT
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

OSI_Find    proc    pOSI:DWORD, bFlag:BYTE

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Find    endp

;---------------------------------------------------------------------
; OSI_Open                                       Find an existing file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pName:    Pointer to a 11-byte blank-padded string with the filename
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

OSI_Open    proc    pOSI:DWORD, pName:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Open    endp

;---------------------------------------------------------------------
; OSI_Create                                         Create a new file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pName:    Pointer to a 11-byte blank-padded string with the filename
; dwSize:   File size
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

OSI_Create  proc    pOSI:DWORD, pName:DWORD, dwSize:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Create  endp

;---------------------------------------------------------------------
; OSI_Seek                                           Move file pointer
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwPos:    New pointer position
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Seek    proc    pOSI:DWORD, pFPDE:DWORD, dwPos:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Seek    endp

;---------------------------------------------------------------------
; OSI_Read                                         Read data from file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to the receiving buffer
; dwBytes:  Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Read    proc    pOSI:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Read    endp

;---------------------------------------------------------------------
; OSI_Write                                         Write data to file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to the data buffer
; dwBytes:  Number of bytes to write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Write   proc    pOSI:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Write   endp

;---------------------------------------------------------------------
; OSI_Rename                                               Rename file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the FPDE of the file to rename
; pName:    Pointer to a 11-byte blank-padded string with the new name
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Rename  proc    pOSI:DWORD, pFPDE:DWORD, pName:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Rename  endp

;---------------------------------------------------------------------
; OSI_Delete                                               Delete file
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the FPDE of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_Delete  proc    pOSI:DWORD, pFPDE:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_Delete  endp

;---------------------------------------------------------------------
; OSI_GetLastError                 Return the last internal error code
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
;---------------------------------------------------------------------
; EAX:      Last error code
;---------------------------------------------------------------------

OSI_GetLastError    proc    pOSI:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret

OSI_GetLastError    endp

;---------------------------------------------------------------------
; OSI_GetDosVersion                             Return the DOS version
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

OSI_GetDosVersion   proc    pOSI:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetDosVersion   endp

;---------------------------------------------------------------------
; OSI_SetDosVersion                              Set a new DOS version
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; bVersion: New DOS version to set
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

OSI_SetDosVersion   proc    pOSI:DWORD, bVersion:BYTE

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_SetDosVersion   endp

;---------------------------------------------------------------------
; OSI_GetDiskName                                 Return the disk name
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

OSI_GetDiskName proc    pOSI:DWORD, pStr:DWORD, dwLen:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetDiskName endp

;---------------------------------------------------------------------
; OSI_SetDiskName                                    Set the disk name
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pName:    Pointer to a 11-byte blank-padded string with the new name
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

OSI_SetDiskName proc    pOSI:DWORD, pName:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_SetDiskName endp

;---------------------------------------------------------------------
; OSI_GetDiskDate                                 Return the disk date
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

OSI_GetDiskDate proc    pOSI:DWORD, pStr:DWORD, dwLen:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetDiskDate endp

;---------------------------------------------------------------------
; OSI_SetDiskDate                                    Set the disk date
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pDate:    Pointer to a 8-byte blank-padded string with the new date
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

OSI_SetDiskDate proc    pOSI:DWORD, pDate:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_SetDiskDate endp

;---------------------------------------------------------------------
; OSI_GetFilePos                           Return current file pointer
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Current pointer position or -1 for an invalid position
;---------------------------------------------------------------------

OSI_GetFilePos  proc    pOSI:DWORD, pFPDE:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFilePos  endp

;---------------------------------------------------------------------
; OSI_GetFileName                                 Return the file name
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_GetFileName proc    pOSI:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFileName endp

;---------------------------------------------------------------------
; OSI_GetFileExt                             Return the file extension
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to the receiving string buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_GetFileExt  proc    pOSI:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFileExt  endp

;---------------------------------------------------------------------
; OSI_GetFileSize                                 Return the file size
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

OSI_GetFileSize proc    pOSI:DWORD, pFPDE:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFileSize endp

;---------------------------------------------------------------------
; OSI_GetFileDate                                 Return the file date
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

OSI_GetFileDate proc    pOSI:DWORD, pFPDE:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFileDate endp

;---------------------------------------------------------------------
; OSI_SetFileDate                                 Change the file date
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwDate:   New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_SetFileDate proc    pOSI:DWORD, pFPDE:DWORD, dwDate:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_SetFileDate endp

;---------------------------------------------------------------------
; OSI_GetFileAttr                           Return the file attributes
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

OSI_GetFileAttr proc    pOSI:DWORD, pFPDE:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_GetFileAttr endp

;---------------------------------------------------------------------
; OSI_SetFileAttr                           Change the file attributes
;---------------------------------------------------------------------
; pOSI:     Pointer to the calling OSI object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwAttr:   New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

OSI_SetFileAttr proc    pOSI:DWORD, pFPDE:DWORD, dwAttr:DWORD

            mov     eax, ERROR_CALL_NOT_IMPLEMENTED
            ret
            
OSI_SetFileAttr endp

