;---------------------------------------------------------------------
; TRSDOS 6.x Operating System Interface     Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; TRSDOS GAT disk configuration flags
;---------------------------------------------------------------------

TD_GAT_Type                 equ         10000000y                       ; 0:System Disk, 1:Data Disk (define # of reserved slots in the HIT)
TD_GAT_Density              equ         01000000y                       ; 0:Single Density, 1:Double Density
TD_GAT_Sides                equ         00100000y                       ; 0:Single Sided, 1:Double Sided
TD_GAT_Granules             equ         00000111y                       ; Number of granules per track minus 1

;---------------------------------------------------------------------
; TRSDOS File Primary Directory Entry attributes
;---------------------------------------------------------------------

TD_FPDE_Extended            equ         10000000y                       ; 0:File Primary Directory Entry, 1:File Extended Directory Entry
TD_FPDE_System              equ         01000000y                       ; 0:Normal file, 1:System file
TD_FPDE_PDS                 equ         00100000y                       ; 1:Partition Data Set (PDS) file
TD_FPDE_Active              equ         00010000y                       ; 0:Directory Record inactive, 1:Directory Record in use
TD_FPDE_Invisible           equ         00001000y                       ; 0:File Visible, 1:File Invisible
TD_FPDE_Access              equ         00000111y                       ; USER protection level (0=Full, 1=Remove, 2=Rename, 3=Write, 4=Update, 5=Read, 6=Execute, 7=No Access)

TD_FPDE_Created             equ         10000000y                       ; 1:File was CREATEd
TD_FPDE_Modified            equ         01000000y                       ; 1:Backup pending
TD_FPDE_Open                equ         00100000y                       ; 1:File in an open condition with UPDATE access or greater
TD_FPDE_BadDate             equ         00010000y                       ; 1:The date of the next bits is inaccurate
TD_FPDE_Month               equ         00001111y                       ; Binary month of the last modification date

TD_FPDE_Day                 equ         11111000y                       ; Binary day of last modification
TD_FPDE_Year                equ         00000111y                       ; Binary year minus 80 (1980 = 000, 1981 = 001...)

;---------------------------------------------------------------------
; TRSDOS directory operations
;---------------------------------------------------------------------

TD_DIR_READ                 equ         0                               ; Read directory from disk
TD_DIR_WRITE                equ         1                               ; Write directory to disk

.data?

;---------------------------------------------------------------------
; TRSDOS Granule Allocation Table (GAT)
;---------------------------------------------------------------------

TD_GAT  struct
    abAllocTable            BYTE        96  dup (?)                     ; Free/Assigned table information
    abLockoutTable          BYTE        96  dup (?)                     ; Used during mirror-image backups to determine if the destination disk has the proper capacity
    abExtendedTable         BYTE        11  dup (?)                     ; Used in hard drive configurations to extend the table from 00 to CA
    bDosVersion             BYTE        ?                               ; Operation System version used in formatting the disk
    bCylinderExcess         BYTE        ?                               ; Number of cylinders in excess of 35
    bDiskConfig             BYTE        ?                               ; Data specific to the formatting of the diskette
    wPasswordHash           WORD        ?                               ; Master password hash
    abDiskName              BYTE        8   dup (?)                     ; Disk name
    abDiskDate              BYTE        8   dup (?)                     ; Disk date
    abReserved              BYTE        21  dup (?)                     ; Reserved for system use
    abMDB                   BYTE        4   dup (?)                     ; Media Data Block (MDB) (0x03, 'LSI')
    abDCT                   BYTE        7   dup (?)                     ; Last 7 bytes of the DCT when the media was formatted
TD_GAT  ends

;---------------------------------------------------------------------
; TRSDOS File Extent Element
;---------------------------------------------------------------------

TD_Extent   struct
    bCylinder               BYTE        ?                               ; Cylinder value for the starting granule of the extent
    bGranules               BYTE        ?                               ; First granule (bits 5-7) and number of contiguous granules (bits 0-4)
TD_Extent   ends

;---------------------------------------------------------------------
; TRSDOS File Primary Directory Entry
;---------------------------------------------------------------------

TD_FPDE struct
    abAttributes            BYTE        3 dup (?)                       ; File attributes
    bEOF                    BYTE        ?                               ; EOF offset byte
    bLRL                    BYTE        ?                               ; Logical Record Length (LRL)
    abFileName              BYTE        8 dup (?)                       ; File name
    abExtension             BYTE        3 dup (?)                       ; File extension
    wOwnerHash              WORD        ?                               ; OWNER password hash
    wUserHash               WORD        ?                               ; USER password hash
    wERN                    WORD        ?                               ; Ending Record Number (ERN)
    Extent                  TD_Extent   4 dup (<>)                      ; Four extent elements
    Link                    TD_Extent   <>                              ; Link to File Extended Directory Entry
TD_FPDE ends

;---------------------------------------------------------------------
; TRSDOS Class
;---------------------------------------------------------------------

CTRSDOS struct
    VPTR                    DWORD       ?                               ; Pointer to the object's VTBL
    pVDI                    DWORD       ?                               ; Pointer to an initialized VDI object
    pDir                    DWORD       ?                               ; Pointer to Directory data
    bDirTrack               BYTE        ?                               ; Number of directory cylinder
    bDirSectors             BYTE        ?                               ; Number of directory sectors
    bFirstTrack             BYTE        ?                               ; Number of the first track on disk (0 or 1)
    bFirstSide              BYTE        ?                               ; Number of the first disk side (0 or 1)
    bFirstSector            BYTE        ?                               ; Number of the first sector on track (0 or 1)
    bSides                  BYTE        ?                               ; Number of disk sides (1 or 2)
    bSectorsPerTrack        BYTE        ?                               ; Number of sectors per track
    bGranulesPerTrack       BYTE        ?                               ; Number of granules per track
    bGranulesPerCylinder    BYTE        ?                               ; Number of granules per cylinder
    bSectorsPerGranule      BYTE        ?                               ; Number of sectors per granule
    dwSeekPointer           DWORD       ?                               ; Current file pointer set by Seek()
    wSeekSector             WORD        ?                               ; Current relative sector computed by Seek()
    wLastScan               WORD        ?                               ; Last Directory Entry Code found by ScanHIT()
    dwLastError             DWORD       ?                               ; Code of last error occurred in an internal function
CTRSDOS ends

.data

;---------------------------------------------------------------------
; TRSDOS Virtual Table
;---------------------------------------------------------------------

TRSDOS_VTBL struct                                                      ; Declare the structure and initializers
    Load                    DWORD       TD_Load                         ; OSI-derived functions
    Unload                  DWORD       TD_Unload
    Find                    DWORD       TD_Find
    Open                    DWORD       TD_Open
    Create                  DWORD       TD_Create
    Seek                    DWORD       TD_Seek
    Read                    DWORD       TD_Read
    Write                   DWORD       TD_Write
    Rename                  DWORD       TD_Rename
    Delete                  DWORD       TD_Delete
    GetLastError            DWORD       TD_GetLastError
    GetDosVersion           DWORD       TD_GetDosVersion
    SetDosVersion           DWORD       TD_SetDosVersion
    GetDiskName             DWORD       TD_GetDiskName
    SetDiskName             DWORD       TD_SetDiskName
    GetDiskDate             DWORD       TD_GetDiskDate
    SetDiskDate             DWORD       TD_SetDiskDate
    GetFilePos              DWORD       TD_GetFilePos
    GetFileName             DWORD       TD_GetFileName
    GetFileExt              DWORD       TD_GetFileExt
    GetFileSize             DWORD       TD_GetFileSize
    GetFileDate             DWORD       TD_GetFileDate
    SetFileDate             DWORD       TD_SetFileDate
    GetFileAttr             DWORD       TD_GetFileAttr
    SetFileAttr             DWORD       TD_SetFileAttr
    DirRW                   DWORD       TD_DirRW                        ; TRSDOS-specific functions
    GetFDE                  DWORD       TD_GetFDE
    GetExtPtr               DWORD       TD_GetExtPtr
    AllocGAT                DWORD       TD_AllocGAT
    FreeGAT                 DWORD       TD_FreeGAT
    ScanHIT                 DWORD       TD_ScanHIT
    SetHIT                  DWORD       TD_SetHIT
    DEC2FDE                 DWORD       TD_DEC2FDE
    FDE2DEC                 DWORD       TD_FDE2DEC
    Hash                    DWORD       TD_Hash
    Sector2CHS              DWORD       TD_Sector2CHS
TRSDOS_VTBL ends

TRSDOS                      TRSDOS_VTBL <>                              ; Define a variable of that type

.code

;---------------------------------------------------------------------
; TD_Load                                         Initialize interface
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Load     proc    pTD:DWORD, pVDI:DWORD

            local   abBoot[256]:BYTE
            local   pDir:DWORD
            local   pDirCopy:DWORD
            local   bDirTrack:BYTE
            local   bDirSectors:BYTE
            local   bFirstTrack:BYTE
            local   bFirstSide:BYTE
            local   bFirstSector:BYTE
            local   bSides:BYTE
            local   bSectorsPerTrack:BYTE
            local   bGranulesPerTrack:BYTE
            local   bGranulesPerCylinder:BYTE
            local   bSectorsPerGranule:BYTE
            local   bDensity:BYTE

    ; --- Get disk information ---

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     bFirstTrack, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     bFirstSide, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     bFirstSector, al

            mcall   VDI_VTBL, GetSides, pVDI
            mov     bSides, al

            mcall   VDI_VTBL, GetDensity, pVDI
            mov     bDensity, al

    ; --- Normalize Sectors Per Track ---

            mcall   VDI_VTBL, GetSPT, pVDI                              ; Get Sectors Per Track

            cmp     al, 26                                              ; 8" DD                                              
            jb      @F

            mov     al, 26
            jmp     SetSPT
            
@@:         cmp     al, 18                                              ; 5 1/4" DD
            jb      @F

            mov     al, 18
            jmp     SetSPT
            
@@:         cmp     al, 15                                              ; 8" SD
            jb      @F

            mov     al, 15
            jmp     SetSPT
            
@@:         mov     al, 10                                              ; 5 1/4" SD

SetSPT:     mov     bSectorsPerTrack, al

    ; --- Load the boot sector to find out the directory cylinder ---

            mcall   VDI_VTBL, Read, pVDI, addr abBoot, sizeof abBoot, bFirstTrack, bFirstSide, bFirstSector
            
            or      eax, eax
            jnz     Done

            cmp     abBoot[1], 0FEh                                     ; Kind of integrity check
            jne     Error

            mov     al, abBoot[2]                                       ; This byte contains the directory cylinder
            mov     bDirTrack, al
            
    ; --- Compute the number of directory sectors ---

            mov     al, bSectorsPerTrack                                ; Directory Sectors = Sectors per Track * Sides
            mul     bSides

            cmp     al, 32+2                                            ; Max sectors limited by HIT: 256 hashes / 8 FDEs per sector = 32 sectors (+2 for GAT and HIT)
            jbe     @F

            mov     al, 32+2                                            ; Set top limit

@@:         mov     bDirSectors, al
            
    ; --- Allocate memory for the entire directory and a copy of it ---

            shl     eax, 8                                              ; Buffer Size = DirSectors * 256 bytes per sector (TRSDOS standard)

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pDir, eax                                           ; Save pointer to Directory buffer

    ; --- Update member class variables ---
    
            mov     edx, pTD
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (CTRSDOS PTR [edx]).pVDI, eax

            mov     eax, pDir                                           ; Update pointer to Directory buffer
            mov     (CTRSDOS PTR [edx]).pDir, eax

            mov     al, bDirTrack                                       ; Update number of the directory cylinder
            mov     (CTRSDOS PTR [edx]).bDirTrack, al

            mov     al, bDirSectors                                     ; Update number of directory sectors
            mov     (CTRSDOS PTR [edx]).bDirSectors, al

            mov     al, bFirstTrack                                     ; Update number of the first track on disk
            mov     (CTRSDOS PTR [edx]).bFirstTrack, al

            mov     al, bFirstSide                                      ; Update number of the first disk side
            mov     (CTRSDOS PTR [edx]).bFirstSide, al

            mov     al, bFirstSector                                    ; Update number of the first sector on track
            mov     (CTRSDOS PTR [edx]).bFirstSector, al

            mov     al, bSides                                          ; Update number of disk sides
            mov     (CTRSDOS PTR [edx]).bSides, al

            mov     al, bSectorsPerTrack                                ; Update number of Sectors per Track
            mov     (CTRSDOS PTR [edx]).bSectorsPerTrack, al

    ; --- Load directory into memory ---

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0
            
            or      eax, eax
            jnz     Done

    ; --- Compute and validate the disk's granule units ---

            mov     edx, pDir                                           ; Pointer to the first directory sector (GAT)

            mov     cl, (TD_GAT PTR [edx]).bDiskConfig                  ; Find the number of granules per track
            and     cl, TD_GAT_Granules
            inc     cl
            mov     bGranulesPerTrack, cl

            mov     al, bSides                                          ; Find the number of granules per cylinder
            mul     cl
            mov     bGranulesPerCylinder, al

            cmp     al, 2                                               ; GranulesPerCylinder must be between 2 and 8
            jb      Error

            cmp     al, 8
            ja      Error
            
            mov     al, bSectorsPerTrack                                ; Compute the number of sectors per granule
            div     bGranulesPerTrack
            mov     bSectorsPerGranule, al

            or      ah, ah                                              ; There must be no remainder after last division
            jnz     Error

            cmp     al, 5                                               ; Sectors per Granule for 5.25" SDEN disks
            je      Single
    
            cmp     al, 6                                               ; Sectors per Granule for 5.25" DDEN disks
            je      Double
    
            cmp     al, 8                                               ; Sectors per Granule for 8" SDEN disks
            je      Single
    
            cmp     al, 10                                              ; Sectors per Granule for 8" DDEN disks
            jne     Error
    
    ; --- Check if the SPG conforms with the disk density ---

Double:     cmp     bDensity, 2                                         ; Check if the disk is double or mixed density
            je      @F

            cmp     bDensity, 1
            je      @F
            jmp     Error

Single:     cmp     bDensity, 2                                         ; Check if the disk is single or mixed density
            je      @F

            cmp     bDensity, 0
            je      @F
            jmp     Error

    ; --- Update remaining member class variables ---

@@:         mov     edx, pTD
            
            mov     al, bGranulesPerTrack                               ; Update number of Granules Per Track
            mov     (CTRSDOS PTR [edx]).bGranulesPerTrack, al

            mov     al, bGranulesPerCylinder                            ; Update number of Granules Per Cylinder
            mov     (CTRSDOS PTR [edx]).bGranulesPerCylinder, al

            mov     al, bSectorsPerGranule                              ; Update number of Sectors Per Granule
            mov     (CTRSDOS PTR [edx]).bSectorsPerGranule, al

            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done

Error:      mov     eax, ERROR_NOT_DOS_DISK                             ; The specified disk or diskette cannot be accessed.

Done:       ret

TD_Load     endp

;---------------------------------------------------------------------
; TD_Unload                                          Release interface
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Unload   proc    pTD:DWORD

            mov     eax, pTD

            fn      GlobalFree, (CTRSDOS PTR [eax]).pDir

            or      eax, eax
            jz      Done

            fn      GetLastError

Done:       ret

TD_Unload   endp

;---------------------------------------------------------------------
; TD_Find                                Search files in the directory
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; bFlag:    OSI_FINDFIRST or OSI_FINDNEXT
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error (use GetLastError)
;---------------------------------------------------------------------

TD_Find     proc    pTD:DWORD, bFlag:BYTE

            local   bMode:BYTE

            mov     bMode, 0
            
            cmp     bFlag, OSI_FINDFIRST
            je      Find

            mov     bMode, 1

Find:       mcall   TRSDOS_VTBL, ScanHIT, pTD, bMode, 0                     ; Scan the first/next non-empty slot

            cmp     eax, -1
            jne     @F
            
            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_NO_MORE_FILES    ; There are no more files.
            mov     eax, NULL
            jmp     Done

@@:         mcall   TRSDOS_VTBL, DEC2FDE, pTD, al                           ; Get pointer to directory entry based on DEC

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_INVALID_ACCESS   ; The access code is invalid.
            mov     eax, NULL
            jmp     Done

@@:         mov     bMode, 1                                                ; Switch to "next file"

            test    (TD_FPDE PTR [eax]).abAttributes[0], TD_FPDE_Active     ; Check if the entry is active
            jz      Find

            test    (TD_FPDE PTR [eax]).abAttributes[0], TD_FPDE_Extended   ; Check if the entry is primary
            jnz     Find

Done:       ret

TD_Find     endp

;---------------------------------------------------------------------
; TD_Open                                        Find an existing file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pName:    Pointer to a 11-byte blank-padded string with the filename
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

TD_Open     proc    pTD:DWORD, pName:DWORD

            local   bMode:BYTE
            local   bHash:BYTE

            mov     bMode, 0
            
            mcall   TRSDOS_VTBL, Hash, pTD, pName                           ; Hash the file name
            mov     bHash, al

Find:       mcall   TRSDOS_VTBL, ScanHIT, pTD, bMode, bHash                 ; Scan the first/next non-empty slot

            cmp     eax, -1
            jne     @F
            
            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_FILE_NOT_FOUND   ; The system cannot find the file specified.
            mov     eax, NULL
            jmp     Done

@@:         mov     bMode, 1                                                ; Set to "Find Next" on next scan
    
            mcall   TRSDOS_VTBL, DEC2FDE, pTD, al                           ; Convert DEC into pointer to FPDE

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_INVALID_ACCESS   ; The access code is invalid.
            mov     eax, NULL
            jmp     Done

@@:         test    (TD_FPDE PTR [eax]).abAttributes[0], TD_FPDE_Active     ; Check if the FPDE is active
            jz      Find

            test    (TD_FPDE PTR [eax]).abAttributes[0], TD_FPDE_Extended   ; Check if the entry is primary
            jnz     Find

            push    esi
            push    edi
            
            lea     esi, (TD_FPDE PTR [eax]).abFileName
            mov     edi, pName
            mov     ecx, sizeof TD_FPDE.abFileName + sizeof TD_FPDE.abExtension
            cld
            repe    cmpsb                                                   ; Compare the two file names
            
            pop     edi
            pop     esi
            jne     Find

Done:       ret
    
TD_Open     endp

;---------------------------------------------------------------------
; TD_Create                                          Create a new file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pName:    Pointer to a 11-byte blank-padded string with the filename
; dwSize:   File size
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

TD_Create   proc    pTD:DWORD, pName:DWORD, dwSize:DWORD

            local   pFPDE:DWORD
            local   pFDE:DWORD
            local   pLastFDE:DWORD
            local   bGranules:BYTE
            local   bExtCount:BYTE
            local   wExtent:WORD
            local   bAllocated:BYTE

            mov     bExtCount, 0

    ; --- Find an available directory entry ---

            mcall   TRSDOS_VTBL, GetFDE, pTD

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
            jmp     Error

@@:         mov     pFPDE, eax
            
    ; --- Fill out the entry ---

            mov     (TD_FPDE PTR [eax]).abAttributes[0], TD_FPDE_Active     ; Mark entry as active

            mov     (TD_FPDE PTR [eax]).wOwnerHash, 4296h                   ; Owner password is blanks
            mov     (TD_FPDE PTR [eax]).wUserHash, 4296h                    ; User password is blanks

            mov     edx, dwSize                                             ; Set End Of File (EOF)
            mov     (TD_FPDE PTR [eax]).bEOF, dl
            
            shr     edx, 8                                                  ; Set Ending Record Number (ERN)
            mov     (TD_FPDE PTR [eax]).wERN, dx

            cmp     (TD_FPDE PTR [eax]).bEOF, 0                             ; If EOF is not zero increment ERN
            je      @F
            
            inc     (TD_FPDE PTR [eax]).wERN
            
@@:         fn      MemCopy, pName, addr (TD_FPDE PTR [eax]).abFileName, sizeof TD_FPDE.abFileName + sizeof TD_FPDE.abExtension

    ; --- Update the Hash Index Table ---
    
            mcall   TRSDOS_VTBL, Hash, pTD, pName                           ; Hash the file name

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFPDE, al                     ; Update HIT slot with the new hash code
            
    ; --- Compute the required number of granules ---
    
            mov     edx, pTD
            
            mov     eax, pFPDE
            mov     pFDE, eax
            
            mov     ax, (TD_FPDE PTR [eax]).wERN                            ; Granules = ERN / SectorsPerGranule
            div     (CTRSDOS PTR [edx]).bSectorsPerGranule
            or      ah, ah
            jz      @F
            inc     al
@@:         mov     bGranules, al

    ; --- Create an extent with as much granules as possible ---
    
NextExt:    mcall   TRSDOS_VTBL, AllocGAT, pTD, bGranules

            or      eax, eax                                                ; Check against null pointer
            jz      Done

            mov     wExtent, ax
            mov     bAllocated, cl

    ; --- Save the new extent in the extents table ---

            inc     bExtCount

            mcall   TRSDOS_VTBL, GetExtPtr, pTD, pFDE, bExtCount, 1         ; Get pointer to next extent

            or      eax, eax                                                ; Check if null
            jz      AllocFDE

            cmp     eax, -1                                                 ; Check for error
            jne     @F

            mov     eax, ERROR_INVALID_TABLE                                ; Invalid or unknown table specified.
            jmp     Error

@@:         push    wExtent                                                 ; Save extent
            pop     word ptr [eax]

    ; --- Check if all required granules have been allocated ---
    
            mov     al, bAllocated
            sub     bGranules, al
            jnz     NextExt

    ; --- Save Directory ---

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0

            or      eax, eax
            jnz     Error

            mov     eax, pFPDE                                              ; Return pointer to FPDE
            jmp     Done

    ; --- Obtain a new directory entry ---

AllocFDE:   mov     eax, pFDE                                               ; Save current pointer
            mov     pLastFDE, eax

            mcall   TRSDOS_VTBL, GetFDE, pTD                                ; Get a new pointer

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
            jmp     Error

@@:         mov     pFDE, eax                                               ; Save new pointer

    ; --- Set a forward link to the next entry ---

            mcall   TRSDOS_VTBL, FDE2DEC, pTD, pFDE                         ; Get its corresponding DEC

            mov     edx, pLastFDE
            
            mov     (TD_FPDE PTR [edx]).Link.bCylinder, 0FEh                ; 0xFE indicates a link exists
            mov     (TD_FPDE PTR [edx]).Link.bGranules, al                  ; This value is the DEC of the next entry
            
    ; --- Set a backward link to the previous entry ---

            mcall   TRSDOS_VTBL, FDE2DEC, pTD, pLastFDE                     ; Get DEC of previous entry

            mov     edx, pFDE

            mov     (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_Active+TD_FPDE_Extended    ; Mark the new entry as active and extended
            mov     (TD_FPDE PTR [edx]).abAttributes[1], al                                 ; In a FXDE this byte contains the backward link

    ; --- Copy the filename and set the HIT slot ---
    
            fn      MemCopy, pName, addr (TD_FPDE PTR [edx]).abFileName, sizeof TD_FPDE.abFileName + sizeof TD_FPDE.abExtension

            mcall   TRSDOS_VTBL, Hash, pTD, pName                           ; Hash the file name

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFDE, al                      ; Update HIT slot with the hash code
            
            jmp     NextExt

Error:      push    eax
            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0                 ; Restore previous directory state
            pop     eax

            mov     edx, pTD
            mov     (CTRSDOS PTR [edx]).dwLastError, eax                    ; Set last error
            
            mov     eax, NULL                                               ; Return null pointer

Done:       ret

TD_Create   endp

;---------------------------------------------------------------------
; TD_Seek               Convert a file position into a relative sector
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwPos:    New pointer position
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Seek     proc    pTD:DWORD, pFPDE:DWORD, dwPos:DWORD

            local   wSector:WORD
            local   bExtent:BYTE
            local   bSPG:BYTE

    ; --- Initialize local variables ---

            mov     edx, pTD
            
            mov     al, (CTRSDOS PTR [edx]).bSectorsPerGranule          ; Copy SectorsPerGranule to a local var
            mov     bSPG, al

            mov     eax, dwPos                                          ; wSector = dwPos / 256
            shr     eax, 8
            mov     wSector, ax

    ; --- Find in which extent the requested sector is ---

            mov     bExtent, 1
            
Scan:       mcall   TRSDOS_VTBL, GetExtPtr, pTD, pFPDE, bExtent, 0      ; Get pointer to extent

            or      eax, eax                                            ; Check if EOF
            jnz     @F
            
            mov     eax, ERROR_HANDLE_EOF                               ; Reached the end of the file.
            mov     wSector, -1
            jmp     Update

@@:         cmp     eax, -1                                             ; Check for error
            jne     @F

            mov     eax, ERROR_FILE_CORRUPT                             ; The file or directory is corrupted and unreadable.
            mov     wSector, -1
            jmp     Update

@@:         mov     ax, word ptr [eax]                                  ; Get the extent
   
            and     ah, 00011111y                                       ; Compute the number of sectors contained in this extent
            inc     ah
            shr     ax, 8
            mul     bSPG

            cmp     ax, wSector                                         ; Check if the requested sector is in the current extent
            ja      Found

            sub     wSector, ax                                         ; Subtract from sectors required in the next extent(s) and repeat
            inc     bExtent
            jmp     Scan

    ; --- Convert the extent information in a relative sector ---

Found:      mov     cx, word ptr [edx]                                  ; Get extent (CH = Initial_Granule, CL = Cylinder)

            mov     eax, pTD
            movzx   eax, (CTRSDOS PTR [eax]).bGranulesPerCylinder       ; Relative Sector = (Cylinder * GranulesPerCylinder + Initial_Granule) * SectorsPerGranule + Sector_Offset
            mul     cl

            and     ch, 11100000y                                       ; Add the initial granule
            shr     cx, 5+8
            add     ax, cx

            movzx   cx, bSPG                                            ; Convert granules into sectors
            mul     cx

            add     wSector, ax                                         ; Add to remaining sectors from previous scan (sector offset inside granule)

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

    ; --- Update member variables ---

Update:     mov     edx, pTD

            push    dwPos
            pop     (CTRSDOS PTR [edx]).dwSeekPointer                   ; Update current file pointer

            push    wSector
            pop     (CTRSDOS PTR [edx]).wSeekSector                     ; Update current relative sector

            ret
            
TD_Seek     endp

;---------------------------------------------------------------------
; TD_Read                                          Read data from file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to a receiving buffer
; dwBytes:  Number of bytes to read
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Read     proc    pTD:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            local   pVDI:DWORD
            local   abBuffer[256]:BYTE
            local   dwSeekPointer:DWORD
            local   dwFileSize:DWORD

    ; --- Initialize local variables ---
    
            mov     edx, pTD

            mov     eax, (CTRSDOS PTR [edx]).pVDI                       ; Copy pVDI to local var
            mov     pVDI, eax

            mov     eax, (CTRSDOS PTR [edx]).dwSeekPointer              ; Copy SeekPointer to local var
            mov     dwSeekPointer, eax

            mcall   TRSDOS_VTBL, GetFileSize, pTD, pFPDE                ; Get the file size
            mov     dwFileSize, eax

    ; --- Check if last seek was successful ---

Next:       mov     edx, pTD

            cmp     (CTRSDOS PTR [edx]).wSeekSector, -1
            jne     @F

            mov     eax, ERROR_INVALID_ADDRESS                          ; Attempt to access invalid address.
            jmp     Done

    ; --- Read the sector pointed by SeekSector ---

@@:         mcall   TRSDOS_VTBL, Sector2CHS, pTD, (CTRSDOS PTR [edx]).wSeekSector       ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Read, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl    ; Read sector

            or      eax, eax
            jnz     Done

    ; --- Determine the starting transfer point ---

            mov     eax, dwSeekPointer                                  ; Compute SeekPointer % 256
            and     eax, 0FFh                                           

            lea     edx, abBuffer                                       ; Advance over internal buffer pointer 
            add     edx, eax

    ; --- Determine the ending transfer point ---

            mov     ecx, dwFileSize                                     ; Compute how far SeekPointer is from the end of file
            sub     ecx, dwSeekPointer
            ja      @F
            
            mov     eax, ERROR_READ_FAULT                               ; The system cannot read from the specified device.
            jmp     Done

@@:         cmp     ecx, 256                                            ; Check if SeekPointer is within the last sector
            jae     @F
            
            mov     ecx, dwFileSize                                     ; Set the ending point to FileSize % 256
            and     ecx, 0FFh
            jmp     J1

@@:         mov     ecx, 256                                            ; Set the ending point to the end of sector

    ; --- Determine the transfer length ---

J1:         sub     ecx, eax

            cmp     ecx, dwBytes                                        ; If it is more than the remaining requested bytes to read
            jbe     @F

            mov     ecx, dwBytes                                        ; Reduce it to the number of remaining requested bytes to read

    ; --- Transfer data from the internal buffer to the user buffer ---

@@:         push    ecx
            fn      MemCopy, edx, pBuffer, ecx
            pop     ecx

    ; --- Update pointers and counters ---

            sub     dwBytes, ecx                                        ; Compute remaining bytes to read
            add     pBuffer, ecx                                        ; Advance buffer pointer
            add     dwSeekPointer, ecx                                  ; Advance file pointer

            mcall   TRSDOS_VTBL, Seek, pTD, pFPDE, dwSeekPointer        ; Move the file pointer to the most current position

            cmp     dwBytes, 0                                          ; Exit when read it all
            jnz     Next
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

TD_Read     endp

;---------------------------------------------------------------------
; TD_Write                                          Write data to file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; pBuffer:  Pointer to the data buffer
; dwBytes:  Number of bytes to write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Write    proc    pTD:DWORD, pFPDE:DWORD, pBuffer:DWORD, dwBytes:DWORD

            local   pVDI:DWORD
            local   abBuffer[256]:BYTE
            local   dwSeekPointer:DWORD
            local   dwFileSize:DWORD
            local   dwPtr:DWORD
            local   dwLen:DWORD

    ; --- Initialize local variables ---
    
            mov     edx, pTD

            mov     eax, (CTRSDOS PTR [edx]).pVDI                       ; Copy pVDI to local var
            mov     pVDI, eax

            mov     eax, (CTRSDOS PTR [edx]).dwSeekPointer              ; Copy SeekPointer to local var
            mov     dwSeekPointer, eax

            mcall   TRSDOS_VTBL, GetFileSize, pTD, pFPDE                ; Get the file size
            mov     dwFileSize, eax

    ; --- Check if last seek was successful ---

Next:       mov     edx, pTD

            cmp     (CTRSDOS PTR [edx]).wSeekSector, -1
            jne     @F

            mov     eax, ERROR_INVALID_ADDRESS                          ; Attempt to access invalid address.
            jmp     Done

    ; --- Determine the starting transfer point ---

@@:         mov     eax, dwSeekPointer                                  ; Compute SeekPointer % 256
            and     eax, 0FFh                                           

            lea     edx, abBuffer                                       ; Advance over internal buffer pointer 
            add     edx, eax
            
            mov     dwPtr, edx                                          ; Save starting point

    ; --- Determine the ending transfer point ---

            mov     ecx, dwFileSize                                     ; Compute how far SeekPointer is from the end of file
            sub     ecx, dwSeekPointer
            ja      @F
            
            mov     eax, ERROR_WRITE_FAULT                              ; The system cannot write to the specified device.
            jmp     Done

@@:         cmp     ecx, 256                                            ; Check if SeekPointer is within the last sector
            jae     @F
            
            mov     ecx, dwFileSize                                     ; Set the ending point to FileSize % 256
            and     ecx, 0FFh
            jmp     J1

@@:         mov     ecx, 256                                            ; Set the ending point to the end of sector

    ; --- Determine the transfer length ---

J1:         sub     ecx, eax

            cmp     ecx, dwBytes                                        ; If it is more than the remaining requested bytes to read
            jbe     @F

            mov     ecx, dwBytes                                        ; Reduce it to the number of remaining requested bytes to read

@@:         mov     dwLen, ecx                                          ; Save transfer length
            
    ; --- If needed, read the sector before writing to it ---

            cmp     ecx, 256                                                            ; Check if transfer length is of one entire 256-byte sector
            je      @F

            mov     edx, pTD

            mcall   TRSDOS_VTBL, Sector2CHS, pTD, (CTRSDOS PTR [edx]).wSeekSector       ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Read, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl    ; Read sector

            or      eax, eax
            jnz     Done

    ; --- Transfer data from the user buffer to the internal buffer ---

@@:         fn      MemCopy, pBuffer, dwPtr, dwLen

    ; --- Write sector ---

            mov     edx, pTD

            mcall   TRSDOS_VTBL, Sector2CHS, pTD, (CTRSDOS PTR [edx]).wSeekSector       ; Convert relative sector to Cylinder/Head/Sector

            mcall   VDI_VTBL, Write, pVDI, addr abBuffer, sizeof abBuffer, cl, dh, dl   ; Write sector

            or      eax, eax
            jnz     Done

    ; --- Update pointers and counters ---

            mov     ecx, dwLen
            
            sub     dwBytes, ecx                                        ; Compute remaining bytes to write
            add     pBuffer, ecx                                        ; Advance buffer pointer
            add     dwSeekPointer, ecx                                  ; Advance file pointer

            mcall   TRSDOS_VTBL, Seek, pTD, pFPDE, dwSeekPointer        ; Move the file pointer to the most current position

            cmp     dwBytes, 0                                          ; Exit when done
            jnz     Next

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

TD_Write    endp

;---------------------------------------------------------------------
; TD_Rename                                                Rename file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the FPDE of the file to rename
; pName:    Pointer to a 11-byte blank-padded string with the new name
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Rename   proc    pTD:DWORD, pFPDE:DWORD, pName:DWORD

    ; --- Replace the old name with the new one ---
   
            mov     eax, pFPDE
            
            fn      MemCopy, pName, addr (TD_FPDE PTR [eax]).abFileName, sizeof TD_FPDE.abFileName + sizeof TD_FPDE.abExtension

    ; --- Update the Directory Entry Code ---
        
            mcall   TRSDOS_VTBL, Hash, pTD, pName                           ; Hash the file name

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFPDE, al                     ; Set HIT slot with the new hash code

    ; --- Save Directory ---

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0

            ret
    
TD_Rename   endp

;---------------------------------------------------------------------
; TD_Delete                                                Delete file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the FPDE of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_Delete   proc    pTD:DWORD, pFPDE:DWORD

            local   bExtent:BYTE
            
    ; --- Inactivate directory entry ---
    
            mov     edx, pFPDE

            and     byte ptr (TD_FPDE PTR [edx]), not TD_FPDE_Active        ; Reset 4th bit of FPDE Attributes[0]

    ; --- Release corresponding HIT slot ---

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFPDE, 0                      ; Replace current hash code by zero

    ; --- Loop through all extents releasing allocated granules ---

            mov     bExtent, 1

Del:        mcall   TRSDOS_VTBL, GetExtPtr, pTD, pFPDE, bExtent, 0          ; Get pointer to extent

            or      eax, eax                                                ; Check if EOF
            jz      Commit

            cmp     eax, -1                                                 ; Check for error
            jne     @F

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0                 ; Restore previous directory state

            mov     eax, ERROR_FILE_CORRUPT                                 ; The file or directory is corrupted and unreadable.
            jmp     Done

@@:         mcall   TRSDOS_VTBL, FreeGAT, pTD, word ptr [eax]               ; Release granules allocated in this extent

            or      eax, eax                                                ; Check if any errors occurred
            jnz     Commit

            inc     bExtent                                                 ; Advance to next extent and repeat
            jmp     Del

    ; --- Save Directory ---
    
Commit:     mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Save directory

Done:       ret

TD_Delete   endp

;---------------------------------------------------------------------
; TD_GetLastError                  Return the last internal error code
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      Last error code
;---------------------------------------------------------------------

TD_GetLastError proc    pTD:DWORD

            mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).dwLastError
            ret

TD_GetLastError endp

;---------------------------------------------------------------------
; TD_GetDosVersion                              Return the DOS version
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      DOS version
;---------------------------------------------------------------------

TD_GetDosVersion    proc    pTD:DWORD

            mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).pDir
            movzx   eax, (TD_GAT PTR [eax]).bDosVersion
            ret
    
TD_GetDosVersion    endp

;---------------------------------------------------------------------
; TD_SetDosVersion                               Set a new DOS version
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; Version:  New DOS version to set
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetDosVersion    proc    pTD:DWORD, bVersion:BYTE

            mov     edx, pTD
            mov     edx, (CTRSDOS PTR [edx]).pDir
            
            mov     al, bVersion
            mov     (TD_GAT PTR [edx]).bDosVersion, al

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Commit changes in the Directory structure

            ret
    
TD_SetDosVersion    endp

;---------------------------------------------------------------------
; TD_GetDiskName                                  Return the disk name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pStr:     Pointer to a 11-byte string receiving buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error Code
;---------------------------------------------------------------------

TD_GetDiskName  proc    pTD:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof TD_GAT.abDiskName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER                          ; The data area passed to a system call is too small.
            jmp     Done

@@:         mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).pDir
            lea     eax, (TD_GAT PTR [eax]).abDiskName

            fn      MemCopy, eax, pStr, sizeof TD_GAT.abDiskName

            mov     eax, NO_ERROR                                           ; The operation completed successfully.

Done:       ret
    
TD_GetDiskName  endp

;---------------------------------------------------------------------
; TD_SetDiskName                                   Set a new disk name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pName:    Pointer to a 11-byte blank-padded string with the new name
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetDiskName  proc    pTD:DWORD, pName:DWORD

            mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).pDir
            lea     eax, (TD_GAT PTR [eax]).abDiskName

            fn      MemCopy, pName, eax, sizeof TD_GAT.abDiskName

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Commit changes in the Directory structure

            ret
    
TD_SetDiskName  endp

;---------------------------------------------------------------------
; TD_GetDiskDate                                  Return the disk date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pStr:     Pointer to a 8-byte string receiving buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetDiskDate  proc    pTD:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof TD_GAT.abDiskDate
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER                          ; The data area passed to a system call is too small.
            jmp     Done

@@:         mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).pDir
            lea     eax, (TD_GAT PTR [eax]).abDiskDate

            fn      MemCopy, eax, pStr, sizeof TD_GAT.abDiskDate

            mov     eax, NO_ERROR                                           ; The operation completed successfully.

Done:       ret
    
TD_GetDiskDate  endp

;---------------------------------------------------------------------
; TD_SetDiskDate                                   Set a new disk date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pDate:    Pointer to a 8-byte blank-padded string with the new date
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetDiskDate  proc    pTD:DWORD, pDate:DWORD

            mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).pDir
            lea     eax, (TD_GAT PTR [eax]).abDiskDate

            fn      MemCopy, pDate, eax, sizeof TD_GAT.abDiskDate

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Commit changes in the Directory structure

            ret
    
TD_SetDiskDate  endp

;---------------------------------------------------------------------
; TD_GetFilePos                            Return current file pointer
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Handle to File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Current pointer position or -1 for an invalid position
;---------------------------------------------------------------------

TD_GetFilePos   proc    pTD:DWORD, pFPDE:DWORD

            mov     eax, pTD
            mov     eax, (CTRSDOS PTR [eax]).dwSeekPointer
            ret
    
TD_GetFilePos   endp

;---------------------------------------------------------------------
; TD_GetFileName                                  Return the file name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to a 8-byte string receiving buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetFileName  proc    pTD:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof TD_FPDE.abFileName
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER                      ; The data area passed to a system call is too small.
            jmp     Done

@@:         mov     eax, pFPDE
            lea     eax, (TD_FPDE PTR [eax]).abFileName

            fn      MemCopy, eax, pStr, sizeof TD_FPDE.abFileName

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret
    
TD_GetFileName  endp

;---------------------------------------------------------------------
; TD_GetFileExt                              Return the file extension
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; pStr:     Pointer to a 3-byte string receiving buffer
; dwLen:    Buffer length
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_GetFileExt   proc    pTD:DWORD, pFPDE:DWORD, pStr:DWORD, dwLen:DWORD

            cmp     dwLen, sizeof TD_FPDE.abExtension
            jae     @F

            mov     eax, ERROR_INSUFFICIENT_BUFFER                      ; The data area passed to a system call is too small.
            jmp     Done

@@:         mov     eax, pFPDE
            lea     eax, (TD_FPDE PTR [eax]).abExtension

            fn      MemCopy, eax, pStr, sizeof TD_FPDE.abExtension

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret
    
TD_GetFileExt   endp

;---------------------------------------------------------------------
; TD_GetFileSize                                  Return the file size
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

TD_GetFileSize  proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (TD_FPDE PTR [edx]).wERN                       ; Get number of sectors (Ending Record Number) and multiply by 256 bytes
            shl     eax, 8
            
            movzx   ecx, (TD_FPDE PTR [edx]).bEOF                       ; Get End of File (EOF) value
            jcxz    Done

            or      eax, eax                                            ; ERN = 0? 
            jnz     @F

            mov     eax, ecx                                            ; Size = EOF
            jmp     Done

@@:         neg     cl                                                  ; Subtract (256 - EOF) from the total
            sub     eax, ecx
            
Done:       ret
    
TD_GetFileSize  endp

;---------------------------------------------------------------------
; TD_GetFileDate                                  Return the file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

TD_GetFileDate  proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE
            
            mov     cl, (TD_FPDE PTR [edx]).abAttributes[2]             ; Get second attribute byte which contains the Year and Day
            
            mov     al, cl                                              ; Extract Year
            and     eax, TD_FPDE_Year
            jz      @F
            add     eax, 1980                                           ; Add 1980 to it
            shl     eax, 16                                             ; Put Year in MSB

@@:         and     cl, TD_FPDE_Day                                     ; Extract Day    
            shr     cl, 3
;            jnz     @F
;            inc     cl                                                  ; If day is zero, set to one
            mov     al, cl                                              ; Put Day in AL

            mov     cl, (TD_FPDE PTR [edx]).abAttributes[1]             ; Get first attribute byte and extract the Month
            and     cl, TD_FPDE_Month
;            jnz     @F
;            inc     cl                                                  ; If month is zero, set to one
            mov     ah, cl                                              ; Put Month in AH

Done:       ret
    
TD_GetFileDate  endp

;---------------------------------------------------------------------
; TD_SetFileDate                                   Set a new file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwDate:   New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetFileDate  proc    pTD:DWORD, pFPDE:DWORD, dwDate:DWORD

            mov     edx, pFPDE
            mov     eax, dwDate

            and     (TD_FPDE PTR [edx]).abAttributes[1], not TD_FPDE_Month  ; Clear current date
            mov     (TD_FPDE PTR [edx]).abAttributes[2], 0

            and     ah, TD_FPDE_Month
            or      (TD_FPDE PTR [edx]).abAttributes[1], ah                 ; Set month

            shl     al, 3
            and     al, TD_FPDE_Day
            or      (TD_FPDE PTR [edx]).abAttributes[2], al                 ; Set day

            shr     eax, 16
            sub     eax, 1980
            and     al, TD_FPDE_Year
            or      (TD_FPDE PTR [edx]).abAttributes[2], al                 ; Set year

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Commit changes in the Directory structure

            ret
    
TD_SetFileDate  endp

;---------------------------------------------------------------------
; TD_GetFileAttr                            Return the file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

TD_GetFileAttr  proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (TD_FPDE PTR [edx]).abAttributes[0]                ; Get the file attributes
            
            and     al, TD_FPDE_Access                                      ; Strip the file access level

            test    (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_System     ; Test the System attribute
            jz      @F

            bts     eax, 4
            
@@:         test    (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_Invisible  ; Test the Invisible attribute
            jz      @F

            bts     eax, 8
            
@@:         test    (TD_FPDE PTR [edx]).abAttributes[1], TD_FPDE_Modified   ; Test the Modified attribute
            jz      Done

            bts     eax, 12
            
Done:       ret
    
TD_GetFileAttr  endp

;---------------------------------------------------------------------
; TD_SetFileAttr                               Set new file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwAttr:   New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_SetFileAttr  proc    pTD:DWORD, pFPDE:DWORD, dwAttr:DWORD

            mov     edx, pFPDE
            mov     eax, dwAttr

            and     (TD_FPDE PTR [edx]).abAttributes[0], not (TD_FPDE_Access+TD_FPDE_System+TD_FPDE_Invisible)
            and     (TD_FPDE PTR [edx]).abAttributes[1], not TD_FPDE_Modified

            and     al, TD_FPDE_Access                                      ; Set the new access level
            or      (TD_FPDE PTR [edx]).abAttributes[0], al

            test    dwAttr, 000000F0h                                       ; Check whether to set the System attribute
            jz      @F
            
            or      (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_System

@@:         test    dwAttr, 00000F00h                                       ; Check whether to set the Invisible attribute
            jz      @F
            
            or      (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_Invisible

@@:         test    dwAttr, 0000F000h                                       ; Check whether to set the Modified attribute
            jz      @F
            
            or      (TD_FPDE PTR [edx]).abAttributes[1], TD_FPDE_Modified

@@:         mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Commit changes in the Directory structure
    
Done:       ret
    
TD_SetFileAttr  endp

;---------------------------------------------------------------------
; TD_DirRW                          Read or Write the entire directory
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; bMode:    0:Read, 1:Write
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------      

TD_DirRW    proc    pTD:DWORD, bMode:BYTE, bFirstSector:BYTE

            local   pVDI:DWORD
            local   bTrack:BYTE
            local   bSide:BYTE
            local   bSector:BYTE
            local   bLastSector:BYTE

            mov     edx, pTD

            mov     eax, (CTRSDOS PTR [edx]).pVDI
            mov     pVDI, eax
            
            mov     al, (CTRSDOS PTR [edx]).bDirTrack
            mov     bTrack, al

            mov     al, (CTRSDOS PTR [edx]).bFirstSide
            mov     bSide, al

            mov     al, bFirstSector

            or      al, al                                              ; If parameter is 0, assure it points to the first sector (either 0 or 1)
            jnz     @F

            add     al, (CTRSDOS PTR [edx]).bFirstSector
            mov     bFirstSector, al

@@:         mov     bSector, al

            add     al, (CTRSDOS PTR [edx]).bSectorsPerTrack
            dec     al
            mov     bLastSector, al

            movzx   ecx, (CTRSDOS PTR [edx]).bDirSectors
            mov     edx, (CTRSDOS PTR [edx]).pDir

Next:       push    ecx
            push    edx

            cmp     bMode, TD_DIR_READ
            jne     DirWrite
            
DirRead:    mcall   VDI_VTBL, Read, pVDI, edx, 256, bTrack, bSide, bSector
            jmp     @F

DirWrite:   mcall   VDI_VTBL, Write, pVDI, edx, 256, bTrack, bSide, bSector

@@:         pop     edx
            pop     ecx
            
            or      eax, eax
            jnz     Done

            add     edx, 256                                            ; Advance pointer to the next memory block

            inc     bSector                                             ; Advance to the next sector

            mov     al, bSector                                         ; Reached the end of the track?
            cmp     al, bLastSector
            jbe     @F

            mov     al, bFirstSector                                    ; Restart sector count
            mov     bSector, al

            inc     bSide                                               ; Continue on the other side

@@:         loop    Next                                                ; Loop through all directory sectors

            mov     eax, NO_ERROR                                       ; The operation completed successfully.

Done:       ret

TD_DirRW    endp

;---------------------------------------------------------------------
; TD_GetFDE      Return a pointer to an available file directory entry
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

TD_GetFDE   proc    pTD:DWORD

            mcall   TRSDOS_VTBL, ScanHIT, pTD, 2, 0                     ; Scan first empty slot

            cmp     eax, -1
            jne     @F

            mov     eax, NULL
            jmp     Done

@@:         mcall   TRSDOS_VTBL, DEC2FDE, pTD, al                       ; Convert DEC into pointer to FPDE

            or      eax, eax                                            ; Check against null pointer
            jz      Done

            push    eax
            mov     ecx, sizeof TD_FPDE
@@:         mov     byte ptr [eax], 0                                   ; Fill entry with zeroes
            inc     eax
            loop    @B
            pop     eax

            push    eax
            lea     eax, (TD_FPDE PTR [eax]).Extent
            mov     ecx, 4+1
@@:         not     word ptr [eax]                                      ; Set extents as "not used" (0xFFFF)
            add     eax, sizeof TD_Extent
            loop    @B
            pop     eax

Done:       ret
            
TD_GetFDE   endp

;---------------------------------------------------------------------
; TD_GetExtPtr                             Return pointer to an extent
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to FPDE
; bExtent:  Required extent (starts at 1)
; bIEOF:    Ignore EOF (0:No, 1:Yes)
;---------------------------------------------------------------------
; EAX:      Pointer to extent or NULL if not found or -1 on error
;---------------------------------------------------------------------

TD_GetExtPtr    proc    pTD:DWORD, pFPDE:DWORD, bExtent:BYTE, bIEOF:BYTE

            mov     eax, pFPDE

Set:        lea     edx, (TD_FPDE PTR [eax]).Extent                     ; Point to the extents table

            mov     ecx, 5                                              ; Scan up to 5 extents

Get:        mov     ax, word ptr [edx]                                  ; Get one extent

            cmp     bIEOF, 0
            jne     @F
            
            cmp     al, 255                                             ; 255 means no more extents for this file
            je      EOF

@@:         cmp     al, 254                                             ; 254 means other byte is a DEC linking to a FXDE
            je      Link

            dec     bExtent                                             ; Check whether the required extent has been achieved
            jnz     Next

            mov     eax, edx                                            ; Return pointer in EAX
            jmp     Done
            
Next:       add     edx, sizeof TD_Extent                               ; Advance to the next extent
            loop    Get

EOF:        mov     eax, NULL                                           ; Return null if not found
            jmp     Done

Link:       mcall   TRSDOS_VTBL, DEC2FDE, pTD, ah                       ; Get pointer to the FXDE

            or      eax, eax                                            ; Check against null pointer
            jnz     Set

            mov     eax, -1

Done:       ret

TD_GetExtPtr    endp

;---------------------------------------------------------------------
; TD_AllocGAT                             Allocate contiguous granules
;---------------------------------------------------------------------
; pTD:          Pointer to the calling TRSDOS object
; bGranules:    Contiguous granules to allocate (up to 32)
;---------------------------------------------------------------------
; EAX:          New extent or NULL on error
; ECX:          Number of contiguous granules allocated
;---------------------------------------------------------------------

TD_AllocGAT proc    pTD:DWORD, bGranules:BYTE

            local   bGransPerCyl:BYTE
            local   bTracks:BYTE
            local   bInitCyl:BYTE
            local   bInitGran:BYTE
            local   bAllocGrans:BYTE
            
    ; --- Validate parameters ---

            cmp     bGranules, 0                                        ; Required number of granules cannot be zero
            ja      Init

            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_INVALID_PARAMETER    ; The parameter is incorrect.
            mov     eax, NULL
            jmp     Done
            
    ; --- Initialize variables ---

Init:       mov     edx, pTD

            mov     al, (CTRSDOS PTR [edx]).bGranulesPerCylinder        ; Get number of granules per cylinder
            mov     bGransPerCyl, al

            mcall   VDI_VTBL, GetTracks, (CTRSDOS PTR [edx]).pVDI       ; Get number of disk tracks
            mov     bTracks, al

            mov     bAllocGrans, 0                                      ; Clear allocated granules counter

            mov     edx, pTD
            mov     edx, (CTRSDOS PTR [edx]).pDir                       ; EDX = Pointer to GAT
            xor     eax, eax                                            ; AL = Current cylinder
            xor     ecx, ecx                                            ; CL = Current granule

    ; --- Look for the first empty slot in GAT ---

Search:     cmp     bAllocGrans, 0                                      ; Find the first empty slot then continue up to the first non-empty slot
            jnz     Allocate
            
            mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Next

            cmp     al, bTracks                                         ; Check if reached a valid slot
            jnb     Error
            
            mov     bInitCyl, al                                        ; Save initial cylinder
            mov     bInitGran, cl                                       ; Save initial granule

    ; --- Allocate as much contiguous granules as required ---

Allocate:   mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Finish

            or      [eax+edx], ch                                       ; Set granule as used

            inc     bAllocGrans                                         ; Increment count of granules allocated
            
            cmp     bAllocGrans, 32                                     ; Count of granules must fit in 5 bits
            je      Finish

            dec     bGranules                                           ; Decrement count of required granules
            jz      Finish                                              ; Exit when done

    ; --- Advance to next granule/cylinder as needed ---
    
Next:       inc     cl                                                  ; Advance to next granule
            cmp     cl, bGransPerCyl                                    ; Repeat for up to GranulesPerCylinder
            jb      Search

            mov     cl, 0                                               ; Reset current granule counter            

            inc     eax                                                 ; Advance to next cylinder

            cmp     al, bTracks                                         ; Check if still a valid slot
            jnb     Error

            cmp     eax, sizeof TD_GAT.abAllocTable                     ; Repeat for all allocation table bytes
            jb      Search

Error:      mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_DISK_FULL    ; There is not enough space on the disk.
            mov     eax, NULL
            jmp     Done

    ; --- Set return registers ---

Finish:     movzx   eax, bInitCyl                                       ; AL = Initial Cylinder

            mov     ah, bAllocGrans                                     ; AH (5 LSB) = Contiguous Granules - 1
            dec     ah

            shl     bInitGran, 5                                        ; AH (3 MSB) = Initial Granule
            or      ah, bInitGran

            movzx   ecx, bAllocGrans                                    ; ECX = Contiguous granules allocated
            
Done:       ret

TD_AllocGAT endp

;---------------------------------------------------------------------
; TD_FreeGAT                            Deallocate contiguous granules
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; wExtent:  Extent to deallocate
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD_FreeGAT  proc    pTD:DWORD, wExtent:WORD

            local   bGranulesPerCylinder:BYTE
            local   bCylinder:BYTE
            local   bGranules:BYTE
            local   bGranule:BYTE

    ; --- Initialize local variables ---
                
            mov     edx, pTD
            
            mov     al, (CTRSDOS PTR [edx]).bGranulesPerCylinder        ; Copy bGranulesPerCylinder to local var
            mov     bGranulesPerCylinder, al

            mov     ax, wExtent

            mov     bCylinder, al                                       ; Get the initial cylinder
                
            mov     bGranules, ah                                       ; Get the number of granules in the extent
            and     bGranules, 00011111y
            inc     bGranules

            mov     bGranule, ah                                        ; Get the initial granule
            shr     bGranule, 5

    ; --- Deallocate all granules in the extent ---

NextCyl:    mov     edx, pTD                                            ; Point to GAT[Cylinder]
            mov     edx, (CTRSDOS PTR [edx]).pDir
            movzx   eax, bCylinder
            add     edx, eax

ResetGran:  mov     cl, bGranule                                        ; Reset bGranule at GAT[Cylinder]
            mov     al, 1
            shl     al, cl
            not     al
            and     [edx], al

            dec     bGranules                                           ; Check whether all have been done
            jnz     @F
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done
            
@@:         inc     bGranule                                            ; Advance to the next granule

            mov     al, bGranule                                        ; Check the granules per cylinder limit
            cmp     al, bGranulesPerCylinder
            jb      ResetGran

            mov     bGranule, 0                                         ; Advance to the next cylinder
            inc     bCylinder
            
            cmp     bCylinder, sizeof TD_GAT.abAllocTable               ; Check the cylinders limit
            jb      NextCyl

            mov     eax, ERROR_FLOPPY_WRONG_CYLINDER                    ; Mismatch between the floppy disk sector ID field and the floppy disk controller track address.

Done:       ret

TD_FreeGAT  endp

;---------------------------------------------------------------------
; TD_ScanHIT                                 Scan the Hash Index Table
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; bMode:    0: Return first non-empty (optionally matching) slot
;           1: Return next non-empty (optionally matching) slot
;           2: Return first empty slot
; bHash:    Optional hash code for matching
;---------------------------------------------------------------------
; EAX:      DEC or -1 if not found
;---------------------------------------------------------------------

TD_ScanHIT  proc    pTD:DWORD, bMode:BYTE, bHash:BYTE

            local   bCols:BYTE
            local   bRows:BYTE
            local   bCol:BYTE
            local   bRow:BYTE

            mov     edx, pTD

            mov     al, (CTRSDOS PTR [edx]).bDirSectors                 ; Compute max HIT cols
            sub     al, 2
            mov     bCols, al

            mov     bRows, 256 / sizeof TD_FPDE                         ; Compute max HIT rows

            mov     ax, (CTRSDOS PTR [edx]).wLastScan                   ; Recover last scan position
            
            mov     bCol, ah
            mov     bRow, al

            mov     edx, (CTRSDOS PTR [edx]).pDir                       ; Point to the HIT
            add     edx, 256

            cmp     bMode, 1                                            ; Jump if "find next" requested
            je      Next

            mov     bCol, 0                                             ; Otherwise restart from 0,0
            mov     bRow, 0

Scan:       movzx   eax, bRow                                           ; Compute current position (Row x sizeof FPDE + Col)
            shl     al, 5
            add     al, bCol
            
            cmp     byte ptr [edx+eax], 0                               ; Test the slot
            setne   cl

            cmp     bMode, 2                                            ; Invert result if requested mode is "2"
            jne     @F

            xor     cl, 1

@@:         or      cl, cl                                              ; Jump if search result has been achieved
            jnz     Found

Next:       inc     bCol                                                ; Advance to the next column

            mov     al, bCol                                            ; Check the end of columns
            cmp     al, bCols
            jb      Scan

            mov     bCol, 0

            inc     bRow                                                ; Advance to the next row

            mov     al, bRow                                            ; Check the end of rows
            cmp     al, bRows
            jb      Scan

            mov     eax, -1                                             ; Not found: return -1
            jmp     Done

Found:      cmp     bHash, 0                                            ; Check whether a hash code has been indicated
            je      @F

            mov     cl, bHash                                           ; Compare with the value in the slot
            cmp     byte ptr [edx+eax], cl
            jne     Next

@@:         mov     ch, bCol                                            ; Save current position for next scan
            mov     cl, bRow
            
            mov     edx, pTD
            mov     (CTRSDOS PTR [edx]).wLastScan, cx
            
Done:       ret

TD_ScanHIT  endp

;---------------------------------------------------------------------
; TD_SetHIT                          Set a HIT slot to a new hash code
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the FPDE whose HIT slot is to be set
; bHash:    New hash code (0 indicates the slot is free)
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

TD_SetHIT   proc    pTD:DWORD, pFPDE:DWORD, bHash:BYTE

            mcall   TRSDOS_VTBL, FDE2DEC, pTD, pFPDE                    ; Compute DEC from pointer to FPDE

            mov     edx, pTD                                            ; Compute pointer to the corresponding HIT slot
            mov     edx, (CTRSDOS PTR [edx]).pDir
            add     edx, 256*1
            add     edx, eax                                            ; EAX = DEC

            mov     al, bHash                                           ; Set the new hash code
            mov     [edx], al

            ret

TD_SetHIT   endp

;---------------------------------------------------------------------
; TD_DEC2FDE                          Convert DEC into pointer to FPDE
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; bDEC:     Directory Entry Code
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

TD_DEC2FDE  proc    pTD:DWORD, bDEC:BYTE

            movzx   eax, bDEC

            mov     ah, al
            and     ah, 00011111y                                       ; Strip column bits to get the sector offset
            and     al, 11100000y                                       ; Strip row bits to get the entry offset (already multiplied by 32 bytes)

            mov     edx, pTD

            cmp     ah, (CTRSDOS PTR [edx]).bDirSectors                 ; Check whether this is will be a valid pointer
            jb      @F

            mov     eax, NULL                                           ; Return NULL if not
            jmp     Done

@@:         add     eax, (CTRSDOS PTR [edx]).pDir                       ; Make a pointer to the directory
            add     eax, 256*2                                          ; Skip GAT and HIT sectors

Done:       ret

TD_DEC2FDE  endp

;---------------------------------------------------------------------
; TD_FDE2DEC                          Convert pointer to FPDE into DEC
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Directory Entry Code (index to HIT slot containg the hash)
;---------------------------------------------------------------------

TD_FDE2DEC  proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pTD
            mov     eax, pFPDE
            sub     eax, (CTRSDOS PTR [edx]).pDir                       ; Subtract pointer to FPDE from pointer to directory buffer
            sub     eax, 256*2                                          ; Subtract GAT and HIT sector sizes
            shr     eax, 5                                              ; Divide by the size of each directory entry (32 bytes)
            ror     al, 3                                               ; Reorganize bits: 11111000 -> 00011111
            ret

TD_FDE2DEC  endp

;---------------------------------------------------------------------
; TD_Hash                                             Hash a file name
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFile:    Pointer to a blank-padded file name and extension
;---------------------------------------------------------------------
; AL:       Computed hash code
;---------------------------------------------------------------------

TD_Hash     proc    pTD:DWORD, pFile:DWORD

            xor     eax, eax
            mov     ecx, 11
            mov     edx, pFile
            
@@:         xor     al, byte ptr [edx]
            inc     edx
            rol     al, 1
            loop    @B
            
            or      al, al
            jnz     Done
            
            inc     al

Done:       ret

TD_Hash     endp

;---------------------------------------------------------------------
; TD_Sector2CHS                       Convert relative sector to C/H/S
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; wSector:  Relative sector number
;---------------------------------------------------------------------
; CL:       Cylinder
; DH:       Head
; DL:       Sector
;---------------------------------------------------------------------

TD_Sector2CHS   proc    pTD:DWORD, wSector:WORD

            mov     edx, pTD

            mov     al, (CTRSDOS PTR [edx]).bSectorsPerTrack            ; SectorsPerCylinder = SectorsPerTrack * Sides
            mul     (CTRSDOS PTR [edx]).bSides
            mov     cl, al                                              ; CL = SectorsPerCylinder

            mov     ax, wSector                                         ; Cylinder (AL) = (RelSector / SectorsPerCylinder) + FirstTrack
            div     cl                                                  ; AH = Remainder
            add     al, (CTRSDOS PTR [edx]).bFirstTrack
            mov     cl, al                                              ; CL = Cylinder
            
            movzx   ax, ah                                              ; Head (AL) = (Remainder / SectorsPerTrack) + FirstSide
            div     (CTRSDOS PTR [edx]).bSectorsPerTrack                ; Sector (AH) = Remainder + FirstSector
            add     al, (CTRSDOS PTR [edx]).bFirstSide
            add     ah, (CTRSDOS PTR [edx]).bFirstSector

            mov     dh, al                                              ; DH = Head
            mov     dl, ah                                              ; DL = Sector

Done:       ret

TD_Sector2CHS   endp
