;---------------------------------------------------------------------
; TRSDOS Model III OS Interface             Copyright (C) Miguel Dutra
;---------------------------------------------------------------------

.data?

;---------------------------------------------------------------------
; TRSDOS File Primary Directory Entry                                 
;---------------------------------------------------------------------

TD3_FPDE    struct
    bAttributes             BYTE        ?                               ; File attributes                   ; Patch
    bMonth                  BYTE        ?                               ; Last modification date (Month)    ; Patch
    bYear                   BYTE        ?                               ; Last modification date (Year)     ; Patch
    bEOF                    BYTE        ?                               ; EOF offset byte
    bLRL                    BYTE        ?                               ; Logical Record Length (LRL)
    abFileName              BYTE        8 dup (?)                       ; File name
    abExtension             BYTE        3 dup (?)                       ; File extension
    wOwnerHash              WORD        ?                               ; OWNER password hash
    wUserHash               WORD        ?                               ; USER password hash
    wERN                    WORD        ?                               ; Ending Record Number (ERN)
    Extent                  TD_Extent   13 dup (<>)                     ; Up to 13 extent elements          ; Patch
TD3_FPDE    ends

.data

;---------------------------------------------------------------------
; TRSDOS3 Virtual Table
;---------------------------------------------------------------------

TRSDOS3_VTBL    struct                                                  ; Declare the structure and initializers
    Load                    DWORD       TD3_Load                        ; OSI-derived functions ; Patch
    Unload                  DWORD       TD_Unload
    Find                    DWORD       TD_Find
    Open                    DWORD       TD_Open
    Create                  DWORD       TD3_Create                      ; Patch
    Seek                    DWORD       TD_Seek
    Read                    DWORD       TD_Read
    Write                   DWORD       TD_Write
    Rename                  DWORD       TD_Rename
    Delete                  DWORD       TD3_Delete                      ; Patch
    GetLastError            DWORD       TD_GetLastError
    GetDosVersion           DWORD       TD_GetDosVersion
    SetDosVersion           DWORD       TD_SetDosVersion
    GetDiskName             DWORD       TD_GetDiskName
    SetDiskName             DWORD       TD_SetDiskName
    GetDiskDate             DWORD       TD_GetDiskDate
    SetDiskDate             DWORD       TD_SetDiskDate
    GetFilePos              DWORD       TD_GetFilePos
    GetFileName             DWORD       TD_GetFileName
    GetFileExt              DWORD       TD_GetFileExt
    GetFileSize             DWORD       TD3_GetFileSize                 ; Patch
    GetFileDate             DWORD       TD3_GetFileDate                 ; Patch
    SetFileDate             DWORD       TD3_SetFileDate                 ; Patch
    GetFileAttr             DWORD       TD3_GetFileAttr                 ; Patch
    SetFileAttr             DWORD       TD3_SetFileAttr                 ; Patch
    DirRW                   DWORD       TD_DirRW                        ; TRSDOS-specific functions
    GetFDE                  DWORD       TD3_GetFDE                      ; Patch
    GetExtPtr               DWORD       TD3_GetExtPtr                   ; Patch
    AllocGAT                DWORD       TD3_AllocGAT                    ; Patch
    FreeGAT                 DWORD       TD3_FreeGAT                     ; Patch
    ScanHIT                 DWORD       TD3_ScanHIT                     ; Patch
    SetHIT                  DWORD       TD_SetHIT
    DEC2FDE                 DWORD       TD3_DEC2FDE                     ; Patch
    FDE2DEC                 DWORD       TD3_FDE2DEC                     ; Patch
    Hash                    DWORD       TD_Hash
    Sector2CHS              DWORD       TD_Sector2CHS
TRSDOS3_VTBL    ends

TRSDOS3                     TRSDOS3_VTBL    <>                          ; Define a variable of that type

.code

;---------------------------------------------------------------------
; TD3_Load                                        Initialize interface
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; pVDI:     Pointer to an initialized VDI object
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD3_Load    proc    pTD:DWORD, pVDI:DWORD

            local   abBoot[256]:BYTE
            local   pDir:DWORD
            local   bDirTrack:BYTE
            local   bDirSectors:BYTE
            local   bFirstTrack:BYTE
            local   bFirstSide:BYTE
            local   bFirstSector:BYTE
            local   bSides:BYTE
            local   bSectorsPerTrack:BYTE
            local   bGranulesPerTrack:BYTE
            local   bGranulesPerCylinder:BYTE
            local   bSectorsPerGranule:BYTE
            local   bDensity:BYTE

    ; --- Get disk information ---

            mcall   VDI_VTBL, GetFirstTrack, pVDI
            mov     bFirstTrack, al

            mcall   VDI_VTBL, GetFirstSide, pVDI
            mov     bFirstSide, al

            mcall   VDI_VTBL, GetFirstSector, pVDI
            mov     bFirstSector, al

            mcall   VDI_VTBL, GetSides, pVDI
            mov     bSides, al

            mcall   VDI_VTBL, GetDensity, pVDI
            mov     bDensity, al

    ; --- Normalize Sectors Per Track ---

            mcall   VDI_VTBL, GetSPT, pVDI                              ; Get Sectors Per Track

            cmp     al, 26                                              ; 8" DD                                              
            jb      @F

            mov     al, 26
            jmp     SetSPT
            
@@:         cmp     al, 18                                              ; 5 1/4" DD
            jb      @F

            mov     al, 18
            jmp     SetSPT
            
@@:         cmp     al, 15                                              ; 8" SD
            jb      @F

            mov     al, 15
            jmp     SetSPT
            
@@:         mov     al, 10                                              ; 5 1/4" SD

SetSPT:     mov     bSectorsPerTrack, al

    ; --- Load the boot sector to find out the directory cylinder ---

            mcall   VDI_VTBL, Read, pVDI, addr abBoot, sizeof abBoot, bFirstTrack, bFirstSide, bFirstSector
            
            or      eax, eax
            jnz     Done

            cmp     abBoot[0], 0FEh                                     ; Kind of integrity check   ; Patch
            jne     Error

            mov     al, abBoot[1]                                       ; This byte contains the directory cylinder ; Patch
            mov     bDirTrack, al
            
    ; --- Compute the number of directory sectors ---

            mov     al, bSectorsPerTrack                                ; Directory Sectors = Sectors per Track * Sides
            mul     bSides

            cmp     al, 32+2                                            ; Max sectors limited by HIT: 256 hashes / 8 FDEs per sector = 32 sectors (+2 for GAT and HIT)
            jbe     @F

            mov     al, 32+2                                            ; Set top limit

@@:         mov     bDirSectors, al
            
    ; --- Allocate memory for the entire directory ---

            shl     eax, 8                                              ; Buffer Size = DirSectors * 256 bytes per sector (TRSDOS standard)

            fn      GlobalAlloc, GPTR, eax

            or      eax, eax
            jnz     @F

            fn      GetLastError
            jmp     Done

@@:         mov     pDir, eax                                           ; Save pointer to Directory buffer

    ; --- Update member class variables ---
    
            mov     edx, pTD
            
            mov     eax, pVDI                                           ; Update pointer to Virtual Disk Interface
            mov     (CTRSDOS PTR [edx]).pVDI, eax

            mov     eax, pDir                                           ; Update pointer to Directory buffer
            mov     (CTRSDOS PTR [edx]).pDir, eax

            mov     al, bDirTrack                                       ; Update number of the directory cylinder
            mov     (CTRSDOS PTR [edx]).bDirTrack, al

            mov     al, bDirSectors                                     ; Update number of directory sectors
            mov     (CTRSDOS PTR [edx]).bDirSectors, al

            mov     al, bFirstTrack                                     ; Update number of the first track on disk
            mov     (CTRSDOS PTR [edx]).bFirstTrack, al

            mov     al, bFirstSide                                      ; Update number of the first disk side
            mov     (CTRSDOS PTR [edx]).bFirstSide, al

            mov     al, bFirstSector                                    ; Update number of the first sector on track
            mov     (CTRSDOS PTR [edx]).bFirstSector, al

            mov     al, bSides                                          ; Update number of disk sides
            mov     (CTRSDOS PTR [edx]).bSides, al

            mov     al, bSectorsPerTrack                                ; Update number of Sectors per Track
            mov     (CTRSDOS PTR [edx]).bSectorsPerTrack, al

    ; --- Load directory into memory ---

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0
            
            or      eax, eax
            jnz     Done

    ; --- Compute and validate the disk's granule units ---

            cmp     bDensity, 0                                         ; Is this disk single density?
            je      SDEN

            cmp     bDensity, 1                                         ; Is this disk single density?
            je      DDEN

            cmp     bSectorsPerTrack, 10                                ; 5 1/4" SD
            je      SDEN

            cmp     bSectorsPerTrack, 18                                ; 5 1/4" DD
            je      DDEN

            cmp     bSectorsPerTrack, 15                                ; 8" SD
            je      SDEN

            cmp     bSectorsPerTrack, 26                                ; 8" DD
            jne     Error
           
DDEN:       mov     al, 6                                               ; Set 6 Granules Per Track
            jmp     @F

SDEN:       mov     al, 5                                               ; Set 5 Granules Per Track

@@:         mov     bGranulesPerTrack, al
            mul     bSides
            mov     bGranulesPerCylinder, al

;            mov     edx, pDir                                           ; Pointer to the first directory sector (GAT)

;            mov     cl, (TD_GAT PTR [edx]).bDiskConfig                  ; Find the number of granules per track
;            and     cl, TD_GAT_Granules
;            inc     cl
;            mov     bGranulesPerTrack, cl

;            mov     al, bSides                                          ; Find the number of granules per cylinder
;            mul     cl
;            mov     bGranulesPerCylinder, al

;            cmp     al, 2                                               ; GranulesPerCylinder must be between 2 and 8
;            jb      Error

;            cmp     al, 8
;            ja      Error
            
            mov     al, bSectorsPerTrack                                ; Compute the number of sectors per granule
            div     bGranulesPerTrack
            mov     bSectorsPerGranule, al

            or      ah, ah                                              ; There must be no remainder after last division
            jnz     Error

            cmp     al, 2                                               ; Sectors per Granule for 5.25" SDEN disks  ; Patch
            je      @F
    
            cmp     al, 3                                               ; Sectors per Granule for 5.25" DDEN disks  ; Patch
            jne     Error
    
;            cmp     al, 8                                               ; Sectors per Granule for 8" SDEN disks
;            je      @F
    
;            cmp     al, 10                                              ; Sectors per Granule for 8" DDEN disks
;            jne     Error
    
    ; --- Update member class variables ---
    
@@:         mov     edx, pTD
            
            mov     al, bGranulesPerTrack                               ; Update number of Granules Per Track
            mov     (CTRSDOS PTR [edx]).bGranulesPerTrack, al

            mov     al, bGranulesPerCylinder                            ; Update number of Granules Per Cylinder
            mov     (CTRSDOS PTR [edx]).bGranulesPerCylinder, al

            mov     al, bSectorsPerGranule                              ; Update number of Sectors Per Granule
            mov     (CTRSDOS PTR [edx]).bSectorsPerGranule, al

            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done

Error:      mov     eax, ERROR_NOT_DOS_DISK                             ; The specified disk or diskette cannot be accessed.

Done:       ret

TD3_Load    endp

;---------------------------------------------------------------------
; TD3_Create                                         Create a new file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; pName:    Pointer to a 11-byte blank-padded string with the filename
; dwSize:   File size
;---------------------------------------------------------------------
; EAX:      Pointer to FPDE or NULL on error
;---------------------------------------------------------------------

TD3_Create  proc    pTD:DWORD, pName:DWORD, dwSize:DWORD

            local   pFPDE:DWORD
            local   pFDE:DWORD
;            local   pLastFDE:DWORD
            local   bGranules:BYTE
            local   bExtCount:BYTE
            local   wExtent:WORD
            local   bAllocated:BYTE

            mov     bExtCount, 0

    ; --- Find an available directory entry ---

            mcall   TRSDOS_VTBL, GetFDE, pTD

            or      eax, eax                                                ; Check against null pointer
            jnz     @F

            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
            jmp     Error

@@:         mov     pFPDE, eax
            
    ; --- Fill out the entry ---

            mov     (TD3_FPDE PTR [eax]).bAttributes, TD_FPDE_Active        ; Mark entry as active

            mov     (TD3_FPDE PTR [eax]).wOwnerHash, 5CEFh                  ; Owner password is blanks  ; Patch
            mov     (TD3_FPDE PTR [eax]).wUserHash, 5CEFh                   ; User password is blanks   ; Patch

            mov     edx, dwSize                                             ; Set End Of File (EOF)     ; Patch
            mov     (TD3_FPDE PTR [eax]).bEOF, dl
            
            shr     edx, 8                                                  ; Set Ending Record Number (ERN)
            mov     (TD3_FPDE PTR [eax]).wERN, dx
 
;            cmp     (TD_FPDE PTR [eax]).bEOF, 0                             ; If EOF is not zero increment ERN
;            je      @F
            
;            inc     (TD_FPDE PTR [eax]).wERN
            
@@:         fn      MemCopy, pName, addr (TD3_FPDE PTR [eax]).abFileName, sizeof TD3_FPDE.abFileName + sizeof TD3_FPDE.abExtension

    ; --- Update the Hash Index Table ---
        
            mcall   TRSDOS_VTBL, Hash, pTD, pName                           ; Hash the file name

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFPDE, al                     ; Update HIT slot with the new hash code
            
    ; --- Compute the required number of granules ---
    
            mov     edx, pTD
            
            mov     eax, pFPDE
            mov     pFDE, eax

            mov     eax, dwSize                                             ; Sectors = Size / 256  ; Patch
            or      al, al
            jz      @F
            add     eax, 0100h                                              ; Add 1 in case of remainder (EOF)
@@:         shr     eax, 8          
            
;            mov     ax, (TD3_FPDE PTR [eax]).wERN                           ; Granules = ERN / SectorsPerGranule
            div     (CTRSDOS PTR [edx]).bSectorsPerGranule
            or      ah, ah                                                  ; Add 1 in case of remainder
            jz      @F
            inc     al
@@:         mov     bGranules, al

    ; --- Create an extent with as much granules as possible ---
    
NextExt:    mcall   TRSDOS_VTBL, AllocGAT, pTD, bGranules

            or      eax, eax                                                ; Check against null pointer
            jz      Done
            
            mov     wExtent, ax
            mov     bAllocated, cl

    ; --- Save the new extent in the extents table ---

            inc     bExtCount

            mcall   TRSDOS_VTBL, GetExtPtr, pTD, pFDE, bExtCount, 1         ; Get pointer to next extent

            or      eax, eax                                                ; Check if null
            jnz     @F                                                      ; Patch

            mov     eax, ERROR_CANNOT_MAKE                                  ; The directory or file cannot be created.
            jmp     Error

@@:         cmp     eax, -1                                                 ; Check for error
            jne     @F

            mov     eax, ERROR_INVALID_TABLE                                ; Invalid or unknown table specified.
            jmp     Error

@@:         push    wExtent                                                 ; Save extent
            pop     word ptr [eax]

    ; --- Check if all required granules have been allocated ---
    
            mov     al, bAllocated
            sub     bGranules, al
            jnz     NextExt

    ; --- Save Directory ---
    
            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0

            or      eax, eax
            jnz     Error

            mov     eax, pFPDE
            jmp     Done

    ; --- Obtain a new directory entry ---

;AllocFDE:   mov     eax, pFDE                                               ; Save current pointer
;            mov     pLastFDE, eax

;            mcall   TRSDOS_VTBL, GetFDE, pTD                               ; Get a new pointer

;            or      eax, eax                                                ; Check against null pointer
;            jnz     @F

;            mov     eax, ERROR_HANDLE_DISK_FULL                             ; The disk is full.
;            jmp     Error

;@@:         mov     pFDE, eax                                               ; Save new pointer

    ; --- Set a forward link to the next entry ---

;            mcall   TRSDOS_VTBL, FDE2DEC, pTD, pFDE                        ; Get its corresponding DEC

;            mov     edx, pLastFDE
            
;            mov     (TD3_FPDE PTR [edx]).Link.bCylinder, 0FEh               ; 0xFE indicates a link exists
;            mov     (TD3_FPDE PTR [edx]).Link.bGranules, al                 ; This value is the DEC of the next entry
            
    ; --- Set a backward link to the previous entry ---

;            mcall   TRSDOS_VTBL, FDE2DEC, pTD, pLastFDE                    ; Get DEC of previous entry

;            mov     edx, pFDE

;            mov     (TD3_FPDE PTR [edx]).abAttributes[0], TD3_FPDE_Active+TD3_FPDE_Extended  ; Mark the new entry as active and extended
;            mov     (TD3_FPDE PTR [edx]).abAttributes[1], al                                 ; In a FXDE this byte contains the backward link

    ; --- Copy the filename and set the HIT slot ---
    
;            fn      MemCopy, pName, addr (TD3_FPDE PTR [edx]).abFileName, sizeof TD3_FPDE.abFileName + sizeof TD3_FPDE.abExtension

;            mcall   TRSDOS_VTBL, Hash, pTD, pName                          ; Hash the file name

;            mcall   TRSDOS_VTBL, SetHIT, pTD, pFDE, al                     ; Update HIT slot with the hash code
            
;            jmp     NextExt

Error:      push    eax
            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0                 ; Restore previous directory state
            pop     eax

            mov     edx, pTD
            mov     (CTRSDOS PTR [edx]).dwLastError, eax                    ; Set last error
            
            mov     eax, NULL                                               ; Return null pointer

Done:       ret

TD3_Create  endp

;---------------------------------------------------------------------
; TD3_Delete                                               Delete file
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the FPDE of the file to delete
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD3_Delete  proc    pTD:DWORD, pFPDE:DWORD

            local   bExtent:BYTE
            
;            mov     edx, pFPDE

;            and     byte ptr (TD_FPDE PTR [edx]), not TD_FPDE_Active        ; Reset 4th bit of FPDE Attributes[0]

    ; --- Release corresponding HIT slot ---

            mcall   TRSDOS_VTBL, SetHIT, pTD, pFPDE, 0                      ; Replace current hash code by zero

    ; --- Loop through all extents releasing allocated granules ---

            mov     bExtent, 1

Del:        mcall   TRSDOS_VTBL, GetExtPtr, pTD, pFPDE, bExtent, 0          ; Get pointer to extent

            or      eax, eax                                                ; Check if EOF
            jz      Fill

            cmp     eax, -1                                                 ; Check for error
            jne     @F

            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_READ, 0                 ; Restore previous directory state

            mov     eax, ERROR_FILE_CORRUPT                                 ; The file or directory is corrupted and unreadable.
            jmp     Done

@@:         mcall   TRSDOS_VTBL, FreeGAT, pTD, word ptr [eax]               ; Release granules allocated in this extent

            or      eax, eax                                                ; Check if any errors occurred
            jnz     Fill

            inc     bExtent                                                 ; Advance to next extent and repeat
            jmp     Del

    ; --- Fill directory entry with zeroes ---
    
Fill:       mov     edx, pFPDE                                              ; Patch
            mov     ecx, sizeof TD3_FPDE
@@:         mov     byte ptr [edx+ecx-1], 0
            loop    @B
            
    ; --- Save Directory ---
    
Commit:     mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0                ; Save directory

Done:       ret

TD3_Delete  endp

;---------------------------------------------------------------------
; TD3_GetFileSize                                 Return the file size
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File size
;---------------------------------------------------------------------

TD3_GetFileSize proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (TD3_FPDE PTR [edx]).wERN                      ; Get number of sectors (Ending Record Number) and multiply by 256 bytes
            shl     eax, 8
            
            movzx   ecx, (TD3_FPDE PTR [edx]).bEOF                      ; Get End of File (EOF) value
;            jcxz    Done

            add     eax, ecx                                            ; Patch

;            or      eax, eax                                            ; ERN = 0? 
;            jnz     @F

;            mov     eax, ecx                                            ; Size = EOF
;            jmp     Done

;@@:         neg     cl                                                  ; Subtract (256 - EOF) from the total
;            sub     eax, ecx
            
Done:       ret
    
TD3_GetFileSize endp

;---------------------------------------------------------------------
; TD3_GetFileDate                                 Return the file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Date (YYYYMMDD)
;---------------------------------------------------------------------

TD3_GetFileDate proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

            movzx   eax, (TD3_FPDE PTR [edx]).bYear                     ; Get Year  ; Patch

            or      al, al                                              ; If not zero
            jz      @F

            add     eax, 1900                                           ; Add 1900 to it
            shl     eax, 16

@@:         mov     ah, (TD3_FPDE PTR [edx]).bMonth                     ; Get Month, leave day as zero (info not available in TRSDOS 1.3)   ; Patch
            
;            mov     cl, (TD_FPDE PTR [edx]).abAttributes[2]             ; Get second attribute byte which contains the Year and Day
            
;            mov     al, cl                                              ; Extract Year
;            and     eax, TD_FPDE_Year
;            add     eax, 1980
;            shl     eax, 16                                             ; Put Year in EAX's MSB

;            and     cl, TD_FPDE_Day                                     ; Extract Day    
;            shr     cl, 3
;            jnz     @F
            
;            inc     cl                                                  ; If day is zero, set to one

;@@:         mov     al, cl                                              ; Put Day in AL

;            mov     cl, (TD_FPDE PTR [edx]).abAttributes[1]             ; Get first attribute byte and extract the Month
;            and     cl, TD_FPDE_Month
;            jnz     @F
            
;            inc     cl                                                  ; If month is zero, set to one
            
;@@:         mov     ah, cl                                              ; Put Month in AH

Done:       ret
    
TD3_GetFileDate endp

;---------------------------------------------------------------------
; TD3_SetFileDate                                  Set a new file date
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwDate:   New file date (YYYYMMDD)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD3_SetFileDate proc    pTD:DWORD, pFPDE:DWORD, dwDate:DWORD

            mov     edx, pFPDE
            mov     eax, dwDate

            mov     (TD3_FPDE PTR [edx]).bMonth, ah                     ; Set Month ; Patch

            shr     eax, 16
            sub     eax, 1900
            mov     (TD3_FPDE PTR [edx]).bYear, al                      ; Set Year (subtracting 1900 first) ; Patch

;            and     (TD_FPDE PTR [edx]).abAttributes[1], not TD_FPDE_Month  ; Clear current date
;            mov     (TD_FPDE PTR [edx]).abAttributes[2], 0

;            and     ah, TD_FPDE_Month
;            or      (TD_FPDE PTR [edx]).abAttributes[1], ah             ; Set month

;            shl     al, 3
;            and     al, TD_FPDE_Day
;            or      (TD_FPDE PTR [edx]).abAttributes[2], al             ; Set day

;            shr     eax, 16
;            sub     eax, 1980
;            and     al, TD_FPDE_Year
;            or      (TD_FPDE PTR [edx]).abAttributes[2], al             ; Set year
            
            mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0            ; Commit changes in the Directory structure

            ret
    
TD3_SetFileDate endp

;---------------------------------------------------------------------
; TD3_GetFileAttr                           Return the file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      File attributes (0000MISA)
;---------------------------------------------------------------------

TD3_GetFileAttr proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pFPDE

;            movzx   eax, (TD_FPDE PTR [edx]).abAttributes[0]                ; Get the file attributes
            movzx    eax, (TD3_FPDE PTR [edx]).bAttributes                  ; Get the file attributes   ; Patch
            
            and     al, TD_FPDE_Access                                      ; Strip the file access level

;            test    (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_System     ; Test the System attribute
            test    (TD3_FPDE PTR [edx]).bAttributes, TD_FPDE_System        ; Test the System attribute    ; Patch
            jz      @F

            bts     eax, 4
            
;@@:         test    (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_Invisible  ; Test the Invisible attribute
;            jz      @F
@@:         test    (TD3_FPDE PTR [edx]).bAttributes, TD_FPDE_Invisible     ; Test the Invisible attribute  ; Patch
            jz      Done

            bts     eax, 8
            
;@@:         test    (TD_FPDE PTR [edx]).abAttributes[1], TD_FPDE_Modified   ; Test the Modified attribute
;            jz      Done

;            bts     eax, 12
            
Done:       ret
    
TD3_GetFileAttr endp

;---------------------------------------------------------------------
; TD3_SetFileAttr                              Set new file attributes
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; pFPDE:    Pointer to the File Primary Directory Entry
; dwAttr:   New file attributes (0000MISA)
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD3_SetFileAttr proc    pTD:DWORD, pFPDE:DWORD, dwAttr:DWORD

            mov     edx, pFPDE
            mov     eax, dwAttr

            and     (TD3_FPDE PTR [edx]).bAttributes, not (TD_FPDE_Access+TD_FPDE_System+TD_FPDE_Invisible) ; Patch
;            and     (TD_FPDE PTR [edx]).abAttributes[0], not (TD_FPDE_Access+TD_FPDE_System+TD_FPDE_Invisible)
;            and     (TD_FPDE PTR [edx]).abAttributes[1], not TD_FPDE_Modified

            and     al, TD_FPDE_Access                                  ; Set the new access level
            or      (TD3_FPDE PTR [edx]).bAttributes, al                ; Patch
;            or      (TD_FPDE PTR [edx]).abAttributes[0], al

            test    dwAttr, 000000F0h                                   ; Check whether to set the System attribute
            jz      @F
            
            or      (TD3_FPDE PTR [edx]).bAttributes, TD_FPDE_System    ; Patch
;            or      (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_System

@@:         test    dwAttr, 00000F00h                                   ; Check whether to set the Invisible attribute
            jz      @F
            
            or      (TD3_FPDE PTR [edx]).bAttributes, TD_FPDE_Invisible ; Patch
;            or      (TD_FPDE PTR [edx]).abAttributes[0], TD_FPDE_Invisible

;@@:         test    dwAttr, 0000F000h                                   ; Check whether to set the Modified attribute
;            jz      @F
            
;            or      (TD_FPDE PTR [edx]).abAttributes[1], TD_FPDE_Modified

@@:         mcall   TRSDOS_VTBL, DirRW, pTD, TD_DIR_WRITE, 0            ; Commit changes in the Directory structure

Done:       ret
    
TD3_SetFileAttr endp

;---------------------------------------------------------------------
; TD3_GetFDE     Return a pointer to an available file directory entry
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

TD3_GetFDE  proc    pTD:DWORD

            mcall   TRSDOS_VTBL, ScanHIT, pTD, 2, 0                     ; Scan first empty slot

            cmp     eax, -1
            jne     @F

            mov     eax, NULL
            jmp     Done

@@:         mcall   TRSDOS_VTBL, DEC2FDE, pTD, al                       ; Convert DEC into pointer to FPDE

            or      eax, eax                                            ; Check against null pointer
            jz      Done

            push    eax
            mov     ecx, sizeof TD3_FPDE                                ; Patch
@@:         mov     byte ptr [eax], 0                                   ; Fill entry with zeroes
            inc     eax
            loop    @B
            pop     eax

            push    eax
            lea     eax, (TD3_FPDE PTR [eax]).Extent                    ; Patch
            mov     ecx, 13                                             ; Patch
@@:         not     word ptr [eax]                                      ; Set extents as "not used" (0xFFFF)
            add     eax, sizeof TD_Extent
            loop    @B
            pop     eax

Done:       ret
            
TD3_GetFDE  endp

;---------------------------------------------------------------------
; TD3_GetExtPtr                             Return pointer to a extent
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; pFPDE:    Pointer to FPDE
; bExtent:  Required extent (starting at 1)
; bIEOF:    Ignore EOF (0:No, 1:Yes)
;---------------------------------------------------------------------
; EAX:      Pointer to extent or NULL if not found or -1 on error
;---------------------------------------------------------------------

TD3_GetExtPtr   proc    pTD:DWORD, pFPDE:DWORD, bExtent:BYTE, bIEOF:BYTE

            mov     eax, pFPDE

Set:        lea     edx, (TD3_FPDE PTR [eax]).Extent                    ; Point to the extents table

            mov     ecx, 13                                             ; Scan up to 13 extents ; Patch

Get:        mov     ax, word ptr [edx]                                  ; Get one extent

            cmp     bIEOF, 0
            jne     @F
            
            cmp     al, 255                                             ; 255 means no more extents for this file
            je      EOF

;@@:         cmp     al, 254                                             ; 254 means other byte is a DEC linking to a FXDE
;            je      Link

@@:         dec     bExtent                                             ; Check whether the required extent has been achieved
            jnz     Next

            mov     eax, edx                                            ; Return pointer in EAX
            jmp     Done
            
Next:       add     edx, sizeof TD_Extent                               ; Advance to the next extent
            loop    Get

EOF:        mov     eax, NULL                                           ; Return null if not found
            jmp     Done

;Link:       mcall   TRSDOS_VTBL, DEC2FDE, pTD, ah                      ; Get pointer to the FXDE

;            or      eax, eax                                            ; Check against null pointer
;            jnz     Set

;            mov     eax, -1

Done:       ret

TD3_GetExtPtr   endp

;---------------------------------------------------------------------
; TD3_AllocGAT                            Allocate contiguous granules
;---------------------------------------------------------------------
; pTD:          Pointer to the calling TRSDOS object
; bGranules:    Contiguous granules to allocate (up to 32)
;---------------------------------------------------------------------
; EAX:          New extent or NULL on error
; ECX:          Number of contiguous granules allocated
;---------------------------------------------------------------------

TD3_AllocGAT    proc    pTD:DWORD, bGranules:BYTE

            local   bGransPerCyl:BYTE
            local   bTracks:BYTE
            local   bInitCyl:BYTE
            local   bInitGran:BYTE
            local   bAllocGrans:BYTE
            
    ; --- Validate parameters ---

            cmp     bGranules, 0                                        ; Required number of granules cannot be zero
            ja      Init

            mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_INVALID_PARAMETER    ; The parameter is incorrect.
            mov     eax, NULL
            jmp     Done
            
    ; --- Initialize variables ---

Init:       mov     edx, pTD

            mov     al, (CTRSDOS PTR [edx]).bGranulesPerCylinder        ; Get number of granules per cylinder
            mov     bGransPerCyl, al

            mcall   VDI_VTBL, GetTracks, (CTRSDOS PTR [edx]).pVDI       ; Get number of disk tracks
            mov     bTracks, al

            mov     bAllocGrans, 0                                      ; Clear allocated granules counter

            mov     edx, pTD
            mov     edx, (CTRSDOS PTR [edx]).pDir                       ; EDX = Pointer to GAT
            xor     eax, eax                                            ; AL = Current cylinder
            xor     ecx, ecx                                            ; CL = Current granule

    ; --- Look for the first empty slot in GAT ---

Search:     cmp     bAllocGrans, 0                                      ; Find the first empty slot then continue up to the first non-empty slot
            jnz     Allocate
            
            mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Next

            cmp     al, bTracks                                         ; Check if reached a valid slot
            jnb     Error
            
            mov     bInitCyl, al                                        ; Save initial cylinder
            mov     bInitGran, cl                                       ; Save initial granule

    ; --- Allocate as much contiguous granules as required ---

Allocate:   mov     ch, 1                                               ; Test granule state
            shl     ch, cl
            test    [eax+edx], ch
            jnz     Finish

            or      [eax+edx], ch                                       ; Set granule as used

            inc     bAllocGrans                                         ; Increment count of granules allocated

            cmp     bAllocGrans, 31                                     ; Count of granules must fit in 5 bits  ; Patch
            je      Finish

            dec     bGranules                                           ; Decrement count of required granules
            jz      Finish                                              ; Exit when done

    ; --- Advance to next granule/cylinder as needed ---
    
Next:       inc     cl                                                  ; Advance to next granule
            cmp     cl, bGransPerCyl                                    ; Repeat for up to GranulesPerCylinder
            jb      Search

            mov     cl, 0                                               ; Reset current granule counter            

            inc     eax                                                 ; Advance to next cylinder

            cmp     al, bTracks                                         ; Check if still a valid slot
            jnb     Error

            cmp     eax, sizeof TD_GAT.abAllocTable                     ; Repeat for all allocation table bytes
            jb      Search

Error:      mov     eax, pTD
            mov     (CTRSDOS PTR [eax]).dwLastError, ERROR_DISK_FULL    ; There is not enough space on the disk.
            mov     eax, NULL
            jmp     Done

    ; --- Set return registers ---

Finish:     movzx   eax, bInitCyl                                       ; AL = Initial Cylinder

            mov     ah, bAllocGrans                                     ; AH (5 LSB) = Contiguous Granules - 1
;            dec     ah                                                 ; Patch

            shl     bInitGran, 5                                        ; AH (3 MSB) = Initial Granule
            or      ah, bInitGran

            movzx   ecx, bAllocGrans                                    ; ECX = Contiguous granules allocated
            
Done:       ret

TD3_AllocGAT    endp

;---------------------------------------------------------------------
; TD3_FreeGAT                           Deallocate contiguous granules
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; wExtent:  Extent to deallocate
;---------------------------------------------------------------------
; EAX:      Error code
;---------------------------------------------------------------------

TD3_FreeGAT proc    pTD:DWORD, wExtent:WORD

            local   bGranulesPerCylinder:BYTE
            local   bCylinder:BYTE
            local   bGranules:BYTE
            local   bGranule:BYTE

    ; --- Initialize local variables ---
                
            mov     edx, pTD
            
            mov     al, (CTRSDOS PTR [edx]).bGranulesPerCylinder        ; Copy bGranulesPerCylinder to local var
            mov     bGranulesPerCylinder, al

            mov     ax, wExtent

            mov     bCylinder, al                                       ; Get the initial cylinder
                
            mov     bGranules, ah                                       ; Get the number of granules in the extent
            and     bGranules, 00011111y
;            inc     bGranules                                          ; Patch

            mov     bGranule, ah                                        ; Get the initial granule
            shr     bGranule, 5

    ; --- Deallocate all granules in the extent ---

NextCyl:    mov     edx, pTD                                            ; Point to GAT[Cylinder]
            mov     edx, (CTRSDOS PTR [edx]).pDir
            movzx   eax, bCylinder
            add     edx, eax

ResetGran:  mov     cl, bGranule                                        ; Reset bGranule at GAT[Cylinder]
            mov     al, 1
            shl     al, cl
            not     al
            and     [edx], al

            dec     bGranules                                           ; Check whether all have been done
            jnz     @F
            
            mov     eax, NO_ERROR                                       ; The operation completed successfully.
            jmp     Done
            
@@:         inc     bGranule                                            ; Advance to the next granule

            mov     al, bGranule                                        ; Check the granules per cylinder limit
            cmp     al, bGranulesPerCylinder
            jb      ResetGran

            mov     bGranule, 0                                         ; Advance to the next cylinder
            inc     bCylinder
            
            cmp     bCylinder, sizeof TD_GAT.abAllocTable               ; Check the cylinders limit
            jb      NextCyl

            mov     eax, ERROR_FLOPPY_WRONG_CYLINDER                    ; Mismatch between the floppy disk sector ID field and the floppy disk controller track address.

Done:       ret

TD3_FreeGAT endp

;---------------------------------------------------------------------
; TD3_ScanHIT                                Scan the Hash Index Table
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS object
; bMode:    0: Return first non-empty (optionally matching) slot
;           1: Return next non-empty (optionally matching) slot
;           2: Return first empty slot
; bHash:    Optional hash code for matching
;---------------------------------------------------------------------
; EAX:      DEC or -1 if not found
;---------------------------------------------------------------------

TD3_ScanHIT  proc    pTD:DWORD, bMode:BYTE, bHash:BYTE

            local   bCols:BYTE
            local   bRows:BYTE
            local   bCol:BYTE
            local   bRow:BYTE

            mov     edx, pTD

            mov     al, (CTRSDOS PTR [edx]).bDirSectors                 ; Compute max HIT cols
            sub     al, 2
            mov     bCols, al

            mov     bRows, 256 / sizeof TD3_FPDE                        ; Compute max HIT rows (sizeof TD3_FPDE = 48 bytes) ; Patch

            mov     ax, (CTRSDOS PTR [edx]).wLastScan                   ; Recover last scan position
            
            mov     bCol, ah
            mov     bRow, al

            mov     edx, (CTRSDOS PTR [edx]).pDir                       ; Point to the HIT
            add     edx, 256

            cmp     bMode, 1                                            ; Jump if "find next" requested
            je      Next

            mov     bCol, 0                                             ; Otherwise restart from 0,0
            mov     bRow, 0

Scan:       movzx   eax, bRow                                           ; Compute current position (Row x sizeof FPDE + Col)
;            shl     al, 5
            mov     cl, 48                                              ; TD3_FPDE is 48 bytes long instead of 32   ; Patch
            mul     cl
            add     al, bCol
            
            cmp     byte ptr [edx+eax], 0                               ; Test the slot
            setne   cl

            cmp     bMode, 2                                            ; Invert result if requested mode is "2"
            jne     @F

            xor     cl, 1

@@:         or      cl, cl                                              ; Jump if search result has been achieved
            jnz     Found

Next:       inc     bCol                                                ; Advance to the next column

            mov     al, bCol                                            ; Check the end of columns
            cmp     al, bCols
            jb      Scan

            mov     bCol, 0

            inc     bRow                                                ; Advance to the next row

            mov     al, bRow                                            ; Check the end of rows
            cmp     al, bRows
            jb      Scan

            mov     eax, -1                                             ; Not found: return -1
            jmp     Done

Found:      cmp     bHash, 0                                            ; Check whether a hash code has been indicated
            je      @F

            mov     cl, bHash                                           ; Compare with the value in the slot
            cmp     byte ptr [edx+eax], cl
            jne     Next

@@:         mov     ch, bCol                                            ; Save current position for next scan
            mov     cl, bRow
            
            mov     edx, pTD
            mov     (CTRSDOS PTR [edx]).wLastScan, cx
            
Done:       ret

TD3_ScanHIT  endp

;---------------------------------------------------------------------
; TD3_DEC2FDE                         Convert DEC into pointer to FPDE
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; bDEC:     Directory Entry Code
;---------------------------------------------------------------------
; EAX:      Pointer to directory entry or NULL on error
;---------------------------------------------------------------------

TD3_DEC2FDE proc    pTD:DWORD, bDEC:BYTE

    ; --- FDE is 48 bytes long (5 FDE per sector) and 16 bytes are unused ---
    
            movzx   eax, bDEC
            
;            mov     ah, al
;            and     ah, 00011111y                                       ; Strip column bits to get the sector offset
;            and     al, 11100000y                                       ; Strip row bits to get the entry offset (already multiplied by 32 bytes)

            mov     dl, 5                                               ; Sector (AH) = DEC / 5 (entries per sector)    ; Patch
            div     dl
            xchg    ah, al
            
            mov     dl, al                                              ; Entry (AL) = Remainder * 48 (entry size)  ; Patch
            shl     al, 4
            shl     dl, 5
            add     al, dl

            mov     edx, pTD

            cmp     ah, (CTRSDOS PTR [edx]).bDirSectors                 ; Check whether this is will be a valid pointer
            jb      @F

            mov     eax, NULL                                           ; Return null if not
            jmp     Done

@@:         add     eax, (CTRSDOS PTR [edx]).pDir                       ; Make a pointer to the directory
            add     eax, 256*2                                          ; Skip GAT and HIT sectors

Done:       ret

TD3_DEC2FDE endp

;---------------------------------------------------------------------
; TD3_FDE2DEC                         Convert pointer to FPDE into DEC
;---------------------------------------------------------------------
; pTD:      Pointer to the calling TRSDOS3 object
; pFPDE:    Pointer to the File Primary Directory Entry
;---------------------------------------------------------------------
; EAX:      Directory Entry Code (index to HIT slot containg the hash)
;---------------------------------------------------------------------

TD3_FDE2DEC proc    pTD:DWORD, pFPDE:DWORD

            mov     edx, pTD
            mov     eax, pFPDE
            sub     eax, (CTRSDOS PTR [edx]).pDir                       ; Subtract pointer to FPDE from pointer to directory buffer
            sub     eax, 256*2                                          ; Subtract GAT and HIT sector sizes

            mov     edx, eax                                            ; Sector = PTR / 256    ; Patch
            shr     edx, 8

            sub     ah, dl                                              ; Entry = (PTR % 256) / 48  ; Patch
            mov     cl, 48
            div     cl

            xchg    al, dl                                              ; DEC = (Sector * 5) + Entry    ; Patch
            mov     cl, 5
            mul     cl
            add     al, dl

            ret

TD3_FDE2DEC endp

