.code

;---------------------------------------------------------------------
; Dump                                 Dump data in both hex and ascii
;---------------------------------------------------------------------
; pData:    Pointer to the data buffer
; dwLen:    Length of the data buffer
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

Dump        proc    pData:DWORD, dwLen:DWORD

            local   Line[65]:BYTE
            
            mov     Line[sizeof Line - 1], 0

            push    ebx
            push    esi
            push    edi
            
            mov     ebx, _T("0123456789ABCDEF")
NewLine:    lea     esi, Line
            lea     edi, Line+48

            mov     ecx, sizeof Line - 1
@@:         mov     byte ptr [esi+ecx-1], ' '
            loop    @B

Get:        mov     edx, pData

            mov     al, [edx]
            shr     al, 4
            xlatb
            mov     [esi], al

            mov     al, [edx]
            and     al, 0Fh
            xlatb
            mov     [esi+1], al

            mov     al, [edx]
            cmp     al, ' '
            ja      @F
            mov     al, '.'
@@:         mov     [edi], al

            add     esi, 3
            inc     edi
            inc     pData
            dec     dwLen

            lea     eax, Line+64
            cmp     eax, edi
            je      @F
            
            cmp     dwLen, 0
            jne     Get

@@:         fn      StdOut, addr Line
            fnc     StdOut, "\n"

            cmp     dwLen, 0
            jne     NewLine
            
            pop     edi
            pop     esi
            pop     ebx

Done:       ret

Dump        endp

;---------------------------------------------------------------------
; AddArg                 Insert pointer in the argument pointers array
;---------------------------------------------------------------------
; pArg:     Pointer to the argument to add
; pArgV:    Pointer to the argument pointer array (argv[16]:DWORD)
; pArgC:    Pointer to the argument counter variable (argc:BYTE)
;---------------------------------------------------------------------
; EAX:      Pointer to the same argument
;---------------------------------------------------------------------

AddArg      proc    pArg:DWORD, pArgV:DWORD, pArgC:DWORD

            mov     edx, pArgC
            cmp     byte ptr [edx], 16
            jnb     Done

            mov     eax, sizeof DWORD                                   ; Compute the address of the next slot
            mul     byte ptr [edx]
            add     eax, pArgV

            mov     edx, pArg                                           ; Save pointer to the argument
            mov     [eax], edx

            mov     edx, pArgC
            inc     byte ptr [edx]                                      ; Increment the argument counter

Done:       mov     eax, pArg
            ret

AddArg      endp

;---------------------------------------------------------------------
; MakeArgList                   Build a list of command-line arguments
;---------------------------------------------------------------------
; pArgV:    Pointer to the argument pointer array (argv[16]:DWORD)
; pArgC:    Pointer to the argument counter variable (argc:BYTE)
; pCmd:     Pointer to the command-line string
;---------------------------------------------------------------------
; No return value
;---------------------------------------------------------------------

MakeArgList proc    pArgV:DWORD, pArgC:DWORD, pCmd:DWORD

            local   IgnoreBlanks:BYTE
            local   QuotesOpen:BYTE

            mov     IgnoreBlanks, 1
            mov     QuotesOpen, 0

            mov     eax, pArgC
            mov     byte ptr [eax], 0
            
            mov     eax, pCmd

Scan:       cmp     byte ptr [eax], 0                                   ; End of line?
            je      Done

            cmp     byte ptr [eax], ' '                                 ; Is blank?
            sete    dl

            cmp     byte ptr [eax], '"'                                 ; Is double quotes?
            sete    dh

            shl     dl, 1                                               ; Assemble a decision bitmap
            add     dl, dh
            shl     dl, 1
            add     dl, IgnoreBlanks
            shl     dl, 1
            add     dl, QuotesOpen

            cmp     dl, 1000y                                           ; End of argument
            je      EndArg

            cmp     dl, 0010y                                           ; Beginning of new argument
            je      NewArg

            cmp     dl, 0110y                                           ; Opening quotes
            je      NewQArg

            cmp     dl, 0111y                                           ; Closing quotes
            je      EndQArg

Next:       inc     eax                                                 ; Advance to the next char
            jmp     Scan

EndArg:     mov     byte ptr [eax], 0                                   ; Add trailing null
            mov     IgnoreBlanks, 1                                     ; Ignore blanks until next argument
            jmp     Next

NewArg:     fn      AddArg, eax, pArgV, pArgC                           ; Add pointer to arguments list
            mov     IgnoreBlanks, 0                                     ; Stop ignoring blanks
            jmp     Next

NewQArg:    inc     eax                                                 ; Skip the quotes mark
            fn      AddArg, eax, pArgV, pArgC                           ; Add pointer to arguments list
            mov     QuotesOpen, 1                                       ; Flag the open quotes
            mov     IgnoreBlanks, 1                                     ; Ignore blanks inside quotes
            jmp     Next

EndQArg:    mov     byte ptr [eax], 0                                   ; Remove the quotes mark
            mov     QuotesOpen, 0                                       ; Reset the open quotes flag
            jmp     Next

Done:       ret

MakeArgList endp

;---------------------------------------------------------------------
; ToLower                                 Convert string to lower case
;---------------------------------------------------------------------
; pStr:     Pointer to a null-terminated string
;---------------------------------------------------------------------
; EAX:      Pointer to the same string
;---------------------------------------------------------------------

ToLower     proc    pStr:DWORD

            mov     eax, pStr
            
Scan:       mov     dl, [eax]

            cmp     dl, 0
            je      Done

            cmp     dl, 'A'
            jb      Next

            cmp     dl, 'Z'
            ja      Next

            or      byte ptr [eax], 'a' - 'A'

Next:       inc     eax
            jmp     Scan

Done:       mov     eax, pStr

            ret

ToLower     endp

;---------------------------------------------------------------------
; PrintError                     Print formatted system error messages
;---------------------------------------------------------------------
; dwError:  Error code to print
;---------------------------------------------------------------------
; EAX:      Original error code
;---------------------------------------------------------------------

PrintError  proc    dwError:DWORD

            local   pMessage:PTR

            fn      FormatMessage, FORMAT_MESSAGE_FROM_SYSTEM+FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, dwError, 0, addr pMessage, 0, NULL

            or      eax ,eax
            jz      @F

            fn      StdOut, pMessage

@@:         fn      LocalFree, pMessage

            mov     eax, dwError
            
            ret

PrintError  endp

